#line 1 "note.yc"
#include "abcp_.h"
/*
  Note parsing

  The following function will return "nominal" characteristics for a note.
  With "nominal", I mean that the real values could depend from other factors.

  A note with pitch "b" is actually a "_b" if the key is Fmaj but the function
  abcNoteAccidental("b") always returns "NONE".

  Returns the uppercase ASCII representation of the pitch (i.e. 'A'-'G')
  or '?' if no note is found.

  Returns an ASCII char representing the accidentals
  n=none, N=natural, f=flat, F=double flat, s=sharp, S=double sharp

 The "nominal" Octave of the note (no trasposition is considered)
   0 C,,,,   -> B,,,
   1 C,,,
   2 C,,    -> B,
   3 C,     -> B
   4 C   -> b
   5 c   -> b'
   6 c'  -> b''
   7 c''
   8 c'''
   9 c'''' -> b'''''


 and so on.

  The maximum extension for octaves is -32768 .. 32767 but I question the usefulness
of using such  extended range!

  A4 = 440Hz

  The function manages equalities like: "C' = c" , "A', = A,' = A", ...


 The "nominal" note length is expressed as a fraction of the
 default length as stated by the L: field

 Also microtones are supported in the same form as abcm2ps:
   ^3/4C _/f'

 As an extension to this charachter ' and , can be used:

   ^,F = ^1/4F
   ^/F = ^1/2F
   ^'F = ^3/4F

   |------------ 1 semitone ---------------|

   |---------,---------/---------'---------|

   |-- 1/4 --|-- 1/4 --|-- 1/4 --|-- 1/4 --|


   _F _,F _/F _'F F ^,F ^/F ^'F ^F

And also the syntax '_^' and '^_' is supported:

   _F     _^F     F     ^_F     ^F

Errors

1= Pitch not found
2= Octave out of range  0 .. 9
3= Length out of range   0 .. 255
4=
5=
6=


  <accidentals><microtone><pitch><octave><duration>



*/


int abc_Note(abcScanner *s)
{
  unsigned short n=0;
  unsigned char *t;

  abcNote(s,duration) = 1 * abcTICKS;

  abcNote(s,octave)=4;

  if (abcNote(s,pitch >= 'a')) abcNote(s,octave)++ ;
  abcNote(s,pitch) = toupper(abcNote(s,pitch));

{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"[',]+" )
        ":\204\5\200\20/\204\5\200\20\0\0",
    NULL};
#line 94
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
    case 1 : t=YYSTART(0);
             n=YYLEN(0);
  }} /* abcStream(s */


  while (n > 0) {
    abcNote(s,octave) = abcNote(s,octave) + (t[n-1] == ','? -1: +1);
    n--;
  }


{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"\d+" )
        ":D/D\0\0",
    NULL};
#line 105
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
    case 1 : abcNote(s,duration) = atoi(YYSTART(0)) * abcTICKS;
  }} /* abcStream(s */


{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"/(\d+)" )
        ":Q/<\1D/D\0>\1\0",
_Y_(  2,"/+" )
        ":Q//Q/\0\0",
    NULL};
#line 109
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {

    case 1 :
         n=atoi(YYSTART(1));
         if (n == 0) n=2;
         abcNote(s,duration) /= n;
         break;


    case 2 :  abcNote(s,duration) /= (1 << YYLEN(0));
              break;
  }} /* abcStream(s */


  s->token  = T_NOTE;
  s->toklen = abcCursor(s) - s->tokstart ;
  return 0;

  YYSTATE(error) {
    abcNote(s,error)=1;
  }

  return 0;
}


int abc_Rest(abcScanner *s, char *start)
{
  abcRest(s,duration) = 1 * abcTICKS;

  abcRest(s,type) = *start;
  s->tokstart=start;

{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"\d+" )
        ":D/D\0\0",
    NULL};
#line 141
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
    case 1 : abcRest(s,duration) = atoi(YYSTART(0)) * abcTICKS;
  }} /* abcStream(s */


{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"/(\d+)" )
        ":Q/<\1D/D\0>\1\0",
_Y_(  2,"/+" )
        ":Q//Q/\0\0",
    NULL};
#line 145
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
    case 1 : abcRest(s,duration) /= atoi(YYSTART(1));
               break;

    case 2 :  abcRest(s,duration) /= (1 << YYLEN(0));
              break;
  }} /* abcStream(s */


  s->token = T_REST;
  s->toklen= abcCursor(s) - s->tokstart;
  return 0;
}

int abc_Annotation(abcScanner *s)
{
  unsigned short len=1;

  s->toklen=1;
  s->tokstart = abcCursor(s) - 1;

{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"[@<>_^](-?\d*),?(-?\d*)\s*" )
        ":\207\13P\1\0\0\300<\1-Q-\0/D\0>\1-Q,\0<\2-Q-\0/D\0>\2/F\0\0",
    NULL};
#line 165
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
    case 1 :
           if (YYLEN(1)>0) abcAnnotation(s,xoffset) = atoi(YYSTART(1));
           if (YYLEN(2)>0) abcAnnotation(s,yoffset) = atoi(YYSTART(2));
           abcAnnotation(s,position) = *YYSTART(0);
           s->toklen += YYLEN(0);
  }} /* abcStream(s */


  abcAnnotation(s,string) = abcCursor(s);
  len = 0;
  YYSTATE (string) {
{ /* abcStream(s */
    static const unsigned char *_Y[] = {
_Y_(  1,"\\[^\c]" )
        ":Q\134\200`\1\0",
_Y_(  2,"[^\"\\]+" )
        ":\244k\373\377\377\377\377\377\377\357/\244k\373\377\377\377\377\377\377\357\0\0",
_Y_(  3,"\\" )
        ":Q\134\0",
_Y_(  4,"[\0\c]" )
        ":\241\1\0\0",
_Y_(  5,"\"" )
        ":Q\42\0",
    NULL};
#line 176
    switch (yyswitch((unsigned char **)_Y,abcStream(s)) ) {
      case 1 :
      case 2 :
      case 3 : len += YYLEN(0);
                   YYGOTO(string);

      case 4 : abcAnnotation(s,error) = 1;
      case 5 : len ++;
                   break;
    }} /* abcStream(s */

  }

  abcAnnotation(s,string_len) = len - 1;
  if (len == 1)
    abcAnnotation(s,string) = abcEmptyString;
  s->toklen += len;

  if (s->state == (S_FIELD | 's'))
    s->token = T_SYM_ANNOTATION;
  else
    s->token = T_ANNOTATION;

  return 0;
}
