/*  YRX - re_exec.c
**  (C) 2005 by Remo Dentato
**
** Permission to use, copy, modify and distribute this code and
** its documentation for any purpose is hereby granted without fee,
** provided that the above copyright notice appear in all copies
** and supporting documentation.
** 
** Copyright holder makes no representations about the suitability
** of this software for any purpose. It is provided "as is" without
** express or implied warranty.
**
** This code is a derivative work of the PUBLIC DOMAIN regex routines
** by Ozan Yigit: http://www.cs.yorku.ca/~oz/
**
*/

#include "re_.h"

static unsigned char *bol;
static unsigned char *bopat[MAXTAG];
static unsigned char *eopat[MAXTAG];
static unsigned char *goal;
static unsigned char escape=0;

static char *pmatch(unsigned char *, unsigned char *);

/*
  .* Character class
  
  .** Representation

  It is assumed that characters are represented with 8 bit (e.g. ISO-8951-x).
  Each character in a character class is represented by a bit, hence an array of
32 bytes are needed (fig. '?bytes).

.fig '!bytes
             1           2            3
  1234 5678 9012 3456 7890 1234 5678 9012
  0000 0000 0000 0000 0000 0000 0000 0000
            a^          ^b
..
  
  To save space, if the first bytes are all 0 they could be omitted and
an index to the first non-0 byte could be stored ('|a| in the figure). 

  The same goes if the last bytes are all 0. They could be omitted and 
the index '|b| could be stored instead. 

  Of course we could apply the same reasoning the eliminate a sequence of 
1 bytes.

  So the new representation is:
  
.fig '!bytes2

        Aa Bb 000 0000 000
..

   
  Where '|a| and '|b| are the indexes described above, '|A| and '|B| are
value of omitted bytes (0 or 1).

  There's another feature character classes should have. They should respect
the locale settings.

  In my locale, for example, isdigit is true for more characters than [0-9].
  
  To correctly represent a character class as [\d+-], a separate byte is used:
  
.fig '!bytes3

        Aa Bb CC 00 00 00 00 00
..

  Each bit in CC represents a class defined by a isxxx() C function. We also
need to remember if we want to include or exclude those characters from our
class (like in [^ABC\d]).

  The following figure details how character class are stored. 
.fig 

  CCL CC2 ISA n_bytes
  
  CCL   10Aa aaaa
        | ||____|
        | |   +---- first byte represented
        | +-------- bytes before a are 0/1
        +---------- Introduces a Character class
                         
  CCL   IMBb bbbb
        ||||____|
        |||   +---- last byte represented
        ||+-------- bytes after b are 0/1
        |+--------- Include/exclude ISA classes
        +---------- 1 if ISA byte present
        
  ISA   psxd qlua
        |||| ||||
        |||| |||+-- isalpha
        |||| ||+--- isupper
        |||| |+---- islower
        |||| +----- isalnum
        |||+------- isdigit
        ||+-------- isxdigit
        |+--------- isspace
        +---------- ispunct
..        
  
*/

/*
  .** CCLSKIP()
  The '{CCLSKIP} function computes the number of bytes needed to
skip over a character class representation.

.fig
     Aa Bb CC 00 00 00 00
    p^                    ^CCLSKIP(p)

         
     Aa Bb CC II 00 00 00 00
    p^                       ^CCLSKIP(p)
..

*/
int CCLSKIP(char *p)
{
  return ((p[1] & 0x1F) - (p[0] & 0x1F) +2) + ((p[1] & 0x80) >> 7);
}

/* .** isa()
*/

int re_isa(unsigned char c, unsigned char what)
{ 
  switch (what) {
    case ALPHA  :  return(isalpha(c));
    case ANYUPR :  return(isupper(c));
    case ANYLWR :  return(islower(c));
    case ALNUM  :  return(isalnum(c));
    case DIGIT  :  return(isdigit(c));
    case XDIGIT :  return(isxdigit(c));
    case SPACE  :  return(isspace(c));
    case PUNCT  :  return(ispunct(c));
    case WORDC  :  return(isalnum(c) || (c == '_'));
    case ANY    :  return(c);
    case CTRL   :  return(c && (c == (c&0x7)));
    case SPCTAB :  return(c == ' ' || c == '\t');
  }
  return 0;
}

/* .** isinset()


*/

int re_isinset(unsigned char *x,int y)
{
  int n;
  
  /* .{ Check character class
   * Note that this is dependent on the codes assigned to 
   * the constants in '<re_.h>. Do not modify!
   */
  if (x[1] & 0x80) {
    for (n=0; n<8; n++) {
       if ((x[2] & (1<<n)) && (re_isa(y,0x41+n)))
         return !(x[1] & 0x60);
    }  
  }
  /* .} */
  
  /* .{ Check into first omitted bytes */
  n= ((*x & 0x1F)*8);
  if ( y < n) return (*x & 0x20);
  /* .}
   */
     
  /* .{ Check into last omitted bytes */
  x++;
  if ( y > ((*x & 0x1F)*8 + 7) ) return (*x & 0x20);
  /* .} */
  
  /* .{ Look into the bitmap */
  if (*x & 0x80) x++;  /* skip ISA byte if present */
  x++;
  y -= n;
  return ((x)[((y)&BLKIND)>>3] & bitarr[(y)&BITIND]);
  /* .} */
  
}

/*
 * re_exec:
 *   execute nfa to find a match.
 *
 *  special cases: (nfa[0])  
 *    BOL
 *      Match only once, starting from the
 *      beginning.
 *    CHR
 *      First locate the character without
 *      calling pmatch, and if found, call
 *      pmatch for the remaining string.
 *    END
 *      re_comp failed, poor luser did not
 *      check for it. Fail fast.
 *
 *  If a match is found, bopat[0] and eopat[0] are set
 *  to the beginning and the end of the matched fragment,
 *  respectively.
 *
 */

int
re_exec(unsigned char *lp, unsigned char *nfa)
{
  register unsigned char c;
  register unsigned char *ep = 0;
  register unsigned char *ap = nfa;

  bol = lp;

  bopat[0] = 0;
  bopat[1] = 0;
  bopat[2] = 0;
  bopat[3] = 0;
  bopat[4] = 0;
  bopat[5] = 0;
  bopat[6] = 0;
  bopat[7] = 0;
  bopat[8] = 0;
  bopat[9] = 0;

  goal = 0;
  
  switch(*ap) {

  case BOL:      /* anchored: match from BOL only */
    ep = pmatch(lp,ap);
    break;
  case CHR:      /* ordinary char: locate it fast */
    c = *(ap+1);
    while (*lp && *lp != c)
      lp++;
    if (!*lp) return 0;    /* if EOS, fail, else fall thru. */
    
  default:      /* regular matching all the way. */
#ifdef OLD
    while (*lp) {
      if ((ep = pmatch(lp,ap))) break;
      lp++;
    }
#else  /* match null string */
    do {
      if ((ep = pmatch(lp,ap))) break;
      if (escape && (escape == *lp)) lp++;
      lp++;
    } while (*lp);
#endif
    break;
  case END:      /* munged automaton. fail always */
    return 0;
  }
  if (!ep)
    return 0;

  if (goal) ep=goal;

  bopat[0] = lp;
  eopat[0] = ep;
  return 1;
}

/* 
 * pmatch: internal routine for the hard part
 *
 *   This code is partly snarfed from an early grep written by
 *  David Conroy. The backref and tag stuff, and various other
 *  innovations are by oz.
 *
 *  special case optimizations: (nfa[n], nfa[n+1])
 *    CLO ANY
 *      We KNOW .* will match everything upto the
 *      end of line. Thus, directly go to the end of
 *      line, without recursive pmatch calls. As in
 *      the other closure cases, the remaining pattern
 *      must be matched by moving backwards on the
 *      string recursively, to find a match for xy
 *      (x is ".*" and y is the remaining pattern)
 *      where the match satisfies the LONGEST match for
 *      x followed by a match for y.
 *    CLO CHR
 *      We can again scan the string forward for the
 *      single char and at the point of failure, we
 *      execute the remaining nfa recursively, same as
 *      above.
 *
 *  At the end of a successful match, bopat[n] and eopat[n]
 *  are set to the beginning and end of subpatterns matched
 *  by tagged expressions (n = 1 to 9).  
 *
 */

void  
re_fail(char *s, char c)
{
  fprintf(stderr, "%s [opcode %o]\n", s, c);
  exit(1);
}

static char *qstr(unsigned char *e)
{
  char c;
  c=*e++;
  if (c == '\'' || c == '"') {
      while (1) {
        if (*e == '\0') break;
        if (*e == c) {
          if (*(e+1) == c) e+=2;
          else break;
        }
        if ((*e == '\\') && (*(e+1) != '\0')) e++;
        e++;
      }
  }
  else
    return NULL;
    
  if (*e != c) return NULL;
  return (e+1);
}

static char *pmatch(unsigned char *lp, unsigned char *ap)
{
  register int op, c, n=ANYSKIP,k;
  register unsigned char *e;      /* extra pointer for CLO */
  register unsigned char *bp;    /* beginning of subpat.. */
  register unsigned char *ep;    /* ending of subpat..   */
  unsigned char *are=0;          /* to save the line ptr. */
  /*unsigned char *old_ap=ap; */

  
  while ((op = *ap++) != END) {
    c=0;
    
    if (op & CCL) {
       c = *lp++;
       if (c == escape) lp++;
       else if (!re_isinset(ap-1,c)) return 0;
       ap += CCLSKIP(ap-1);
    }
    else if ((op & 0xF0) == 0x40) {
       c = *lp++;
       if (c == escape) lp++;
       else if (!re_isa(c,op)) return 0;
    }
    else switch(op) {
      case END: break;
      
      case ESC:  escape=*ap++;
                 if (escape==' ') escape = '\0';
                 break;
                 
      case CHI:  
      case CHR:  if (escape && (escape == *lp)) {
                   lp+=2;
                   ap++;
                 }
                 else {
                   c=*lp++;
                   if (op == CHI) c = tolower(c);
                   if (c != *ap++) return 0;
                 }
                 break;
      
      case STR:  while (*ap)
                   if (*ap++ != *lp++) return 0;
                 ap++;
                 break;
                 
      case STI:  while (*ap)
                   if (*ap++ != tolower(*lp++)) return 0;
                 ap++;
                 break;               
                 
      case BOL:  if (lp != bol) return 0;  break;
                 
      case EOL:  if (*lp) return 0;        break;
                 
      case BOT:  bopat[*ap++] = lp;        break;
                 
      case EOT:  eopat[*ap++] = lp;        break;
        
      case GOAL: goal = lp;                break;
        
      case BOW:  if ((lp!=bol && re_isa(lp[-1],WORDC)) || !re_isa(*lp,WORDC))
                   return 0;
                 break;
        
      case EOW:  if (lp==bol || !re_isa(lp[-1],WORDC) || re_isa(*lp,WORDC))
                   return 0;
                 break;
        
      case REF:  n = *ap++;
                 bp = bopat[n];
                 ep = eopat[n];
                 while (bp < ep)
                   if (*bp++ != *lp++) return 0;
                 break;
        
      case QSTR: if ((e=qstr(lp)) == NULL ) return 0;
                 lp=e;
                 break;
          
      case NINT: e=lp;
                 if ((*e == '+' || *e == '-') && isdigit(e[1])) e+=2;
                 while (isdigit(*e)) e++;
                 if (e==lp) return 0;
                 lp=e;
                 break;
          
      case SPCS: while (isspace(*lp)) lp++;
                 break;
                 
      case IDENT: e=lp;
                 while (re_isa(*e,WORDC)) e++;
                 if (e==lp) return 0;
                 lp=e;
                 break;
                 
      case NHEX: e=lp;
                 if ((*e == '0') && (e[1] == 'x' || e[1] == 'X') && isxdigit(e[2])) e+=3;
                 while (isxdigit(*e)) e++;
                 if (e==lp) return 0;
                 lp=e;
                 break;
          
      case OPT:
      case CLO:  are = lp;
                 k=0;
                 if ((*ap & 0xF0) == 0x40) {  /* 0100 xxxx */
                   while (*lp) {
                     if ((*lp == escape) && (lp[1])) lp+=2;
                     else if (re_isa(*lp,*ap)) lp++;
                     else break;
                     if (op == OPT) break;
                   }
                   n = ANYSKIP;
                 }
                 else if (*ap & CCL) {
                   while (*lp) {
                     if ((*lp == escape) && (lp[1])) lp+=2;
                     else if (re_isinset(ap,*lp)) lp++;
                     else break;
                     if (op == OPT) break;
                   }
                   n = CCLSKIP(ap)+2;
                 }
                 else {
                   switch(*ap) {

                     case CHI:  c = *(ap+1);
                                while (*lp) {
                                  if ((*lp == escape) && (lp[1])) lp+=2;
                                  else if (c == tolower(*lp)) lp++;
                                  else break;
                                  if (op == OPT) break;
                                }
                                n = CHRSKIP;
                                break;
                                
                     case CHR:  c = *(ap+1);
                                while (*lp)  {
                                  if ((*lp == escape) && (lp[1])) lp+=2;
                                  else if (c == *lp) lp++;
                                  else break;
                                  if (op == OPT) break;
                                }
                                n = CHRSKIP;
                                break;
                                
                     default:   re_fail("closure: bad nfa.", *ap);
                                return 0;
                   }
                 }
                 ap += n;
                 while (lp >= are) {
                   if ((e = pmatch(lp, ap)))
                     return e;
                   --lp;
                   if ((lp > are) && (lp[-1] == escape)) --lp;
                   /*fprintf(stderr, "++ retry '%s' at %d\n",lp,ap-old_ap);*/
                 }
                 return 0;
                 
      default:   re_fail("re_exec: bad nfa.", op);
                 return 0;
    }
  }
  return lp;
}

char *re_ncpy(unsigned char *dst,unsigned char n,int l)
{
  unsigned char *bp;
  unsigned char *r=dst;
  int m;
  
  if (dst!= NULL) {
    if (n<10) {
      bp=bopat[n];
      if (bp) {
        m=eopat[n]-bopat[n]+1; 
        if ((l == 0) || (m < l)) l=m;
        while (--l)
          *dst++ = *bp++;
      }
    }
    *dst='\0';
  }
  return r;
}

char *re_cpy(unsigned char *dst,unsigned char n)
{
  return(re_ncpy(dst,n,0));
}

int re_len(unsigned char n)
{
  if (n>9 || bopat[n]==NULL || *bopat[n] == '\0') return 0; 
  return (eopat[n]-bopat[n]);
}

static char *emptystr="";

char *re_start(unsigned char n)
{
  if (n>9 || bopat[n] == NULL) return emptystr;
  return bopat[n];
}

char *re_end(unsigned char n)
{
  if (n>9 || eopat[n] == NULL) return emptystr;
  return eopat[n];
}
