/*  YRX - yrxrt.c
**  (C) 2005 by Remo Dentato
**
** Permission to use, copy, modify and distribute this code and
** its documentation for any purpose is hereby granted without fee,
** provided that the above copyright notice appear in all copies
** and supporting documentation.
** 
** Copyright holder makes no representations about the suitability
** of this software for any purpose. It is provided "as is" without
** express or implied warranty.
** 
*/


#include <stdio.h>
#include <string.h>

#include "yrx_.h"


/* 
 *  This functions assumes the following:
 *     - Trailing spaces are uninmportant
 *       This implies that lines comprising only spaces are reported
 *       as empty lines
 * 
 *     - Lines ending with \ (possibly followed by spaces) continue
 *       on next line.
 *       The ending \ is replaced by a newline (0x0A).
 */ 
    
 int yygetline(YYSTREAM *y)
{
  int max=y->size;
  char *t,*b;
  int k=0;
  unsigned char c;

  if (y == NULL) return -1;
  
  y->buffer[0]='\02';
  y->cursor=y->buffer;
  b=y->cursor+1; 
  *b='\0';
  
  if (y->file) {
    do {
      t=fgets(b,max,y->file);
      if (t) {
        k++;
        while (*b) b++;
        while ((b>t) && isspace(b[-1])) b--;
        *b='\0';
        max -= (b-t);
        if (b[-1]=='\\') {
          y->linenum++;
          b[-1]='\n';
        }
        else
          t=NULL;
      }
    } while (t != NULL);
    
    if (k == 0) {
      *(y->cursor) = '\0';
      return -1;
    } 

  }
  else { /* string buffer */
    if (y->sptr == NULL || *(y->sptr) == '\0') return -1;
    c=1;
    while (c) {
      c=*(y->sptr);
      switch (c) {
        case '\r': if (*(y->sptr +1 ) == '\n') c = * ++y->sptr; break;
        case '\n': if (*(y->sptr +1 ) == '\r') y->sptr++; break;
      }
      switch (c) {
        case '\n': if (b[-1] == '\\') {
                     y->linenum++;
                     b[-1]='\n';
                   }
                   else c=0;
                   y->sptr++;
        case '\0': break;
        default  : *b++ = *y->sptr++;
                   break;
      }
    }
  }
  
  *b++='\03'; *b='\0';
  y->linenum++;
  
  #if 0
  printf("+%s+\n",y->cursor);
  fflush(stdout);
  #endif
  
  return 0;
} 

int yyswitch(unsigned char **r,YYSTREAM *y)
{
  int i=0;

  if (y == NULL) return -1;
  
  if ((y->cursor == NULL) || (*(y->cursor) == '\0'))
    if (yygetline(y) == -1) return -1;
  
  for (i=0;r[i] != NULL ;i++) {
    if (re_exec(y->cursor,r[i])) {
      y->cursor = re_end(0);
      return (i+1);
    }
  }
  return 0;
}

char *yysearch(YYSTREAM *y,unsigned char *r)
{
  if ((*r== YYSOL) && (r[1] == '\0')) {
    while (*(y->cursor) != YYSOL) y->cursor--;
    return (y->cursor);
  }
  
  if ((*r == '\0') || ((*r== YYEOL) && (r[1] == '\0'))) {
    while (*(y->cursor)) y->cursor++;
    return (y->cursor);
  }
  
  if (re_exec(y->cursor,r))
    return(y->cursor=re_start(0));
  
  return NULL;
}

char *yyseek(YYSTREAM *y,int n)
{
  int i=1;
  
  if (n<0) {
    n=-n; i=-1;
  }
  while (n > 0 && *(y->cursor)) {
    y->cursor += i;
    n--;
  }
  return y->cursor;
}

/*
** Buffer limit
*/

#define MINBUFSIZE     1024
#define MAXBUFSIZE  5242880 

static YYSTREAM *yynew(int size)
{
  YYSTREAM *y=NULL;
  
  if (size < MINBUFSIZE) size = MINBUFSIZE;
  if (size > MAXBUFSIZE) size = MAXBUFSIZE; 
  
  y=(YYSTREAM *)malloc(sizeof(YYSTREAM)+size);
  
  if (y) {
    y->linenum=0;
    y->size=size;
    y->file=NULL;
    y->cursor=y->buffer;
    y->buffer[0]='\0';
  }
  
  return y;
}

YYSTREAM *YYFILE(char *filename,int size)
{
  YYSTREAM *y=NULL;
  FILE *f=NULL;

  if (filename==NULL || *filename == '\0') {
    f=stdin;
    filename="(stdin)";
  }
  else
    f=fopen(filename,"r");
  
  if (f != NULL) {
    y=yynew(size+16+strlen(filename)+1);
  
    if (y) {
      y->size -= (strlen(filename)+1);
      y->file=f;
      y->sptr=y->buffer+ y->size;
      strcpy(y->sptr,filename);
    }
    else {
      if (f != stdin) fclose(f);
    }
  }
  
  return y;
}


/*
           +-----+
 buffer->  |     |  
           +-----+
           |     | 16 bytes gap
           +-----+
 sptr ->   |     |
           +-----+
*/

YYSTREAM *YYSTRING(unsigned char *string)
{
  YYSTREAM *y=NULL;

  if (string != NULL) {
    y=yynew(strlen(string)+1+16);
  
    if (y) {
      y->file=NULL;
      y->sptr=y->buffer+ 16;
      strcpy(y->sptr,string);
      /*fprintf(stderr,"<%s>\n",y->sptr);
      fflush(stderr);*/
    }
  }
  
  return y;
}

YYSTREAM *YYCLOSE(YYSTREAM *y)
{
 if (y) {
   if (y->file && (y->file != stdin)) fclose(y->file);
   y->size=0;
   free(y); 
 }
 return NULL;
}

int YYPOS(YYSTREAM *y)
{
  int p=-1;
  if (y) {
    if (y->file) p=ftell(y->file);
  }
  return p;
}

const char *YYFILENAME(YYSTREAM *y)
{
  return (((y) &&  (y)->file)?(const char *)((y)->sptr):"(string)");
}

unsigned int YYLINENUMBER(YYSTREAM *y)
{
  return (y?y->linenum:0);
}

