/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef AUDIOENGINE_H
#define AUDIOENGINE_H

#include <jack/jack.h>

#include "audioprocessor.h"
#include "clock.h"
#include "jacktransport.h"
#include "processclosure.h"
#include "processorthread.h"
#include "regextype.h"

namespace bipscript {

namespace jack {
class Timebase;
class Clock;
enum PortFlag { IsInput, IsOutput };
} // namespace jack

class XrunClosure : public ProcessClosure {
  protected:
    void addParameters() { addTimeParameter(); }

  public:
    XrunClosure(ScriptFunction function) : ProcessClosure(function, 0) {}
};

class AudioEngine : public ProcessorEngine {
    // Jack client + info
    jack_client_t *client;
    jack_nframes_t msampleRate;

    // transport and restart
    uint16_t syncPeriods;
    std::atomic<bool> mrestarting; // script requests restart
    std::atomic<ScriptFunction *> onXrunHandler;

    // clocks
    jack::Timebase *timebase;
    jack::Clock *externalClock;
    QueueList<time::Clock> activeClocks;
    std::atomic<samples_t> mlatency;

    // processors
    std::atomic<bool> mactiveProcessors;
    QueueList<AudioProcessor> activeProcessors;
    ThreadQueue<AudioProcessor> deletedProcessors;

    // private methods
    bool flushComplete();
    void removeDeletedProcessors();
    bool masterSync(jack_transport_state_t state, jack_position_t *pos);

    // singleton
    AudioEngine();
    AudioEngine(AudioEngine const &);
    void operator=(AudioEngine const &);

  public:
    static bool loaded;
    static AudioEngine &instance() {
        static AudioEngine instance;
        return instance;
    }
    jack_nframes_t sampleRate() { return msampleRate; }
    void setBufferSize(jack_nframes_t size);
    void addClock(time::Clock *clock) { activeClocks.add(clock); }
    void addProcessor(AudioProcessor *obj) { activeProcessors.add(obj); }
    void removeProcessor(AudioProcessor *obj) { deletedProcessors.add(obj); }

    // public methods
    int process(jack_nframes_t nframes);
    void onshutdown();
    void shutdown();
    HSQOBJECT getPorts(const char *name, const char *type, unsigned long flags);

    // for callbacks
    int sync(jack_transport_state_t state, jack_position_t *pos);
    void latency(jack_latency_callback_mode_t mode);
    samples_t latency() { return mlatency.load(); }

    // ProcessorThread interface
    void restart() override { mrestarting.store(true); }
    bool restarting() override { return mrestarting.load(); }
    bool active() override { return mactiveProcessors.load(); }

    // system ports
    jack_port_t *registerMidiInputPort(const char *name) {
        return jack_port_register(client, name, JACK_DEFAULT_MIDI_TYPE, JackPortIsInput, 0);
    }
    jack_port_t *registerMidiOutputPort(const char *name) {
        return jack_port_register(client, name, JACK_DEFAULT_MIDI_TYPE, JackPortIsOutput, 0);
    }
    jack_port_t *registerAudioInputPort(const char *name) {
        return jack_port_register(client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
    }
    jack_port_t *registerAudioOutputPort(const char *name) {
        return jack_port_register(client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    }
    void connectPort(jack_port_t *port, const char *connection) {
        jack_connect(client, jack_port_name(port), connection);
    }
    void connectPort(const char *connection, jack_port_t *port) {
        jack_connect(client, connection, jack_port_name(port));
    }
    void connectPort(jack_port_t *port, Regex &regex, const char *type);
    void connectPort(Regex &regex, jack_port_t *port, const char *type);
    void disconnectPort(jack_port_t *port, const char *connection) {
        jack_disconnect(client, jack_port_name(port), connection);
    }
    void disconnectPort(const char *connection, jack_port_t *port) {
        jack_disconnect(client, connection, jack_port_name(port));
    }
    void disconnectOutputPorts(jack_port_t *port);
    void disconnectInputPorts(jack_port_t *port);
    void unregisterPort(jack_port_t *port) { jack_port_unregister(client, port); }
    // jack transport
    jack::Transport *newTransport() { return new jack::Transport(client); }
    jack::Timebase *getTimebase(float bpm, time::Signature &signature);
    jack::Clock *externalTimeClock();
    // xruns
    void handleXrun();
    void onXrun(ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            throw std::logic_error("onXrun handler should take one argument");
        }
        onXrunHandler.store(new ScriptFunction(handler));
    }
};

} // namespace bipscript

#endif // AUDIOENGINE_H
