/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef AUDIOOUTPUT_H
#define AUDIOOUTPUT_H

#include "audiosink.h"
#include "processorcache.h"

namespace bipscript {
namespace audio {

class AudioInputPort : public Source {
    jack_port_t *port;

  public:
    AudioInputPort(jack_port_t *jackPort, const char *name);
    void connect(const char *name) { AudioEngine::instance().connectPort(name, port); }
    void connect(Regex &regex) {
        AudioEngine::instance().connectPort(regex, port, JACK_DEFAULT_AUDIO_TYPE);
    }
    void disconnect() { AudioEngine::instance().disconnectOutputPorts(port); }
    void disconnect(const char *name) { AudioEngine::instance().disconnectPort(name, port); }
    // Source interface
    bool connectsTo(AbstractSink *) override { return false; }
    bool doProcess(samples_t time) override {
        audioOutput(0)->setBuffer((float *)jack_port_get_buffer(port, Clock::period()));
        grabClips(Clock::period());
        reportLevels(this, time);
        return Source::activeEvents();
    }
    void restart() override {}
};

class AudioInputPortFactory : public ProcessorCache<AudioInputPort> {
  public:
    static AudioInputPortFactory &instance() {
        static AudioInputPortFactory instance;
        return instance;
    }
    AudioInputPort *newAudioInputPort(const char *name, const char *connectTo);
    AudioInputPort *newAudioInputPort(const char *name) { return newAudioInputPort(name, 0); }
};

class InputPortSet : public Source {
    jack_port_t **mport;

  public:
    InputPortSet(const char *name, uint32_t channels);
    ~InputPortSet();
    void autoconnect();                        // connect to first N system capture ports
    void connect(int index, const char *name); // connect to system port
    void connect(int index, Regex &pattern);   // connect to system port
    void unregister();
    // Source interface
    bool connectsTo(AbstractSink *) override { return false; }
    bool doProcess(samples_t time) override;
    void restart() override {}
};

typedef InputPortSet StereoInput;
typedef InputPortSet MultiInput;

class InputPortSetFactory : public ProcessorCache<InputPortSet> {
  protected:
    void removeObject(InputPortSet *set);

  public:
    static InputPortSetFactory &instance() {
        static InputPortSetFactory instance;
        return instance;
    }
    InputPortSet *newInputPortSet(const char *name, uint32_t channels, bool autoconnect = false);
    InputPortSet *newStereoInputPort(const char *name, bool autoconnect = false) {
        return newInputPortSet(name, 2, autoconnect);
    }
};

/**
 * represents a set of system outputs
 *
 */
class OutputPortSet : public Sink, public AudioProcessor {
    jack_port_t **mport;
    // status reporting
    samples_t lastreport;
    Meter *meter;

  public:
    OutputPortSet(const char *name, uint32_t channels);
    ~OutputPortSet();
    void connect(const char *name) { AudioEngine::instance().connectPort(mport[0], name); }
    void connect(Regex &regex) {
        AudioEngine::instance().connectPort(mport[0], regex, JACK_DEFAULT_AUDIO_TYPE);
    }
    void connect(const char **names);
    void disconnect();
    void disconnect(const char *name) { AudioEngine::instance().disconnectPort(mport[0], name); }
    void autoconnect();
    void unregister();
    // process interface
    bool doProcess(samples_t time) override;
    void restart() override { Sink::processRestart(); }
    void restartScript() override { Sink::scriptRestart(); }
};

typedef OutputPortSet Output;
typedef OutputPortSet StereoOutput;
typedef OutputPortSet MultiOutput;

class OutputPortSetFactory : public ProcessorCache<OutputPortSet> {
  protected:
    void removeObject(OutputPortSet *);

  public:
    static OutputPortSetFactory &instance() {
        static OutputPortSetFactory instance;
        return instance;
    }
    MultiOutput *newOutputPortSet(const char *name, uint32_t channels, bool autoconnect = false);
    Output *newAudioOutputPort(const char *name, bool autoconnect = false) {
        return newOutputPortSet(name, 1, autoconnect);
    }
    Output *newAudioOutputPort(const char *name, const char *connect);
    StereoOutput *newStereoOutputPort(const char *name, bool autoconnect) {
        return newOutputPortSet(name, 2, autoconnect);
    }
    StereoOutput *newStereoOutputPort(const char *name) { return newOutputPortSet(name, 2); }
};

} // namespace audio
} // namespace bipscript

#endif // AUDIOOUTPUT_H
