/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>

#include "beattracker.h"

namespace bipscript {

void BaseBeatTracker::tick(samples_t time, bool isbeat, samples_t period) {
    // apply incoming commands
    BasicClock::clockChanges();

    if (mtimer.running()) {
        // beat during this period - might need to adjust
        if (isbeat) {
            current = following;
            // calculate current measure location for this beat
            current.bpm(btrack.getCurrentTempoEstimate());
            // assume beat location is start of this period
            // TODO: find more precise via onset detection?
            double beat = current.measure();
            // guess the closest beat - which one should it be?
            double closest = closestBeat(beat);
            // ideal period end measure is ideal beat location
            // plus measure offset at current estimated bpm
            double end = closest + current.moffset(period);
            if (current.measure() >= end) {
                // start of this period already past ideal end
                current.bpm(0.1); // slow way down to catch up
            } else {
                // what BPM do we need to get to ideal end
                time::Signature signature = current.signature();
                double bpm = current.sampleRate() * 15 * signature.numerator() *
                             signature.denominator() * (end - current.measure()) / period;
                current.bpm(bpm);
            }
            current.size(period);
            // compute following window
            following.frame(current.frame() + period);
            following.measure(current.measure() + current.moffset(period));
            following.bpm(btrack.getCurrentTempoEstimate());
            following.reposition(false);

        } else {
            BasicClock::update(time);
        }
    }
}

double BaseBeatTracker::closestBeat(double location) {
    double ret;
    double closest = -1;
    double start = std::trunc(location);
    uint16_t num = current.signature().numerator();
    for (float i = 0; i <= num; i++) {
        double beat = start + i / num;
        double delta = std::abs(location - beat);
        if (closest < 0 || delta < closest) {
            closest = delta;
            ret = beat;
        }
    }
    return ret;
}

namespace audio {

void BeatTracker::update(samples_t time) {
    // get audio from audio connection
    minput.process(time);
    Connection &connection = minput.outputbus()[0];
    float *audio = connection.audio();

    // add new audio to buffer
    bool beat = false;
    samples_t period = Clock::period();
    for (samples_t i = 0; i < period; i++) {
        btbuffer[index++] = (double)audio[i];
        if (index == BT_HOP_SIZE) {
            // process buffer if full
            btrack.processAudioFrame(btbuffer);
            if (btrack.beatDueInCurrentFrame()) {
                beat = true;
            }
            index = 0;
        }
    }

    BaseBeatTracker::tick(time, beat, period);
}

} // end namespace audio
namespace midi {

void BeatTracker::update(samples_t time) {
    // process MIDI input
    minput.process(time);
    Connection *connection = &minput.midioutbus()[0];

    // loop over frames
    bool beat = false;
    samples_t period = Clock::period();
    RawEvent *nextEvent = connection->firstEvent();
    for (samples_t i = 0; i < period; i++) {

        // add events at this frame to current onset
        while (nextEvent && nextEvent->start() == i) {
            if (nextEvent->message().command() == NoteOn::COMMAND) {
                NoteOn noteOn(nextEvent->message().bytes());
                onsetSample += noteOn.velocity() * noteWeight[noteOn.pitch()];
            }
            nextEvent = connection->nextEvent();
        }

        // full buffer, run beat tracker
        if (index == BT_HOP_SIZE) {
            // process onset sample
            btrack.processOnsetDetectionFunctionSample(onsetSample);
            if (btrack.beatDueInCurrentFrame()) {
                beat = true;
            }
            index = 0;
            onsetSample = 0;
        }
        index++;
    }
    BaseBeatTracker::tick(time, beat, period);
}

} // namespace midi
} // namespace bipscript
