/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIXER_H
#define MIXER_H

#include "audiosink.h"
#include "processorcache.h"

using std::map;

namespace bipscript {

namespace audio {

class MixerGain : public RefCountable<MixerGain> {
    const uint32_t minput;
    const uint32_t moutput;
    const float mamount;

  public:
    MixerGain(uint32_t input, uint32_t output, float amount)
        : minput(input), moutput(output), mamount(amount) {}
    bool applies(uint32_t inputs, uint32_t outputs) { return minput < inputs && moutput < outputs; }
    uint32_t input() { return minput; }
    uint32_t output() { return moutput; }
    float amount() { return mamount; }
};

class Mixer : public Sink, public Source {
    std::atomic<float> **mgain;
    TimelineSet<MixerGain> gainEventBuffer; // gain change events

  public:
    Mixer(uint32_t inputs, uint32_t outputs, HSQOBJECT *gains);
    ~Mixer();
    void add(Connection &connection);
    void add(Connection &connection, float gain);
    void add(Connection &connection, HSQOBJECT &gains);
    void add(Source &source);
    void add(Source &source, float gain);
    void add(Source &source, HSQOBJECT &gains);
    void schedule(MixerGain &gain, float measure, Clock &clock = Clock::defaultClock());
    void gain(uint32_t input, uint32_t output, float gain);
    void gain(uint32_t input, HSQOBJECT &gains);
    // Source interface
    bool connectsTo(AbstractSink *source) override;
    bool doProcess(samples_t time) override;
    void restart() override;
    void restartScript() override;

  private:
    void add(Connection &connection, uint32_t input);
    void add(Source &source, uint32_t start);
    uint32_t connectedInputs();
    void validateInputChannel(uint32_t input);
    void validateOutputChannel(uint32_t output);
    void reportGain(uint32_t input);
};

class MixerFactory : public ProcessorCache<Mixer> {
    std::map<int, int> instanceCount;
    void restartCache() { instanceCount.clear(); }

  public:
    static MixerFactory &instance() {
        static MixerFactory instance;
        return instance;
    }
    Mixer *newMixer(uint16_t inputs, uint16_t outputs, HSQOBJECT *gains);
    Mixer *newMixer(uint16_t inputs, uint16_t outputs) {
        return newMixer(inputs, outputs, nullptr);
    }
    Mixer *newMixer(uint16_t inputs, uint16_t outputs, HSQOBJECT &gains) {
        return newMixer(inputs, outputs, &gains);
    }
};

} // namespace audio
} // namespace bipscript

#endif // MIXER_H
