/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BINDUTIL_H
#define BINDUTIL_H

#include "squirrel.h"

namespace bipscript {

// macros for args
#define BIND_NARGS (sq_gettop(vm) - 1)

#define BIND_GET_NARGS() SQInteger numargs = sq_gettop(vm);

#define BIND_NARGS_CHECK_EXACT(Num)                                                                \
    if (numargs != Num + 1) {                                                                      \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME                              \
                                                 " method expected exactly " #Num " parameters");  \
    }

#define BIND_NARGS_CHECK_MIN(Min)                                                                  \
    if (numargs < Min + 1) {                                                                       \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME                              \
                                                 " method expected at least " #Min " parameters"); \
    }

#define BIND_NARGS_CHECK_MAX(Max)                                                                  \
    if (numargs > Max + 1) {                                                                       \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME                              \
                                                 " method expected at most " #Max " parameters");  \
    }

// macros for get
#define BIND_GET_THIS(Class, Tag)                                                                  \
    SQUserPointer userPtr = 0;                                                                     \
    if (SQ_FAILED(sq_getinstanceup(vm, 1, &userPtr, &Tag, false))) {                               \
        return sq_throwerror(vm, BIND_CLASS_NAME " method needs an instance of " BIND_CLASS_NAME); \
    }                                                                                              \
    Class *obj = static_cast<Class *>(userPtr);

#define BIND_GET_THIS_INSTANCE(Class)                                                              \
    SQUserPointer userPtr = 0;                                                                     \
    if (SQ_FAILED(sq_getinstanceup(vm, 1, &userPtr, 0, false))) {                                  \
        return sq_throwerror(vm,                                                                   \
                             BIND_METHOD_NAME " method needs an instance of " BIND_CLASS_NAME);    \
    }                                                                                              \
    Class *obj = static_cast<Class *>(userPtr);                                                    \
    if (!obj) {                                                                                    \
        return sq_throwerror(vm, BIND_METHOD_NAME " method called before " BIND_CLASS_NAME         \
                                                  " constructor");                                 \
    }

#define BIND_GET_THIS_PROPERTY(Class)                                                              \
    SQUserPointer userPtr = 0;                                                                     \
    if (SQ_FAILED(sq_getinstanceup(vm, 1, &userPtr, 0, false))) {                                  \
        return sq_throwerror(vm, "property needs an instance of " BIND_CLASS_NAME);                \
    }                                                                                              \
    Class *obj = static_cast<Class *>(userPtr);                                                    \
    if (!obj) {                                                                                    \
        sq_pushnull(vm);                                                                           \
        return sq_throwobject(vm);                                                                 \
    }

#define BIND_GET_CHECK_ANY(Name, Pos) HSQUIRRELVM &Name = vm;

#define BIND_GET_ANY(Name, Pos)                                                                    \
    HSQOBJECT Name;                                                                                \
    sq_getstackobj(vm, Pos + 1, &Name);

#define BIND_CHECK_ANY(Pos) true

#define BIND_GET_BOOL(Name, Pos) SQBool Name = arg##Pos._unVal.nInteger;

#define BIND_GET_CHECK_BOOL(Name, Pos)                                                             \
    SQBool Name;                                                                                   \
    if (SQ_FAILED(sq_getbool(vm, Pos + 1, &Name))) {                                               \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type boolean");         \
    }

#define BIND_GET_FLOAT(Name, Pos)                                                                  \
    SQFloat Name = arg##Pos._type == OT_FLOAT ? arg##Pos._unVal.fFloat : arg##Pos._unVal.nInteger;

#define BIND_CHECK_FLOAT(Pos) arg##Pos._type == OT_FLOAT || arg##Pos._type == OT_INTEGER

#define BIND_GET_CHECK_FLOAT(Name, Pos)                                                            \
    SQFloat Name;                                                                                  \
    if (SQ_FAILED(sq_getfloat(vm, Pos + 1, &Name))) {                                              \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type float");           \
    }

#define BIND_GET_INSTANCE(Name, Pos)                                                               \
    SQInstance *Name;                                                                              \
    HSQOBJECT obj;                                                                                 \
    sq_getstackobj(vm, Pos + 1, &obj);                                                             \
    Name = obj._unVal.pInstance;

#define BIND_GET_INTEGER(Name, Pos) SQInteger Name = arg##Pos._unVal.nInteger;

#define BIND_GET_CHECK_INTEGER(Name, Pos)                                                          \
    SQInteger Name;                                                                                \
    if (SQ_FAILED(sq_getinteger(vm, Pos + 1, &Name))) {                                            \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type integer");         \
    }

#define BIND_GET_NULL(Name, Pos) const SQChar *Name = 0;

#define BIND_GET_CHECK_NULL(Name, Pos)                                                             \
    const SQChar *Name = 0;                                                                        \
    if (sq_gettype(vm, Pos + 1) != OT_NULL) {                                                      \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type null");            \
    }

#define BIND_GET_STRING(Name, Pos) const SQChar *Name = sq_objtostring(&arg##Pos);

#define BIND_GET_CHECK_STRING(Name, Pos)                                                           \
    const SQChar *Name;                                                                            \
    if (SQ_FAILED(sq_getstring(vm, Pos + 1, &Name))) {                                             \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type string");          \
    }

#define BIND_GET_CHECK_NULLABLE_STRING(Name, Pos)                                                  \
    const SQChar *Name;                                                                            \
    if (SQ_FAILED(sq_getstring(vm, Pos + 1, &Name))) {                                             \
        if (sq_gettype(vm, Pos + 1) == OT_NULL) {                                                  \
            Name = nullptr;                                                                        \
        } else {                                                                                   \
            return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos        \
                                                     " \"" #Name "\" is not of type string");      \
        }                                                                                          \
    }

#define BIND_GET_ARRAY(Name, Pos)                                                                  \
    HSQOBJECT Name;                                                                                \
    sq_getstackobj(vm, Pos + 1, &Name);

#define BIND_GET_CHECK_ARRAY(Name, Pos)                                                            \
    HSQOBJECT Name;                                                                                \
    sq_getstackobj(vm, Pos + 1, &Name);                                                            \
    if (sq_gettype(vm, Pos + 1) != OT_ARRAY) {                                                     \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type array");           \
    }

#define BIND_GET_CHECK_CLASS(Name, Pos)                                                            \
    HSQOBJECT Name;                                                                                \
    sq_getstackobj(vm, Pos + 1, &Name);                                                            \
    if (sq_gettype(vm, Pos + 1) != OT_CLASS) {                                                     \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type class");           \
    }

#define BIND_GET_CHECK_FUNCTION(Name, Pos)                                                         \
    HSQOBJECT NameObj;                                                                             \
    sq_getstackobj(vm, Pos + 1, &NameObj);                                                         \
    if (sq_gettype(vm, Pos + 1) != OT_CLOSURE) {                                                   \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type function");        \
    }                                                                                              \
    ScriptFunction Name(vm, NameObj, Pos + 1);

#define BIND_GET_TABLE(Name, Pos) ScriptHashIterator Name(vm, arg##Pos);

#define BIND_GET_CHECK_TABLE(Name, Pos)                                                            \
    HSQOBJECT NameObj;                                                                             \
    sq_getstackobj(vm, Pos + 1, &NameObj);                                                         \
    if (sq_gettype(vm, Pos + 1) != OT_TABLE) {                                                     \
        return sq_throwerror(vm, BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos            \
                                                 " \"" #Name "\" is not of type table");           \
    }                                                                                              \
    ScriptHashIterator Name(vm, NameObj);

#define BIND_GET_CHECK_INTERFACE(Name, Pos, Pack, Class, Ns, Cname)                                \
    Ns::Cname *Name = Ns::binding::get##Class(vm, Pos + 1);                                        \
    if (Name == 0) {                                                                               \
        return sq_throwerror(vm,                                                                   \
                             BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos " \"" #Name    \
                                             "\" is not of type " #Pack "." #Class);               \
    }

#define BIND_GET_OBJECT(Name, Pos, Pack, Class, Ns, Cname)                                         \
    Ns::Cname *Name = BindUtil::getObject<Ns::Cname>(vm, Pos + 1, Ns::binding::Class##Object);

#define BIND_GET_CHECK_OBJECT(Name, Pos, Pack, Class, Ns, Cname)                                   \
    Ns::Cname *Name = BindUtil::getObject<Ns::Cname>(vm, Pos + 1, Ns::binding::Class##Object);     \
    if (Name == 0) {                                                                               \
        return sq_throwerror(vm,                                                                   \
                             BIND_CLASS_NAME " " BIND_METHOD_NAME " argument " #Pos " \"" #Name    \
                                             "\" is not of type " #Pack "." #Class);               \
    }

// macros for put
#define BIND_PUSH_BOOL(Name) sq_pushbool(vm, Name);

#define BIND_PUSH_FLOAT(Name) sq_pushfloat(vm, Name);

#define BIND_PUSH_INTEGER(Name) sq_pushinteger(vm, Name);

#define BIND_PUSH_CSTRING(Name) sq_pushstring(vm, Name, -1);

#define BIND_PUSH_STRING(Name) sq_pushstring(vm, Name.c_str(), Name.size());

#define BIND_PUSH_OBJECT(Name, Class, Ns, Cname, Release)                                          \
    sq_pushobject(vm, Ns::binding::Class##Object);                                                 \
    sq_createinstance(vm, -1);                                                                     \
    sq_remove(vm, -2);                                                                             \
    sq_setinstanceup(vm, -1, Name);                                                                \
    sq_setreleasehook(vm, -1, BindUtil::Release<Ns::Cname>);

#define BIND_WRAP_TRY(Exp)                                                                         \
    try {                                                                                          \
        Exp;                                                                                       \
    } catch (std::exception const &e) {                                                            \
        return sq_throwerror(vm, e.what());                                                        \
    }

class BindUtil {
    HSQUIRRELVM vm;

  public:
    BindUtil(HSQUIRRELVM vm) : vm(vm) {}
    void startTable(const char *name, SQInteger namelen) {
        sq_pushstring(vm, name, namelen);
        sq_newtable(vm);
    }
    void endTable() { endObject(); }
    void startClass(const char *name, SQInteger namelen, HSQOBJECT &object) {
        sq_pushstring(vm, _SC(name), namelen);
        sq_newclass(vm, false);
        sq_getstackobj(vm, -1, &object);
        sq_settypetag(vm, -1, &object);
    }
    void endClass() {
        // TODO: check type
        endObject();
    }
    void endObject() {
        sq_newslot(vm, -3, false); // push class to underlying table
    }
    void addMethod(const char *name, SQInteger namelen, SQFUNCTION func) {
        sq_pushstring(vm, _SC(name), namelen);
        sq_newclosure(vm, func, 0);
        sq_newslot(vm, -3, false);
    }
    // templates
    template <class T> static T *getObject(HSQUIRRELVM vm, int index, HSQOBJECT &Object) {
        SQUserPointer objPtr;
        if (!SQ_FAILED(sq_getinstanceup(vm, index, (SQUserPointer *)&objPtr, &Object, false))) {
            return static_cast<T *>(objPtr);
        }
        return nullptr;
    }
    template <class T> static SQInteger Dtor(SQUserPointer p, SQInteger) {
        delete static_cast<T *>(p);
        return 0;
    }
    template <class T> static SQInteger Clone(HSQUIRRELVM vm) {
        // get instance ptr of original
        SQUserPointer userPtr;
        sq_getinstanceup(vm, 2, &userPtr, 0, false);
        // set instance ptr to a copy
        sq_setinstanceup(vm, 1, new T(*(T *)userPtr));
        sq_setreleasehook(vm, 1, &Dtor<T>);
        return 0;
    }
    template <class T> static SQInteger Decrement(SQUserPointer p, SQInteger) {
        T *obj = static_cast<T *>(p);
        if (obj->delref() == 0) {
            delete obj;
        }
        return 0;
    }
    template <class T> static SQInteger Delete(SQUserPointer p, SQInteger) {
        delete static_cast<T *>(p);
        return 0;
    }
    template <class T> static SQInteger Push(HSQUIRRELVM vm, T *obj, HSQOBJECT &Object) {
        sq_pushobject(vm, Object);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, obj);
        sq_setreleasehook(vm, -1, BindUtil::Delete<T>);
        return 0;
    }
    template <class T> static SQInteger PushCached(HSQUIRRELVM vm, T *obj, HSQOBJECT &Object) {
        sq_pushobject(vm, Object);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, obj);
        return 0;
    }
    template <class T> static SQInteger PushCopy(HSQUIRRELVM vm, T &obj, HSQOBJECT &Object) {
        sq_pushobject(vm, Object);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, new T(obj));
        sq_setreleasehook(vm, -1, BindUtil::Delete<T>);
        return 0;
    }
    template <class T> static SQInteger PushRef(HSQUIRRELVM vm, T *obj, HSQOBJECT &Object) {
        sq_pushobject(vm, Object);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, obj->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<T>);
        return 0;
    }
};

} // namespace bipscript

#endif // BINDUTIL_H
