/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BUFFERTYPE_H
#define BUFFERTYPE_H

#include "bindutil.h"
#include "reader.h"
#include "refcountable.h"
#include "writer.h"
#include <boost/crc.hpp>
#include <stdexcept>
#include <string.h>

namespace bipscript {

namespace binding {
extern HSQOBJECT bufferObject;
}

class BufferReader;
class BufferWriter;

class Buffer : public RefCountable<Buffer> {
    unsigned char *mbytes;
    size_t msize;

  public:
    static void bindType(HSQUIRRELVM vm) {
        BindUtil binding(vm);
        binding.startClass("buffer", 6, binding::bufferObject);
        binding.addMethod("constructor", 11, Ctor);
        binding.addMethod("_cloned", 7, BindUtil::Clone<Buffer>); // TODO: copy ctor
        binding.addMethod("_tostring", 9, ToString);
        binding.addMethod("_get", 4, Getters);
        binding.addMethod("_set", 4, Setters);
        binding.addMethod("reader", 6, GetReader);
        binding.addMethod("size", 4, Size);
        binding.addMethod("slice", 5, Slice);
        binding.addMethod("writer", 6, GetWriter);
        binding.endClass();
    }
    //    static Buffer *getBuffer(HSQUIRRELVM vm, SQInteger index) {
    //        return BindUtil::getObject<Buffer>(vm, index, binding::bufferObject);
    //    }
    Buffer(size_t size) : msize(size) {
        mbytes = new unsigned char[size];
        memset(mbytes, 0, size);
    }
    ~Buffer() { delete[] mbytes; }
    size_t size() { return msize; }
    unsigned char *bytes() { return mbytes; }
    Buffer *slice(size_t start, size_t end) {
        if (end > msize) {
            throw std::logic_error("slice end index greater than buffer size");
        }
        size_t size = end - start;
        Buffer *ret = new Buffer(size);
        size_t index = 0;
        for (size_t i = start; i < end; i++) {
            ret->mbytes[index++] = mbytes[i];
        }
        return ret;
    }
    Writer *writer();
    Reader *reader();
    int crc() {
        boost::crc_basic<16> crc(0x8005u, 0u, 0u, true, true);
        crc.process_bytes(mbytes, msize);
        return crc.checksum();
    }
    unsigned char &operator[](size_t index) {
        if (index >= msize) {
            throw std::logic_error("buffer index out of range");
        }
        return mbytes[index];
    }

  private:
    // binding methods
#define BIND_CLASS_NAME "buffer"
#define BIND_METHOD_NAME "constructor"
    static SQInteger Ctor(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        BIND_GET_CHECK_INTEGER(size, 1);
        try {
            Buffer *obj = new Buffer(size);
            // return pointer to new object
            sq_setinstanceup(vm, 1, (SQUserPointer *)obj->addref());
            sq_setreleasehook(vm, 1, BindUtil::Decrement<Buffer>);
            return 1;
        } catch (std::exception const &e) {
            return sq_throwerror(vm, e.what());
        }
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "_tostring"
    static SQInteger ToString(HSQUIRRELVM vm) {
        BIND_GET_THIS(Buffer, binding::bufferObject)
        char message[128];
        sprintf(message, "(buffer %ld bytes)", obj->msize);
        sq_pushstring(vm, message, -1);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "reader"
    static SQInteger GetReader(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Buffer, binding::bufferObject)
        Reader *ret = obj->reader();
        // push return value
        sq_pushobject(vm, binding::readerObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<bipscript::Reader>);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "size"
    static SQInteger Size(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Buffer, binding::bufferObject)
        sq_pushinteger(vm, obj->msize);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "slice"
    static SQInteger Slice(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(2)
        BIND_GET_THIS(Buffer, binding::bufferObject)
        BIND_GET_CHECK_INTEGER(start, 1)
        BIND_GET_CHECK_INTEGER(end, 2)
        if (end < 0 || start < 0) {
            throw std::logic_error("slice parameters must be greater than zero");
        }
        if (end <= start) {
            throw std::logic_error("slice end index must be greater than start");
        }
        Buffer *ret = obj->slice(start, end);
        // push return value
        sq_pushobject(vm, binding::bufferObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<Buffer>);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "writer"
    static SQInteger GetWriter(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        BIND_GET_THIS(Buffer, binding::bufferObject)
        Writer *ret = obj->writer();
        // push return value
        sq_pushobject(vm, binding::writerObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<bipscript::Writer>);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "getter"
    static SQInteger Getters(HSQUIRRELVM vm) {
        SQObjectType argtype = sq_gettype(vm, 2);
        if (argtype == OT_INTEGER) {
            SQInteger index;
            sq_getinteger(vm, 2, &index);
            BIND_GET_THIS_PROPERTY(Buffer)
            try {
                unsigned char ret = (*obj)[index];
                sq_pushinteger(vm, ret);
            } catch (std::exception const &e) {
                return sq_throwerror(vm, e.what());
            }
            return 1;
        }
        sq_pushnull(vm);
        return sq_throwobject(vm);
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "setter"
    static SQInteger Setters(HSQUIRRELVM vm) {
        SQObjectType argtype = sq_gettype(vm, 2);
        if (argtype == OT_INTEGER) {
            SQInteger index, value;
            BIND_GET_THIS_PROPERTY(Buffer)
            sq_getinteger(vm, 2, &index);
            sq_getinteger(vm, 3, &value);
            try {
                (*obj)[index] = value;
            } catch (std::exception const &e) {
                return sq_throwerror(vm, e.what());
            }
            return 1;
        }
        sq_pushnull(vm);
        return sq_throwobject(vm);
    }
#undef BIND_METHOD_NAME
#undef BIND_CLASS_NAME
};

class BufferReader : public Reader {
    Buffer *mbuf;
    size_t mpos;

  public:
    BufferReader(Buffer *buf) : mbuf(buf->addref()), mpos(0) {}
    ~BufferReader() {
        if (!mbuf->delref()) {
            delete mbuf;
        }
    }
    Buffer *buffer() const { return mbuf; }
    bool eos() override { return mpos == mbuf->size(); }
    void line(std::string &str) override {
        unsigned char *ptr = &(mbuf->bytes()[mpos]);
        while (mpos < mbuf->size() && *ptr != '\n') {
            str += *ptr;
            mpos++;
            ptr++;
        }
        while (mpos < mbuf->size() && *ptr == '\n') {
            mpos++;
            ptr++;
        }
    }
    bool readBool(bool &value) override {
        size_t end = mpos + sizeof(bool);
        if (end > mbuf->size()) {
            return false;
        }
        value = *((bool *)ptr());
        mpos = end;
        return true;
    }
    bool readByte(unsigned char &value) override {
        if (mpos >= mbuf->size()) {
            return false;
        }
        value = *((unsigned char *)ptr());
        mpos++;
        return true;
    }
    bool readFloat(float &value) override {
        size_t end = mpos + sizeof(float);
        if (end > mbuf->size()) {
            return false;
        }
        value = *((float *)ptr());
        mpos = end;
        return true;
    }
    bool readInt(long &value, size_t size) override {
        if (size > sizeof(long)) {
            std::string mesg("integer size parameter may not be greater than ");
            mesg.append(std::to_string(sizeof(long)));
            throw std::logic_error(mesg);
        }
        size_t end = mpos + size;
        if (end > mbuf->size()) {
            return false;
        }
        memcpy(&value, ptr(), size);
        mpos = end;
        return true;
    }
    bool readString(std::string &value) override {
        if (mpos >= mbuf->size()) {
            return false;
        }
        char *ptr = (char *)&(mbuf->bytes()[mpos]);
        while (mpos < mbuf->size() && *ptr != '\0') {
            value += *ptr;
            mpos++;
            ptr++;
        }
        if (mpos < mbuf->size()) { // skip over '\0'
            mpos++;
        }
        return true;
    }
    bool readBuffer(Buffer **value, size_t size) override {
        if (mpos >= mbuf->size()) {
            return false;
        }
        *value = (new Buffer(size))->addref();
        size_t toread = std::min(size, mbuf->size() - mpos);
        memcpy((*value)->bytes(), ptr(), toread);
        mpos += toread;
        return true;
    }
    size_t tell() override { return mpos; }
    void seek(size_t pos) { mpos = pos; }

  private:
    unsigned char *ptr() { return &(mbuf->bytes()[mpos]); }
};

class BufferWriter : public Writer {
    Buffer *mbuf;
    size_t mpos;

  public:
    BufferWriter(Buffer *buf) : mbuf(buf->addref()), mpos(0) {}
    ~BufferWriter() {
        if (!mbuf->delref()) {
            delete mbuf;
        }
    }
    void print(const char *str) override {
        size_t len = strlen(str);
        size_t end = mpos + len;
        checkPosition(end);
        char *ptr = (char *)(&(mbuf->bytes()[mpos]));
        strncpy(ptr, str, len);
        mpos = end;
    }
    void println(const char *str) override {
        size_t len = strlen(str);
        size_t end = mpos + len + 1;
        checkPosition(end);
        char *ptr = (char *)(&(mbuf->bytes()[mpos]));
        strncpy(ptr, str, len);
        ptr[len] = '\0';
        mpos = end;
    }
    void write(bool b) override {
        size_t end = mpos + sizeof(bool);
        checkPosition(end);
        unsigned char *ptr = &(mbuf->bytes()[mpos]);
        *((bool *)ptr) = b;
        mpos = end;
    }
    void write(const char *str) override {
        size_t end = mpos + strlen(str) + 1;
        checkPosition(end);
        char *ptr = (char *)(&(mbuf->bytes()[mpos]));
        sprintf(ptr, "%s", str);
        mpos = end;
    }
    void write(float f) override {
        size_t end = mpos + sizeof(float);
        checkPosition(end);
        unsigned char *ptr = &(mbuf->bytes()[mpos]);
        *((float *)ptr) = f;
        mpos = end;
    }
    void write(long l) override {
        size_t end = mpos + sizeof(long);
        checkPosition(end);
        unsigned char *ptr = &(mbuf->bytes()[mpos]);
        *((long *)ptr) = l;
        mpos = end;
    }
    void write(Buffer *buffer) override {
        size_t end = mpos + buffer->size();
        checkPosition(end);
        unsigned char *ptr = &(mbuf->bytes()[mpos]);
        memcpy(ptr, buffer->bytes(), buffer->size());
        mpos = end;
    }

  private:
    void checkPosition(size_t pos) {
        if (pos > mbuf->size()) {
            throw std::logic_error("writer attempt to write beyond end of buffer");
        }
    }
};

inline Writer *Buffer::writer() { return new BufferWriter(this); }

inline Reader *Buffer::reader() { return new BufferReader(this); }

} // namespace bipscript

#endif // BUFFERTYPE_H
