/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef FILEIO_H
#define FILEIO_H

#include "buffertype.h"
#include "reader.h"
#include "writer.h"

#include <errno.h>
#include <stdio.h>
#include <string.h>

namespace bipscript {

class FileHandleReader : public Reader {
    FILE *mfile;

  protected:
    FILE *filehandle() { return mfile; }

  public:
    FileHandleReader(FILE *file) : mfile(file) {}
    bool eos() override { return feof(mfile); }
    void line(std::string &str) override {
        char next;
        while (!feof(mfile) && fread(&next, 1, 1, mfile)) {
            if (next == '\n' || next == '\r') {
                break;
            }
            str += next;
        }
        if (ferror(mfile)) {
            throw std::logic_error("file reader error when reading line");
        }
    }
    bool readBool(bool &value) override {
        if (fread(&value, sizeof(bool), 1, mfile) == 1) {
            return true;
        }
        if (ferror(mfile)) {
            throw std::logic_error("file reader error when reading boolean");
        }
        return false;
    }
    bool readByte(unsigned char &value) override {
        if (fread(&value, 1, 1, mfile) == 1) {
            return true;
        }
        if (ferror(mfile)) {
            throw std::logic_error("file reader error when reading byte");
        }
        return false;
    }
    bool readFloat(float &value) override {
        if (fread(&value, sizeof(float), 1, mfile) == 1) {
            return true;
        }
        if (ferror(mfile)) {
            throw std::logic_error("file reader error when reading float");
        }
        return false;
    }
    bool readInt(long &value, size_t size) override {
        if (size > sizeof(long)) {
            std::string mesg("integer size parameter may not be greater than ");
            mesg.append(std::to_string(sizeof(long)));
            throw std::logic_error(mesg);
        }
        value = 0;
        if (fread(&value, size, 1, mfile) == 1) {
            return true;
        }
        if (ferror(mfile)) {
            throw std::logic_error("file reader error when reading integer");
        }
        return false;
    }
    bool readString(std::string &value) override {
        char next;
        while (true) {
            if (fread(&next, 1, 1, mfile) != 1) {
                if (ferror(mfile)) {
                    throw std::logic_error("file reader error when reading string");
                }
                return false;
            }
            if (next == '\0') {
                break;
            }
            value += next;
        }
        return true;
    }
    bool readBuffer(Buffer **value, size_t size) override {
        *value = (new Buffer(size))->addref();
        size_t read = fread((*value)->bytes(), sizeof(char), size, mfile);
        if (!read) {
            delete *value;
            if (ferror(mfile)) {
                throw std::logic_error("file reader error when reading buffer");
            }
            return false;
        }
        return true;
    }
    size_t tell() override {
        long ret = ftell(filehandle());
        if (ret == -1L) {
            std::string mesg("reader tell() failed: ");
            mesg.append(strerror(errno));
            throw std::logic_error(mesg);
        }
        return ret;
    }
};

class FileReader : public FileHandleReader {
  public:
    FileReader(std::string path) : FileHandleReader(fopen(path.c_str(), "rb")) {
        if (!filehandle()) {
            std::string mesg("Could not open reader for file: ");
            mesg.append(path);
            throw std::logic_error(mesg);
        }
    }
    ~FileReader() { fclose(filehandle()); }
};

class FileHandleWriter : public Writer {
    FILE *mfile;

  protected:
    FILE *filehandle() { return mfile; }

  public:
    FileHandleWriter(FILE *file) : mfile(file) {}
    void write(bool b) override { fwrite(&b, sizeof(bool), 1, mfile); }
    void write(float f) override { fwrite(&f, sizeof(float), 1, mfile); }
    void write(long l) override { fwrite(&l, sizeof(long), 1, mfile); }
    void write(const char *str) override { fwrite(str, sizeof(char), strlen(str) + 1, mfile); }
    void write(Buffer *buffer) override {
        fwrite(buffer->bytes(), sizeof(char), buffer->size(), mfile);
    }
    void print(const char *str) override { fprintf(mfile, "%s", str); }
    void println(const char *str) override { fprintf(mfile, "%s\n", str); }
};

class FileWriter : public FileHandleWriter {
  public:
    FileWriter(const char *path) : FileHandleWriter(fopen(path, "wb")) {
        if (!filehandle()) {
            std::string mesg("Could not open writer for file: ");
            mesg.append(path);
            throw std::logic_error(mesg);
        }
    }
    ~FileWriter() { fclose(filehandle()); }
};

} // namespace bipscript

#endif // FILEIO_H
