/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "io.h"
#include "bindio.h"
#include "fileio.h"

#include <fstream>
#include <stdexcept>

namespace bipscript {
namespace io {

void Path::testexists() {
    if (!exists()) {
        std::string mesg("Path does not exist: ");
        throw std::logic_error(mesg + mpath);
    }
}

Folder *Path::parent() { return new Folder(bfs::path(mpath).parent_path().c_str()); }

File::File(const char *fpath) : FSPath(fpath) {
    if (exists()) {
        testfile();
    }
}

uintmax_t File::len() {
    testexists();
    testfile();
    return bfs::file_size(mpath);
}

const char *File::text() {
    testexists();
    testfile();
    std::ifstream stream;
    stream.open(mpath);
    if (!stream) {
        throw std::logic_error("Could not read file " + mpath);
    }
    stream.seekg(0, std::ios::end);
    mlength = stream.tellg();
    textbuffer.resize(mlength + 1);
    stream.seekg(0, std::ios::beg);
    stream.read((char *)textbuffer.ptr(), mlength);
    stream.close();
    textbuffer[mlength] = '\0';
    return (char *)textbuffer.ptr();
}

Reader *File::reader() { return new FileReader(this->path()); }

Buffer *File::bytes() {
    testexists();
    testfile();
    std::ifstream stream;
    stream.open(mpath);
    if (!stream) {
        throw std::logic_error("Could not read file " + mpath);
    }
    stream.seekg(0, std::ios::end);
    mlength = stream.tellg();
    Buffer *ret = new Buffer(mlength); // deleted by release hook
    stream.seekg(0, std::ios::beg);
    stream.read((char *)ret->bytes(), mlength);
    stream.close();
    return ret;
}

void File::writeAll(const char *body) {
    fprintf(stderr, "warning: writeAll() method is deprecated, use writer() instead\n");
    std::ofstream stream;
    stream.open(mpath);
    if (!stream) {
        throw std::logic_error("Could not write to file " + mpath);
    }
    stream.write(body, strlen(body));
    stream.close();
}

void File::testfile() {
    if (!bfs::is_regular_file(mpath)) {
        std::string mesg("Path is not a file: ");
        throw std::logic_error(mesg + mpath);
    }
}

void FolderIterator::next(HSQUIRRELVM vm) {
    if (bfs::is_directory(itr->status())) {
        BindUtil::Push(vm, new Folder(itr->path().c_str()), binding::FolderObject);
    } else {
        BindUtil::Push(vm, new File(itr->path().c_str()), binding::FileObject);
    }
    ++itr;
}

Folder::Folder(const char *fpath) : FSPath(fpath) {
    if (exists()) {
        testfolder();
    }
}

Iterator *Folder::children() {
    testexists();
    testfolder();
    try {
        return new FolderIterator(mpath);
    } catch (bfs::filesystem_error &e) {
        throw std::logic_error("Error creating folder object");
    }
}

void Folder::testfolder() {
    if (!bfs::is_directory(mpath)) {
        std::string mesg("Path is not a folder: ");
        throw std::logic_error(mesg + mpath);
    }
}

Writer *bipscript::io::File::writer() { return new FileWriter(this->path().c_str()); }

bool File::remove() { return bfs::remove(mpath); }

} // namespace io
} // namespace bipscript
