/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef JSONTYPE_H
#define JSONTYPE_H

#include "io.h"
#include "pdjson.h"
#include "scripttypes.h"

#include <list>

#define JSON_PP_INDENT "  "

namespace bipscript {
namespace io {

class JsonReadHandler {
    HSQUIRRELVM vm;
    enum ReadState { ROOT, ARRAY, OBJECT, KEY };
    ReadState parent;
    uint32_t depth;

  public:
    JsonReadHandler(HSQUIRRELVM &vm) : vm(vm), parent(ROOT), depth(0) {}

    void parse(json_stream *s) {
        bool done = false;
        while (!done) {
            enum json_type type = json_next(s);
            switch (type) {
            case JSON_NULL:
                sq_pushnull(vm);
                appendToParent();
                break;
            case JSON_TRUE:
                sq_pushbool(vm, true);
                appendToParent();
                break;
            case JSON_FALSE:
                sq_pushbool(vm, false);
                appendToParent();
                break;
            case JSON_NUMBER: {
                const char *numstr = json_get_string(s, 0);
                if (strchr(numstr, '.')) {
                    sq_pushfloat(vm, json_get_number(s));
                } else {
                    sq_pushinteger(vm, std::stoi(numstr));
                }
                appendToParent();
                break;
            }
            case JSON_STRING: {
                const char *str = json_get_string(s, 0);
                sq_pushstring(vm, str, strlen(str));
                if (parent == OBJECT) {
                    parent = KEY;
                } else {
                    appendToParent();
                }
                break;
            }
            case JSON_ARRAY:
                sq_newarray(vm, 0);
                parent = ARRAY;
                depth++;
                break;
            case JSON_OBJECT:
                sq_newtable(vm);
                parent = OBJECT;
                depth++;
                break;
            case JSON_OBJECT_END:
                endComposite();
                break;
            case JSON_ARRAY_END:
                endComposite();
                break;
            case JSON_ERROR:
            case JSON_DONE:
                break;
            }
            if (type == JSON_DONE) {
                done = true;
            } else if (type == JSON_ERROR) {
                // TODO: pop stack?
                std::string err("JSON parse error byte ");
                err.append(std::to_string(json_get_position(s)));
                err.append(": ");
                err.append(json_get_error(s));
                throw std::logic_error(err);
            }
        }
        json_close(s);
    }

  private:
    void appendToParent() {
        if (parent == ARRAY) {
            sq_arrayappend(vm, -2);
        } else if (parent == KEY) {
            sq_newslot(vm, -3, false);
            parent = OBJECT;
        }
    }
    // end an object or array
    void endComposite() {
        if (--depth) {
            // pop parent
            switch (sq_gettype(vm, -2)) {
            case OT_STRING: // = key within object
                parent = KEY;
                break;
            case OT_ARRAY:
                parent = ARRAY;
                break;
            default:
                throw std::logic_error("Unknown type not supported here");
            }
            appendToParent();
        }
    }
};

class JsonBindHandler {
    HSQUIRRELVM vm;
    enum ReadState { ROOT, ARRAY, OBJECT, KEY };
    ReadState parent;
    uint32_t depth;
    int typeloc[4096]; // TODO: resize with VM stack
    uint32_t typetop;  // type definition position on stack
    bool mverbose;

  public:
    JsonBindHandler(HSQUIRRELVM &vm, HSQOBJECT &clazz, bool verbose)
        : vm(vm), parent(ROOT), depth(0), typetop(0), mverbose(verbose) {
        // find class argument in stack (should be there!)
        for (int i = sq_gettop(vm); i > 0; i--) {
            HSQOBJECT obj;
            sq_getstackobj(vm, i, &obj);
            if (obj._type == OT_CLASS && obj._unVal.pClass == clazz._unVal.pClass) {
                typeloc[0] = i;
                break;
            }
        }
    }

    void parse(json_stream *s) {
        bool done = false;
        enum json_type type = json_next(s);
        while (!done) {
            switch (type) {
            case JSON_NULL:
                sq_pushnull(vm);
                appendToParent();
                break;
            case JSON_TRUE:
                sq_pushbool(vm, true);
                appendToParent();
                break;
            case JSON_FALSE:
                sq_pushbool(vm, false);
                appendToParent();
                break;
            case JSON_NUMBER: {
                const char *numstr = json_get_string(s, 0);
                if (strchr(numstr, '.')) {
                    double number = json_get_number(s);
                    sq_pushfloat(vm, json_get_number(s));
                } else {
                    long long number = std::stoll(numstr);
                    sq_pushinteger(vm, number);
                }
                appendToParent();
                break;
            }
            case JSON_STRING: {
                const char *str = json_get_string(s, 0);
                if (parent == OBJECT) {
                    sq_pushstring(vm, str, strlen(str));
                    // pop key and push attributes table for this class
                    uint32_t type = 0;
                    if (SQ_SUCCEEDED(sq_getattributes(vm, typeloc[typetop]))) {
                        // attributes table for this key now on stack
                        sq_pushstring(vm, "type", 4);
                        if (SQ_SUCCEEDED(sq_get(vm, -2))) {
                            sq_remove(vm, -2); // remove attributes table
                            if (sq_gettype(vm, -1) == OT_ARRAY) {
                                sq_pushinteger(vm, 0); // [0] array index
                                sq_get(vm, -2);        // get first element
                                sq_remove(vm, -2);     // remove array
                            }
                            type = sq_gettop(vm); // store location of typedef
                        } else if (mverbose) {
                            printf("field has no type mapping, mapping to table\n");
                        }
                    } else {
                        if (mverbose) {
                            const SQChar *name;
                            sq_getstring(vm, -1, &name);
                            printf("warning: field %s does not exist in object\n", name);
                        }
                        sq_poptop(vm);
                        sq_pushnull(vm); // no type mapping
                    }
                    sq_pushstring(vm, str, strlen(str));
                    typeloc[++typetop] = type;
                    parent = KEY;
                } else {
                    sq_pushstring(vm, str, strlen(str));
                    appendToParent();
                }
                break;
            }
            case JSON_ARRAY:
                sq_newarray(vm, 0);
                parent = ARRAY;
                depth++;
                break;
            case JSON_OBJECT: {
                uint32_t type = typeloc[typetop];
                if (type) {
                    sq_createinstance(vm, type);
                } else {
                    sq_newtable(vm);
                }
                parent = OBJECT;
                depth++;
                break;
            }
            case JSON_OBJECT_END:
                endComposite();
                break;
            case JSON_ARRAY_END:
                endComposite();
                break;
            case JSON_ERROR:
            case JSON_DONE:
                break;
            }
            if (type == JSON_DONE) {
                done = true;
            } else if (type == JSON_ERROR) {
                // TODO: pop stack?
                std::string err("JSON parse error byte ");
                err.append(std::to_string(json_get_position(s)));
                err.append(": ");
                err.append(json_get_error(s));
                throw std::logic_error(err);
            } else {
                type = json_next(s);
            }
        }
        json_close(s);
    }

  private:
    void appendToParent() {
        if (parent == ARRAY) {
            sq_arrayappend(vm, -2);
        } else if (parent == KEY) {
            if (typetop && typeloc[typetop - 1]) { // in a mapped object
                if (SQ_FAILED(sq_set(vm, -4))) {
                    if (mverbose) {
                        const SQChar *name;
                        sq_getstring(vm, -2, &name);
                        printf("bind to object with key %s failed\n", name);
                    }
                    sq_pop(vm, 3); // pop type, key, value
                } else {
                    sq_pop(vm, 1); // pop type for this key
                }
            } else { // in a table
                if (SQ_FAILED(sq_newslot(vm, -4, false))) {
                    if (mverbose) {
                        const SQChar *name;
                        sq_getstring(vm, -2, &name);
                        printf("bind to object with key %s failed\n", name);
                    }
                    sq_pop(vm, 3); // pop type, key, value
                } else {
                    sq_pop(vm, 1); // pop type for this key
                }
            }
            parent = OBJECT;
            typetop--;
        }
    }
    // end an object or array
    void endComposite() {
        if (--depth) { // else parse complete
            // pop parent
            switch (sq_gettype(vm, -2)) {
            case OT_STRING: { // = key within object
                parent = KEY;
                break;
            }
            case OT_ARRAY:
                parent = ARRAY;
                break;
            default:
                throw std::logic_error("JSON parse error: unknown type not supported here");
            }
            appendToParent();
        }
    }
};

class Json {
    bool mverbose;

  public:
    Json() : mverbose(false) {}
    bool verbose() { return mverbose; }
    void verbose(bool verbose) { mverbose = verbose; }
    void read(HSQUIRRELVM &vm, const char *json) {
        JsonReadHandler handler(vm);
        json_stream s[1];
        json_open_string(s, json);
        json_set_streaming(s, 0);
        handler.parse(s);
    }
    void read(HSQUIRRELVM &vm, io::File &file) {
        if (file.exists()) {
            JsonReadHandler handler(vm);
            json_stream s[1];
            FILE *f = fopen(file.path().c_str(), "rb");
            json_open_stream(s, f);
            json_set_streaming(s, 0);
            handler.parse(s);
            fclose(f); // TODO: exception handling
        } else {
            std::string mesg("JSON file does not exist: ");
            mesg.append(file.path());
            throw std::logic_error(mesg);
        }
    }
    void read(HSQUIRRELVM &vm, const char *json, HSQOBJECT &clazz) {
        JsonBindHandler handler(vm, clazz, mverbose);
        json_stream s[1];
        json_open_string(s, json);
        json_set_streaming(s, 0);
        handler.parse(s);
    }

    void read(HSQUIRRELVM &vm, io::File &file, HSQOBJECT &clazz) {
        if (file.exists()) {
            JsonBindHandler handler(vm, clazz, mverbose);
            json_stream s[1];
            FILE *f = fopen(file.path().c_str(), "rb");
            json_open_stream(s, f);
            json_set_streaming(s, 0);
            handler.parse(s);
            fclose(f);
        } else {
            std::string mesg("JSON file does not exist: ");
            mesg.append(file.path());
            throw std::logic_error(mesg);
        }
    }

    FILE *openfile(std::string path) {
        FILE *f = fopen(path.c_str(), "w");
        if (!f) {
            std::string mesg("Cannot write to ");
            mesg.append(path);
            mesg.append(": ");
            mesg.append(strerror(errno));
            throw std::logic_error(mesg);
        }
        return f;
    }

    void write(HSQUIRRELVM vm, io::File &file) {
        FILE *f = openfile(file.path());
        // table is -2 in stack, behind file
        writeValue(vm, -2, f, 1);
        fclose(f);
    }

    void write(HSQUIRRELVM vm, io::File &file, bool pretty) {
        FILE *f = openfile(file.path());
        // table is -3 in stack
        writeValue(vm, -3, f, pretty ? 1 : 0);
        fclose(f);
    }

    void write(HSQUIRRELVM vm) {
        // table is -1 in stack
        writeValue(vm, -1, stdout, true);
    }

    void write(HSQUIRRELVM vm, bool pretty) {
        // table is -2 in stack
        writeValue(vm, -2, stdout, pretty ? 1 : 0);
    }

    void writeValue(HSQUIRRELVM vm, SQInteger pos, FILE *f, int level) {
        switch (sq_gettype(vm, pos)) {
        case OT_NULL:
            fprintf(f, "null");
            break;
        case OT_BOOL:
            SQBool b;
            sq_getbool(vm, pos, &b);
            fprintf(f, b ? "true" : "false");
            break;
        case OT_INTEGER:
            SQInteger i;
            sq_getinteger(vm, pos, &i);
            fprintf(f, "%lld", i);
            break;
        case OT_FLOAT:
            SQFloat fl;
            sq_getfloat(vm, pos, &fl);
            fprintf(f, "%f", fl);
            break;
        case OT_STRING:
            const SQChar *str;
            sq_getstring(vm, pos, &str);
            fprintf(f, "\"%s\"", str);
            break;
        case OT_TABLE:
        case OT_CLASS:
            writeTable(vm, pos, f, level ? level + 1 : 0);
            break;
        case OT_ARRAY: {
            fprintf(f, "[");
            if (level) {
                fprintf(f, "\n");
            }
            sq_pushnull(vm);
            bool first = true;
            while (SQ_SUCCEEDED(sq_next(vm, pos - 1))) {
                // comma between entries
                if (!first) {
                    fprintf(f, ",");
                    if (level) {
                        fprintf(f, "\n");
                    }
                }
                first = false;
                for (int i = 0; i < level - 1; i++) {
                    fprintf(f, JSON_PP_INDENT);
                }
                writeValue(vm, -1, f, level);
                sq_pop(vm, 2);
            }
            sq_pop(vm, 1); // pop the iterator
            if (level) {
                fprintf(f, "\n");
                for (int i = 0; i < level - 2; i++) {
                    fprintf(f, JSON_PP_INDENT);
                }
            }
            fprintf(f, "]");
            break;
        }
        case OT_INSTANCE:
            // TODO: only works for class that defines _nexti?
            writeTable(vm, pos, f, level);
            break;
        default:
            break;
        }
    }

    void writeTable(HSQUIRRELVM vm, SQInteger pos, FILE *f, int level) {
        // iterate and get keys
        std::list<std::string> keylist;
        sq_pushnull(vm); // push iterator
        while (SQ_SUCCEEDED(sq_next(vm, pos - 1))) {
            const char *key;
            if (SQ_FAILED(sq_getstring(vm, -2, &key))) {
                throw std::logic_error("JSON table key must be a string");
            }
            keylist.push_back(key);
            sq_pop(vm, 2);
        }
        sq_pop(vm, 1);  // pop the iterator
        keylist.sort(); // sort keys

        // start table structure
        fprintf(f, "{");
        if (level) {
            fprintf(f, "\n");
        }
        // loop over keys
        for (auto it = keylist.begin(); it != keylist.end(); ++it) {
            // comma between key/value pairs
            if (it != keylist.begin()) {
                fprintf(f, ",");
                if (level) {
                    fprintf(f, "\n");
                }
            }
            // indent
            for (int i = 0; i < level - 1; i++) {
                fprintf(f, JSON_PP_INDENT);
            }
            const char *keystr = it->c_str();
            fprintf(f, "\"%s\":", keystr);
            if (level) {
                fprintf(f, " ");
            }
            // get and write value for this key
            sq_pushstring(vm, keystr, it->size());
            if (SQ_SUCCEEDED(sq_get(vm, pos - 1))) {
                writeValue(vm, -1, f, level ? level + 1 : 0);
                sq_poptop(vm);
            }
        }
        if (level) {
            fprintf(f, "\n");
            for (int i = 0; i < level - 2; i++) {
                fprintf(f, JSON_PP_INDENT);
            }
        }
        fprintf(f, "}");
    }
};

} // namespace io
} // namespace bipscript

#endif // JSONTYPE_H
