/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef XMLTYPE_H
#define XMLTYPE_H

#include "io.h"
#include "squirrel.h"
#include "sxml.h"
#include "tinyxml/tinyxml2.h"

#include <algorithm>
#include <map>

namespace bipscript {
namespace io {

struct taginfo {
    uint32_t typeloc;      // type defintion position on VM stack
    const char *textfield; // name of inner text field, if any
    std::string text;      // inner text for this tag
    taginfo() : typeloc(0), textfield(0) {}
};

class XmlBinder {

    HSQUIRRELVM vm;
    bool mverbose;
    std::string mtextfield; // table field to store tag body text
    uint32_t depth = 0;     // current depth in XML document
    uint32_t tagtop = 0;    // tag definition position on tag stack
    taginfo tagstack[4096]; // TODO: resize type stack with VM stack
    enum ReadState { ARRAY, OBJECT };
    ReadState parent;
    std::map<std::string, char> entities;
    std::map<int, std::map<std::string, int>> attributetypes;

  public:
    XmlBinder(HSQUIRRELVM vm, HSQOBJECT *clazz, bool verbose, std::string textfield);
    void processTokens(const char *buffer, const sxmltok_t tokens[], uint32_t num_tokens);
    static void typeTextField(HSQUIRRELVM vm, uint32_t typeloc, const SQChar **fieldname);

  private:
    char escape(const char *text, uint32_t textlen);
    void parseAttributes(const char *buffer, const sxmltok_t *token, uint32_t type);
};

class Xml {

    const size_t BUFFERSIZE = 16384; // XML comments cannot be larger than this size
    const size_t TOKENCOUNT = 128;   // TODO: how to size this

    bool mverbose;
    bool mwritedefault;
    std::string mtextfield;

  public:
    Xml() : mverbose(false), mwritedefault(false), mtextfield("text") {}
    bool verbose() { return mverbose; }
    void verbose(bool verbose) { mverbose = verbose; }
    std::string textfield() { return mtextfield; }
    void textfield(const char *textfield) { mtextfield = textfield; }
    bool writedefault() { return mwritedefault; }
    void writedefault(bool writedefault) { mwritedefault = writedefault; }

    void read(HSQUIRRELVM &vm, const char *xml) { read(vm, xml, nullptr, nullptr); }
    void read(HSQUIRRELVM &vm, const char *xml, HSQOBJECT &clazz) {
        read(vm, xml, nullptr, &clazz);
    }
    void read(HSQUIRRELVM &vm, io::File &iofile) { read(vm, nullptr, &iofile, nullptr); }
    void read(HSQUIRRELVM &vm, io::File &iofile, HSQOBJECT &clazz) {
        read(vm, nullptr, &iofile, &clazz);
    }
    void write(HSQUIRRELVM vm, const char *ename, io::File &file);
    std::string write(HSQUIRRELVM vm, const char *ename);

  private:
    std::string findClassName(HSQUIRRELVM vm, SQInteger idx);
    void read(HSQUIRRELVM &vm, const char *string, io::File *iofile, HSQOBJECT *clazz);
    void instance2xml(HSQUIRRELVM vm, SQInteger pos, const char *, tinyxml2::XMLPrinter &printer);
    static uint32_t linecount(const char *buffer, uint32_t bufferlen);
    FILE *openfile(std::string path) { // TODO: shared with JsonType
        FILE *f = fopen(path.c_str(), "w");
        if (!f) {
            std::string mesg("Cannot write to ");
            mesg.append(path);
            mesg.append(": ");
            mesg.append(strerror(errno));
            throw std::logic_error(mesg);
        }
        return f;
    }
};

} // namespace io
} // namespace bipscript

#endif // XMLTYPE_H
