/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "zipfile.h"

#include <stdexcept>

namespace bipscript {
namespace io {

Zip::Zip(const char *fpath) {
    memset(&zip, 0, sizeof(zip));
    mz_bool status = mz_zip_reader_init_file(&zip, fpath, 0);
    if (!status) {
        mz_zip_error errnum = mz_zip_get_last_error(&zip);
        const char *err = mz_zip_get_error_string(errnum);
        std::string mesg("Zip file error, ");
        mesg.append(err);
        mesg.append(": ");
        mesg.append(fpath);
        throw std::logic_error(mesg);
    }
    mentries = mz_zip_reader_get_num_files(&zip);
}

mz_zip_archive_file_stat Zip::statFile(int index) {
    mz_zip_archive_file_stat file_stat;
    if (!mz_zip_reader_file_stat(&zip, index, &file_stat)) {
        throw std::logic_error("Failure reading zip entry mz_zip_reader_file_stat()");
    }
    return file_stat;
}

Path *Zip::entry(const char *path) {
    int index = mz_zip_reader_locate_file(&zip, path, nullptr, MZ_ZIP_FLAG_CASE_SENSITIVE);
    mz_zip_archive_file_stat file_stat = statFile(index);
    if (mz_zip_reader_is_file_a_directory(&zip, index)) {
        std::string dirname(file_stat.m_filename);
        dirname = dirname.substr(0, dirname.length() - 1); // chop off final '/'
        return new ZipFolder(dirname.c_str());
    }
    return new ZipFile(file_stat.m_filename, this, file_stat.m_uncomp_size);
}

Buffer *Zip::bytes(const std::string &path) {
    int index = mz_zip_reader_locate_file(&zip, path.c_str(), nullptr, MZ_ZIP_FLAG_CASE_SENSITIVE);
    mz_zip_archive_file_stat file_stat = statFile(index);
    size_t size = file_stat.m_uncomp_size;
    Buffer *ret = new Buffer(size); // refcounted
    if (!mz_zip_reader_extract_to_mem(&zip, index, ret->bytes(), size, 0)) {
        throw std::logic_error("Failure reading zip entry");
    }
    return ret;
}

const char *Zip::text(const std::string &path) {
    int index = mz_zip_reader_locate_file(&zip, path.c_str(), nullptr, MZ_ZIP_FLAG_CASE_SENSITIVE);
    mz_zip_archive_file_stat file_stat = statFile(index);
    size_t size = file_stat.m_uncomp_size;
    textbuffer.resize(size + 1);
    if (!mz_zip_reader_extract_to_mem(&zip, index, textbuffer.ptr(), size, 0)) {
        throw std::logic_error("Failure reading zip entry");
    }
    textbuffer[size] = '\0';
    return (char *)textbuffer.ptr();
}

io::Path *Zip::operator[](int index) {
    mz_zip_archive_file_stat file_stat = statFile(index);
    if (mz_zip_reader_is_file_a_directory(&zip, index)) {
        std::string dirname(file_stat.m_filename);
        dirname = dirname.substr(0, dirname.length() - 1); // chop off final '/'
        return new ZipFolder(dirname.c_str());
    }
    return new ZipFile(file_stat.m_filename, this, file_stat.m_uncomp_size);
}

ZipFile::~ZipFile() {
    if (!mparent->delref()) {
        delete mparent;
    }
}

} // namespace io
} // namespace bipscript
