/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef Timebase_H
#define Timebase_H

#include "bindjack.h"
#include "bindutil.h"
#include "jacktransport.h"
#include "objectcache.h"
#include "processclosure.h"
#include "scripttypes.h"
#include "variableclock.h"

namespace bipscript {
namespace jack {

class SyncHandler : public ScriptFunction {
    bool mwait;

  public:
    SyncHandler(ScriptFunction &function, bool wait) : ScriptFunction(function), mwait(wait) {}
    bool wait() { return mwait; }
};

class Timebase;

class SyncClosure : public ProcessClosure {
    uint16_t state;
    Position position; // new position
    float measure;     // when the repo happened
    Timebase *clock;

  protected:
    void addParameters() {
        addInteger(state);
        BindUtil::PushCopy(vm, position, binding::PositionObject);
        addFloat(measure);
    }

  public:
    SyncClosure(ScriptFunction function, uint16_t state, jack_position_t &position, float measure,
                Timebase *master)
        : ProcessClosure(function, 0), state(state), position(position), measure(measure),
          clock(master) {}
    void finished();
};

class Timebase : public time::AdvancedClock {
    Transport transport;
    // on reposition
    std::atomic<bool> mrepositionComplete;
    std::atomic<SyncHandler *> onSyncHandler;

  public:
    Timebase(jack_client_t *client, float bpm, time::Signature &signature)
        : AdvancedClock(bpm, signature), transport(client), mrepositionComplete(true),
          onSyncHandler(nullptr) {}
    // jack API
    void timebase(jack_transport_state_t state, jack_nframes_t nframes, jack_position_t *pos,
                  int new_pos);
    // internal API
    void update(samples_t) override;
    // script API
    void start() override {
        transport.start();
        statusUpdate(mtempo);
    }
    void stop() override {
        transport.stop();
        statusUpdate(mtempo);
    }
    void reposition(float location);
    // this reposition called from jack transport in process thread
    void handlerCompleted() { mrepositionComplete.store(true); }
    void sync(jack_transport_state_t state, jack_position_t *pos);
    bool syncComplete() { return mrepositionComplete.load(); }
    void onSync(ScriptFunction &handler, bool wait = false);
    void recycle() override { clearRemotes(); } // lifecycle managed by cache
};

class TimebaseCache : public ObjectCache {
    bool active;
    std::atomic<Timebase *> cachedMaster; // TODO: why is this atomic
    TimebaseCache() : active(false) {}

  public:
    static TimebaseCache &instance() {
        static TimebaseCache instance;
        return instance;
    }
    void scriptComplete() override;
    void reset(bool) override {}
    Timebase *newTimebase(float bpm);
    Timebase *newTimebase(float bpm, time::Signature sig);
    ~TimebaseCache();
};

} // namespace jack
} // namespace bipscript

#endif // Timebase_H
