/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef AUDIOPLUGIN_H
#define AUDIOPLUGIN_H

#include "lv2/atom/atom.h"
#include "lv2/atom/util.h"
#include "lv2/options/options.h"
#include "lv2/state/state.h"
#include "lv2/worker/worker.h"
#include <jack/ringbuffer.h>
#include <lilv-0/lilv/lilv.h>
#include <semaphore.h>

#include "audiosink.h"
#include "midisink.h"
#include "processorcache.h"

namespace bipscript {

namespace lv2 {

typedef std::map<std::string, uint32_t> Lv2UridMap;

class UridMapper {
    Lv2UridMap uridMap;

  public:
    LV2_URID uriToId(const char *uri);
    const char *idToUri(LV2_URID urid);
};

class PathMapper {
  public:
    char *mapAbsolutePath(const char *abstractPath);
    char *mapAbstractPath(const char *absolutePath);
};

class Constants {
  public:
    Constants(LilvWorld *);
    ~Constants();
    // port types
    LilvNode *lv2AtomPort;
    LilvNode *lv2AudioPort;
    LilvNode *lv2InputPort;
    LilvNode *lv2OutputPort;
    LilvNode *lv2ControlPort;
    // midi
    LilvNode *lv2MidiEvent;
    // atom
    LilvNode *lv2AtomSequence;
    LilvNode *lv2AtomSupports;
    LilvNode *lv2AtomBufferType;
    // presets
    LilvNode *lv2Presets;
    // worker
    LilvNode *lv2WorkerInterface;
    LilvNode *lv2WorkerSchedule;
    // rdf
    LilvNode *lv2RdfsLabel;
};

class AtomTypes {
  public:
    LV2_URID doubleType;
    LV2_URID floatType;
    LV2_URID intType;
    LV2_URID longType;
    LV2_URID stringType;
};

class MidiEvent {
  public:
    static uint32_t midiEventTypeId;
    LV2_Atom_Event event;
    uint8_t buffer[3];
    MidiEvent() {
        event.body.type = midiEventTypeId;
        event.body.size = 3;
    }
};

class MidiOutput : public Listable<MidiOutput>, public midi::Connection {
    LV2_Atom_Sequence *atomSequence;
    const uint32_t atomSize;
    midi::RawMessage *message;
    Event<midi::RawMessage> event;
    LV2_Atom_Event *iterator;
    static const size_t MAX_EVENT_SIZE = 3;

  public:
    MidiOutput(midi::Source *source, LV2_Atom_Sequence *seq, uint32_t size)
        : midi::Connection(source), atomSequence(seq), atomSize(size),
          message(new midi::RawMessage(MAX_EVENT_SIZE)),
          event(message, Command::Type::TRANSIENT, 1.0) {}
    virtual ~MidiOutput() {}
    LV2_Atom_Sequence *getAtomSequence() { return atomSequence; }
    void clear() { atomSequence->atom.size = atomSize; }
    // EventConnection interface
    midi::RawEvent *firstEvent() override;
    midi::RawEvent *nextEvent() override;
};

class PortDef {
    // const char *name;
    uint32_t mindex;

  public:
    uint32_t index() const { return mindex; }
    void index(uint32_t index) { mindex = index; }
};

class AtomPortDef : public PortDef {
    bool mmidi = false;

  public:
    void midi(bool midi) { mmidi = midi; }
    bool midi() const { return mmidi; }
};

class ControlPortDef : public PortDef {
    const char *mname;
    float mvalue;
    float mdfault;
    float mminimum;
    float mmaximum;

  public:
    const char *name() const { return mname; }
    void name(const char *name) { mname = name; }
    float value() { return mvalue; }
    void value(float value) { mvalue = value; }
    float dfault() { return mdfault; }
    void dfault(float dfault) { mdfault = dfault; }
    float minimum() { return mminimum; }
    void minimum(float minimum) { mminimum = minimum; }
    float maximum() { return mmaximum; }
    void maximum(float maximum) { mmaximum = maximum; }
};

class ControlPort {
    const ControlPortDef *mdef;

  public:
    ControlPort(const ControlPortDef *def) : mdef(def) {}
    float value;
};

class ControlChange {
    const char *mport;
    const float mvalue;

  public:
    ControlChange(const char *port, float value) : mport(port), mvalue(value) {}
    const char *port() { return mport; }
    float value() { return mvalue; }
};

class ControlPortChange {
    ControlPort *mport;
    float mvalue;

  public:
    ControlPortChange(ControlPort *port, float value) : mport(port), mvalue(value) {}
    ControlPort *port() { return mport; }
    float value() { return mvalue; }
};

class State {
    std::map<uint32_t, std::string> stateMap;

  public:
    State(const char *filename);
    State(ScriptHashIterator &state);
    const void *retrieveState(uint32_t key, size_t *size, uint32_t *type, uint32_t *flags);
    int getHash();
};

class WorkResponseClosure : public ProcessClosure {

  protected:
    void addParameters() { addTimeParameter(); }

  public:
    WorkResponseClosure(ScriptFunction function) : ProcessClosure(function, 0) {}
};

class Worker {
    LV2_Handle handle;
    LV2_Worker_Interface *interface;
    jack_ringbuffer_t *requestBuffer;
    jack_ringbuffer_t *responseBuffer;
    char response[2048];
    pthread_t thread;
    sem_t semaphore;
    std::atomic<bool> active;
    std::atomic<ScriptFunction *> workResponseHandler;

  public:
    Worker();
    ~Worker();
    void start();
    void run();
    LV2_Worker_Status scheduleWork(const void *, uint32_t size);
    LV2_Worker_Status queueResponse(const void *, uint32_t size);
    void respond();
    void setInstance(LilvInstance *instance);
    void onWorkResponse(ScriptFunction &handler);
};

class PluginType {
    std::string muri;
    std::string mname;
    const LilvPlugin *mplugin;
    PortDef *maudioInput;
    const uint32_t maudioInputCount;
    PortDef *maudioOutput;
    const uint32_t maudioOutputCount;
    ControlPortDef *mcontrolInput;
    const uint32_t mcontrolInputCount;
    ControlPortDef *mcontrolOutput;
    const uint32_t mcontrolOutputCount;
    AtomPortDef *matomInput;
    const uint32_t matomInputCount;
    AtomPortDef *matomOutput;
    const uint32_t matomOutputCount;
    uint32_t mmidiInputCount;
    uint32_t mmidiOutputCount;

  public:
    PluginType(std::string &uri, const LilvPlugin *plugin, const Constants &uris);
    ~PluginType();
    const LilvPlugin *plugin() const { return mplugin; }
    uint32_t audioInputCount() const { return maudioInputCount; }
    uint32_t audioOutputCount() const { return maudioOutputCount; }
    uint32_t controlInputCount() const { return mcontrolInputCount; }
    uint32_t controlOutputCount() const { return mcontrolOutputCount; }
    uint32_t atomInputCount() const { return matomInputCount; }
    uint32_t atomOutputCount() const { return matomOutputCount; }
    uint32_t midiInputCount() const { return mmidiInputCount; }
    uint32_t midiOutputCount() const { return mmidiOutputCount; }
    const PortDef &audioInput(uint32_t i) const { return maudioInput[i]; }
    const PortDef &audioOutput(uint32_t i) const { return maudioOutput[i]; }
    const ControlPortDef &controlInput(uint32_t i) const { return mcontrolInput[i]; }
    const ControlPortDef &controlOutput(uint32_t i) const { return mcontrolOutput[i]; }
    const AtomPortDef &atomInput(uint32_t i) const { return matomInput[i]; }
    const AtomPortDef &atomOutput(uint32_t i) const { return matomOutput[i]; }
    const std::string uri() const { return muri; }
    const std::string name() const { return mname; }
};

class Plugin : public audio::Source, public audio::Sink, public midi::Source, public midi::Sink {
    const PluginType *type;
    LilvInstance *instance;
    LV2_Feature **lv2Features;
    const uint32_t ATOM_CAPACITY = 4096;
    LV2_Atom_Sequence **matomInput;
    LV2_Atom_Sequence **midiInput;
    LV2_Atom_Sequence **matomOutput;
    MidiOutput **mmidiOutput;
    // control ports
    TimelineSet<ControlPortChange> controlBuffer;
    std::map<std::string, ControlPort *> controlMap;
    boost::lockfree::spsc_queue<ControlPortChange *> controlChanges;
    Collector<ControlPortChange> controlCollector;
    float *controlOutput; // TODO: currently unused
    // worker
    Worker *worker;
    pthread_mutex_t mmutex; // for preset loading synchronization

  public:
    static AtomTypes atomTypes;
    Plugin(const PluginType *type, LilvInstance *instance, LV2_Feature **lv2Features,
           Worker *worker);
    ~Plugin();
    // LV2 methods
    void setPortValue(const char *, const void *, uint32_t);
    void *portValue(const char *, uint32_t *, uint32_t *);
    // script API
    void control(const char *symbol, float value);
    void preset(const char *name);
    void saveState(const char *dir);
    void schedule(ControlChange &change, float measure, Clock &clock = Clock::defaultClock());
    // Source interface
    bool connectsTo(AbstractSink *source) override;
    // Processor interface
    bool doProcess(samples_t time) override;
    void restart() override;
    void restartScript() override;
    // callbacks
    void onWorkResponse(ScriptFunction &handler) { worker->onWorkResponse(handler); }

  private:
    ControlPort *getPort(const char *symbol);
    void connectPort(int index, Plugin *other, int otherIndex);
    void print();
};

class PluginFactory : public ProcessorCache<Plugin> {
    LilvWorld *world;
    const LilvPlugins *mplugins; // lilv_world_get_all_plugins
    jack_nframes_t msampleRate;  // from AudioEngine
    const Constants lv2Constants;
    // for caching
    std::map<std::string, const PluginType *> typeMap;
    std::map<int, int> instanceCount;
    // for features
    LV2_Feature *mlv2Features[6];
    std::map<std::string, bool> supported;
    // options feature
    jack_nframes_t blockLength;
    LV2_Options_Option options[3];
    // urid features
    LV2_URID_Map map;
    LV2_URID_Unmap unmap;
    PathMapper pathMapper;
    LV2_State_Map_Path mapPath;
    // use instance()
    PluginFactory();
    // private methods
    void restartCache() override { instanceCount.clear(); }
    const PluginType *getPluginType(std::string &uriString);
    LV2_Feature *workFeature(Worker *worker);
    const LilvPlugins *plugins();

  public:
    UridMapper uridMapper;
    ~PluginFactory();
    static PluginFactory &instance() {
        static PluginFactory instance;
        return instance;
    }
    void sampleRate(jack_nframes_t rate) { this->msampleRate = rate; }
    void bufferSize(jack_nframes_t size) { this->blockLength = size; }
    Plugin *newPlugin(const char *uri, const char *preset, State *state);
    Plugin *newPlugin(const char *uri, const char *preset) { return newPlugin(uri, preset, 0); }
    Plugin *newPlugin(const char *uri, State &state) { return newPlugin(uri, 0, &state); }
    Plugin *newPlugin(const char *uri) { return newPlugin(uri, 0, 0); }
    Plugin *newPlugin(io::File &file);
    LilvWorld *getWorld() { return world; }
    LV2_URID_Map *getMap() { return &map; }
    LV2_URID_Unmap *getUnmap() { return &unmap; }
    const Constants &getConstants() { return lv2Constants; }
};

} // namespace lv2
} // namespace bipscript

#endif // AUDIOPLUGIN_H
