/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <getopt.h>
#include <iostream>
#include <signal.h>

#include "script.h"
#include "scripthost.h"
#include "version.h"

#include "audioport.h"
#include "clip.h"
#include "extension.h"
#include "lv2plugin.h"
#include "midiport.h"
#include "mixer.h"
#include "onsetdetector.h"
#include "oscinput.h"
#include "oscoutput.h"
#include "statusserver.h"
#include "xtprocessor.h"

namespace bipscript {

int sighupRestart = 0;

void signal_handler(int sig) {
    if (sig == SIGHUP && sighupRestart) {
        ScriptHost::instance().requestRestart();
    } else {
        std::cerr << "caught signal " << sig << ", exiting" << std::endl;
        ScriptHost::instance().requestExit(0);
    }
}
} // namespace bipscript

using namespace bipscript;

// option definitions
int unbufferStdout = 0;
static struct option long_options[] = {
    {"unbuffer-stdout", no_argument, &unbufferStdout, 1},
    {"sighup-restart", no_argument, &sighupRestart, 1},
    {"status-port", required_argument, 0, 's'},
    {"version", no_argument, nullptr, 'v'},
    {nullptr, 0, nullptr, 0} // null terminator
};

int main(int argc, char **argv) {
    // option flags
    bool version = false;
    char *statusPort = nullptr;

    // read options
    while (true) {
        int option_index = 0;
        int c = getopt_long(argc, argv, "v", long_options, &option_index);
        if (c == -1) {
            break;
        } // end of options
        switch (c) {
        case 0:
            break;
        case 's':
            statusPort = strdup(optarg);
            break;
        case 'v':
            version = true;
            break;
        case '?':
            return 1;
        default:
            abort();
        }
    }

    // show version
    if (version) {
        std::cout << "bipscript version ";
        std::cout << BIPSCRIPT_VERSION << std::endl;
        return 0;
    }

    const int fileindex = optind;
    // not enough arguments
    if (fileindex >= argc) {
        std::cerr << "error: no script file given" << std::endl;
        return 1;
    }

    // check given script file
    const char *filearg = argv[fileindex];
    bfs::path filePath(filearg);
    if (!exists(filePath)) {
        std::cerr << "error: script file does not exist: " << filearg << std::endl;
        return 2;
    }
    if (is_directory(filePath)) {
        std::cerr << "error: script file is a directory: " << filearg << std::endl;
        return 2;
    }

    // stdout unbuffering
    if (unbufferStdout) {
        setvbuf(stdout, static_cast<char *>(nullptr), _IONBF, 0);
    }
    // status port
    StatusServer &statusserver = StatusServer::instance();
    if (statusPort) {
        try {
            statusserver.port(statusPort);
            statusserver.activate();
            free(statusPort);
        } catch (std::invalid_argument e) {
            std::cerr << "bad status port argument: " << statusPort << std::endl;
            return 3;
        }
        //        catch(std::logic_error e) {
        //        }
    }

    // create script host
    ScriptHost &host = ScriptHost::instance();
    host.init(filePath);

    // initialize script package
    script::Script::instance().init(argc, argv, fileindex);

    // add object caches
    ObjectCache *caches[] = {&audio::AudioInputPortFactory::instance(),
                             &audio::ClipFactory::instance(),
                             &audio::MixerFactory::instance(),
                             &lv2::PluginFactory::instance(),
                             &midi::MidiInputPortFactory::instance(),
                             &midi::MidiOutputPortFactory::instance(),
                             &osc::InputFactory::instance(),
                             &osc::OutputFactory::instance(),
                             &audio::OnsetDetectorFactory::instance(),
                             &audio::OutputPortSetFactory::instance(),
                             &XtAudioProcessorMap::instance(),
                             &script::Script::instance()};
    host.objectCaches(12, caches);

    signal(SIGQUIT, bipscript::signal_handler);
    signal(SIGTERM, bipscript::signal_handler);
    signal(SIGHUP, bipscript::signal_handler);
    signal(SIGINT, bipscript::signal_handler);

    // run script
    int status = host.run();
    // script has ended
    ExtensionManager::instance().shutdown();
    if (AudioEngine::loaded) {
        AudioEngine::instance().shutdown();
    }
    if (StatusServer::instance().active()) {
        StatusServer::instance().shutdown();
    }
    CollectorThread::instance().shutdown();
    return status;
}
