/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "abc.h"
#include <regex>

#include "midifile.h"
#include "module_midi.h"

namespace bipscript {
namespace midi {

std::map<char, int> AbcKey::pitches = {{'C', 0}, {'D', 2}, {'E', 4}, {'F', 5},
                                       {'G', 7}, {'A', 9}, {'B', 11}};

std::map<int, std::string> Abc::notename = {{0, "C"},  {1, "^C"}, {2, "D"},   {3, "_E"},
                                            {4, "E"},  {5, "F"},  {6, "^F"},  {7, "G"},
                                            {8, "_A"}, {9, "A"},  {10, "_B"}, {11, "B"}};

AbcKey::AbcKey(char tonic, int accidental, std::string mode, std::vector<std::string> *extraaccs) {

    // white keys
    std::map<char, int> steps = {{'C', 0}, {'D', 1}, {'E', 2}, {'F', 3},
                                 {'G', 4}, {'A', 5}, {'B', 6}};
    const std::vector<char> notenames = {'C', 'D', 'E', 'F', 'G', 'A', 'B'};
    const std::vector<int> semitones = {0, 2, 4, 5, 7, 9, 11};

    // interval map
    std::map<std::string, std::vector<int>> intmap;
    intmap["maj"] = {2, 2, 1, 2, 2, 2, 0};
    intmap["exp"] = intmap["maj"];
    intmap["min"] = {2, 1, 2, 2, 1, 2, 0};
    intmap["aeo"] = intmap["min"];
    intmap["loc"] = {1, 2, 2, 1, 2, 2, 0};
    intmap["ion"] = intmap["maj"];
    intmap["dor"] = {2, 1, 2, 2, 2, 1, 0};
    intmap["phr"] = {1, 2, 2, 2, 1, 2, 0};
    intmap["lyd"] = {2, 2, 2, 1, 2, 2, 0};
    intmap["mix"] = {2, 2, 1, 2, 2, 1, 0};
    const std::vector<int> *interval;

    // choose key
    if (mode == "") {
        if (tonic == tolower(tonic)) {
            interval = &intmap["min"];
        } else {
            interval = &intmap["maj"];
        }
    } else {
        if (mode.length() > 3) {
            mode = mode.substr(0, 3);
        }
        if (intmap.find(mode) != intmap.end()) {
            interval = &intmap[mode];
        } else {
            std::string mesg("unrecognized mode in key signature: ");
            mesg.append(mode);
            throw std::logic_error(mesg);
        }
    }
    tonic = toupper(tonic);

    // create modmap from key
    int step = steps[tonic];
    int pitch = pitches[tonic] + accidental;
    for (int i = 0; i < 7; i++) {
        char notename = notenames[step];
        int natural = semitones[step];
        modmap[notename] = pitchdiff(pitch, natural);
        pitch = (pitch + (*interval)[i]) % 12;
        step = (step + 1) % 7;
    }

    // add extra accidentals
    if (extraaccs) {
        std::vector<std::string> &list = *extraaccs;
        for (std::string eacc : list) {
            if (eacc.length() > 1) {
                char note = toupper(eacc[1]);
                switch (eacc[0]) {
                case '_':
                    modmap[note] = -1;
                    break;
                case '^':
                    modmap[note] = 1;
                    break;
                case '=':
                    modmap[note] = 0;
                    break;
                }
            }
        }
    }
}

int AbcKey::pitchdiff(int pitch, int natural) {
    if (!pitch && natural == 11) {
        return 1;
    } else if (!natural && pitch == 11) {
        return -1;
    }
    return pitch - natural;
}

Sequence *Abc::sequence(const char *abc) {
    abcScanner *s = scanner(abc);
    const char *c = abc;
    if (*c == '\n') {
        abcerror(s, "sequence abc should not start with newline(s)");
    }
    s->state = S_TUNE;
    SequenceList *seqlist = parse(s, nullptr);
    Sequence *ret = (*seqlist)[0];
    delete seqlist;
    return ret;
}

File *Abc::file(const char *abc) {
    abcScanner *s = scanner(abc);
    Sequence *metaseq = new Sequence(); // file track 0 for meta events
    SequenceList *seqlist = parse(s, metaseq);
    seqlist->insert(seqlist->begin(), metaseq);
    File *ret = new File(*seqlist);
    delete seqlist;
    return ret;
}

std::string Abc::abc(Note &note) {

    std::string abcbuffer;

    int pitch = note.pitch() % 12;
    int octave = note.pitch() / 12;
    abcbuffer = notename[pitch];

    // lower case higher notes
    if (octave > 5) {
        int end = abcbuffer.length() - 1;
        abcbuffer[end] = tolower(abcbuffer[end]);
    }

    // add octave markers
    for (int i = 0; i < (5 - octave); i++) {
        abcbuffer += ",";
    }
    for (int i = 6; i < octave; i++) {
        abcbuffer += "'";
    }

    // add duration
    const long precision = 10000;
    long duration = round(4.0 * note.duration() * precision);
    long divisor = gcd(duration, precision);
    long denominator = precision / divisor;
    long numerator = duration / divisor;
    if (numerator != 1) {
        abcbuffer += std::to_string(numerator);
    }
    if (denominator != 1) {
        abcbuffer += "/";
        abcbuffer += std::to_string(denominator);
    }

    return abcbuffer;
}

uint Abc::gcd(uint x, uint y) { return y > 0 ? gcd(y, x % y) : x; }

SequenceList *Abc::parse(abcScanner *s, Sequence *metaseq) {

    time::Signature defaultsig(4, 4);
    std::vector<AbcKey *> keylist;
    AbcKey *defkey = new AbcKey();
    keylist.push_back(defkey);
    int abcnote = 8; // default note is an 8th note

    // basic stress pattern, mimic abcmidi
    StressPattern sp(80);
    sp.addStress(0.0, 105);
    sp.addStress(0.5, 95);

    std::map<std::string, AbcVoice *> voicemap;                    // map of voices
    voicemap["1"] = new AbcVoice(abcnote, defkey, defaultsig, sp); // add default voice
    AbcVoice *activeVoice = voicemap["1"];                         // set default voice to active
    std::vector<AbcVoice *> voicelist;                             // list of voices
    voicelist.push_back(activeVoice);

    bool header = metaseq;
    int fieldcount = 0;
    int tunecount = 0;
    try {
        while (abcNextToken(s) != T_EOF) {
            switch (abcToken(s)) {

            case T_ENDLINE:
                break;
            case T_EMPTYLINE:
                break;

            case T_SPACE:
                switch (abcSpace(s, type)) {
                case 'W':
                    break;
                case 'B':
                    break;
                }
                break;

            case T_TEXT:
                break;

            case T_COMMENT: {
                std::string text((char *)abcComment(s, string), abcComment(s, string_len));
                activeVoice->add(*(new Text(text.c_str())));
            } break;

            case T_NOTE:
                activeVoice->note(abcNote(s, pitch), abcNote(s, octave), abcNote(s, bending),
                                  abcNote(s, bend), (double)abcNote(s, duration));
                header = false;
                break;

            case T_BROKENLEFT: // broken rhythms
            case T_BROKENRIGHT:
                activeVoice->brokenRhythm(abcBroken(s, type), abcBroken(s, dots));
                header = false;
                break;

            case T_BAR:
                activeVoice->bar(abcBar(s, starts), abcBar(s, ends), abcBar(s, string_len));
                header = false;
                break;

            case T_ENDING:
                activeVoice->ending(abcEnding(s, num));
                break;

            case T_MIDI:
                break;

            case T_REST:
                activeVoice->rest(abcRest(s, type), abcRest(s, duration));
                header = false;
                break;

            // apparently unused, T_REST handles these cases
            case T_INVREST:
            case T_MULTIREST:
                break;

            case T_CHORDSTART:
                activeVoice->chordStart();
                header = false;
                break;

            case T_CHORDEND:
                activeVoice->chordEnd();
                break;

            case T_GCHORD:
                header = false;
                break;

            case T_SPACER:
                break;

            case T_TIE:
                activeVoice->tieNote();
                break;

            case T_TUPLET:
                activeVoice->tuplet(abcTuplet(s, notes), abcTuplet(s, time), abcTuplet(s, span));
                break;

            case T_BEGINTEXT:
                break;
            case T_ENDTEXT:
                break;

            case T_ANNOTATION: {
                std::string text((char *)abcAnnotation(s, string), abcAnnotation(s, string_len));
                activeVoice->add(*(new Marker(text.c_str())));
            } break;

            case T_LYR_SYLLABLE:
                break;
            case T_LYR_BLANK:
                break;
            case T_LYR_BAR:
                break;
            case T_LYR_SPACE:
                break;
            case T_LYR_HOLD:
                break;
            case T_LYR_VERSE:
                break;

            case T_CONTINUE:
            case T_LYR_CONTINUE:
                break;

            case T_DECORATION:
                break;

            case T_SYM_BAR:
                break;
            case T_SYM_BLANK:
                break;
            case T_SYM_VERSE:
                break;

            case T_XFIELD:
                break;

            case T_FIELD:

                switch (abcField(s, field)) {

                case 'K': {
                    std::string keystr((char *)abcField(s, string), abcField(s, string_len));
                    // abcp has issues with parsing keys, use custom method
                    AbcKey *key = parseKey(keystr);
                    keylist.push_back(key);
                    if (header) {
                        defkey = key;
                    }
                    activeVoice->key(key);
                    header = false;
                } break;

                case 'L': {
                    int notelen = abcTICKS / (double)abcLength(s, duration);
                    if (header) {
                        abcnote = notelen;
                    }
                    activeVoice->noteLength(notelen);
                } break;

                case 'M': {
                    int num = abcMeter(s, num);
                    int den = abcMeter(s, den);
                    activeVoice->timeSignature(num, den);
                    TimeSignature *evt = new TimeSignature(num, log2(den), 24, 8);
                    // TODO: uncomment when midi files can save properly
                    //                    if (metaseq) {
                    //                        metaseq->add(*evt, activeVoice->measure());
                    //                    } else {
                    //                        activeVoice->add(*evt);
                    //                    }
                    if (metaseq) {
                        metaseq->add(*evt, activeVoice->measure());
                    } else {
                        activeVoice->add(*evt);
                    }
                    if (header) {
                        defaultsig = time::Signature(num, den);
                    }
                } break;

                case 'Q': {
                    TempoChange *evt = new TempoChange(abcTempo(s, beats));
                    // tempo: always goes on track 0
                    if (metaseq) {
                        metaseq->add(*evt, activeVoice->measure());
                    } else {
                        activeVoice->add(*evt);
                    }
                } break;

                case 'T': {
                    std::string title((char *)abcField(s, string), abcField(s, string_len));
                    TrackName *evt = new TrackName(title.c_str());
                    if (header) {
                        metaseq->add(*evt, activeVoice->measure());
                    } else {
                        activeVoice->add(*evt);
                    }
                } break;

                case 'V': {
                    std::string name((char *)abcVoice(s, id), abcVoice(s, id_len));
                    activeVoice = voicemap[name];
                    if (!activeVoice) {
                        activeVoice = new AbcVoice(abcnote, defkey, defaultsig, sp); // add key
                        voicemap[name] = activeVoice;
                        voicelist.push_back(activeVoice);
                    }
                } break;

                case 'X':
                    if (++tunecount > 1) {
                        abcerror(s, "X: cannot parse multiple tunes");
                    }

                default:
                    if (abcField(s, field)) {
                        // abcField(s, field)
                    }
                }
                fieldcount++;
                break;

            case T_INCLUDE:
                break;

            case T_INCLUDE_EOF:
                break;

            case T_UNKNOWN:
                if (metaseq && !fieldcount) {
                    abcerror(s, "unable to parse, no header field found");
                } else {
                    abcerror(s, "unable to parse");
                }
                break;

            default:
                break;
            }
        }
    } catch (std::logic_error ex) {
        abcerror(s, ex.what());
    }
    abcClose(s);
    for (AbcKey *key : keylist) {
        delete key;
    }
    // create and return list of sequences
    SequenceList *ret = new std::vector<Sequence *>();
    for (AbcVoice *voice : voicelist) {
        ret->push_back(voice->sequence());
        delete voice;
    }
    return ret;
}

void Abc::abcerror(abcScanner *s, const char *mesg) {
    std::string error("ABC notation error, line ");
    error.append(std::to_string(abcLineNum(s)));
    error.append(": ");
    error.append(mesg);
    throw std::logic_error(error);
}

abcScanner *Abc::scanner(const char *abc) {
    abcScanner *s = abcStringScanner((char *)abc);
    if (!s) {
        throw std::logic_error("failed to create ABC scanner");
    }
    return s;
}

AbcKey *Abc::parseKey(std::string &str) {
    std::regex keyex(
        "([A-Ga-g])([b#]?)\\s*([A-Za-z]*)((\\s+[_^=][a-g])*)((\\s+[a-z]+=[a-z]+)*)\\s*");
    std::smatch match;
    if (regex_match(str, match, keyex)) {
        char tonic = match.str(1)[0];
        auto astr = match.str(2);
        int accidental = 0;
        if (astr.length()) {
            accidental = astr[0] == '#' ? 1 : -1;
        }
        std::string mode = match.str(3);
        std::string extrastr = match.str(4);
        // split
        std::vector<std::string> extraaccs;
        std::regex delim("\\s+");
        std::sregex_token_iterator splitter(extrastr.begin(), extrastr.end(), delim, -1);
        std::sregex_token_iterator splitend; // default ctor = end
        for (; splitter != splitend; ++splitter) {
            extraaccs.push_back(*splitter);
        }
        std::string fields = match.str(6); // TODO: handle?
        return new AbcKey(tonic, accidental, mode, &extraaccs);
    } else if (str == "none") {
        return new AbcKey();
    } else {
        throw std::logic_error("cannot parse key");
    }
}

void AbcVoice::bar(int starts, int ends, int barsize) {
    if (starts > 1) {
        if (activeseq == repeatseq || endingnum) {
            throw std::logic_error("no nested repeats");
        }
        repeatseq->clear();
        activeseq = repeatseq; // now writing to repeat section
        repstart = mmeasure;
        mmeasure = 1.0;
    }
    if (ends > 1) {
        if (!endingnum) {
            if (activeseq != repeatseq) {
                throw std::logic_error("end repeat without start");
            }
            mainseq->add(*repeatseq, repstart);
            activeseq = mainseq;
            mmeasure = repstart + repeatseq->length();
        } else {
            endingnum = 0;
        }
        mainseq->add(*repeatseq, mmeasure);
        mmeasure += repeatseq->length();
    } else if (barsize == 1 && starts < 2 && !endingnum && mmeasure >= 2.0 &&
               mmeasure != lastbar) { // sometimes apply bar length sanity check
        double delta = abs(mmeasure - lastbar - 1.0);
        if (delta > 0.00001) { // TODO: what's a good cutoff here? less than 1 frame/tick..
            std::string mesg("bar line at ");
            float loc = mmeasure + (activeseq == repeatseq ? repstart : 0);
            mesg.append(std::to_string(loc));
            mesg.append(" denotes measure of size ");
            mesg.append(std::to_string(mmeasure - lastbar));
            throw std::logic_error(mesg);
        }
    }
    lastbar = mmeasure;
}

void AbcVoice::brokenRhythm(char type, int dots) {
    if (lastoffs.size() < 1) {
        throw std::logic_error("dotted note does not exist");
    }
    int sign = (type == '<') ? 1 : -1;
    broken = sign * dots;

    // calculate amount to modify last notes
    double dotdelta = sign * duration / 2.0;
    for (int i = 1; i < abs(broken); i++) {
        dotdelta += dotdelta / 2;
    }

    // adjust duration of last notes by moving last offs
    for (size_t i = 0; i < lastoffs.size(); i++) {
        activeseq->remove(lastoffs[i]);
        activeseq->add(*lastoffs[i]->message(), lastoffs[i]->measure() - dotdelta);
    }
    lastoffs.clear();

    // adjust measure
    mmeasure -= dotdelta;
}

void AbcVoice::chordStart() {
    ingroup = true;
    lastoffs.clear();
}

void AbcVoice::chordEnd() {
    ingroup = false;
    mmeasure += grouplen;
    grouplen = 0;
    tiednote = false;
    if (tupletspan) {
        tupletspan--;
    }
    broken = 0;
}

void AbcVoice::ending(int num) {
    if (!repstart) {
        throw std::logic_error("ending label without start repeat");
    }
    if (endingnum) {
        throw std::logic_error("multiple ending labels without intervening repeat");
    }
    endingnum = num;
    // first pass
    if (activeseq == repeatseq) {
        mainseq->add(*repeatseq, repstart);
        activeseq = mainseq;
    }
}

void AbcVoice::note(unsigned char pitch, char octave, unsigned char bending, short bend,
                    double ticks) {
    char velocity = stresspat.velocity(mmeasure - lastbar);

    // map to MIDI key
    uint8_t key = note2pitch(pitch, octave);

    // handle bends/accidentals
    switch (bending) {
    case 's':
        key -= bend / 8192;
        break;
    case 'b':
        key += bend / 8192;
        break;
    }

    // convert to measure duration in 8th note increments
    double notesPerMeasure = (mnotelen / 4.0) * msig.beatsPerMeasure();
    duration = ticks / abcTICKS / notesPerMeasure;

    // factor in tuplets like triplets etc
    if (tupletspan) {
        duration *= tupletratio;
    }

    // handle dotted notes
    if (broken != 0) {
        float delta = duration / 2;
        int sign = broken > 0 ? 1 : -1;
        for (int i = 0; i < abs(broken); i++) {
            duration += delta * sign;
            delta = delta / 2;
        }
    }

    // add NOTEON at current measure unless tied
    if (!tiednote) {
        activeseq->add(*(new NoteOn(key, velocity)), mmeasure);
    }

    // new note: clear dots + ties, decrement tuples and reset NOTEOFFs
    if (!ingroup) {
        broken = 0;
        tiednote = false;
        lastoffs.clear();
        if (tupletspan) {
            tupletspan--;
        }
    }
    lastoffs.push_back(activeseq->add(*(new NoteOff(key, 64)), mmeasure + duration));

    // measure moves forward
    if (!ingroup) {
        mmeasure += duration;
    } else {
        grouplen = std::max(grouplen, duration);
    }
}

void AbcVoice::rest(int type, double duration) {
    switch (type) {
    case 'x':
    case 'z': {
        float notesPerMeasure = (mnotelen / 4.0) * msig.beatsPerMeasure();
        double skip = duration / abcTICKS / notesPerMeasure;
        mmeasure += skip;
    } break;
    case 'Z': { // multi measure
        mmeasure += (int)duration;
        lastbar = mmeasure - 1.0;
    } break;
    default:
        break;
    }
}

void AbcVoice::tieNote() {
    // remove last NOTEOFF for tied notes
    for (size_t i = 0; i < lastoffs.size(); i++) {
        activeseq->remove(lastoffs[i]);
    }
    lastoffs.clear();
    tiednote = true;
}

void AbcVoice::tuplet(int notes, int time, int span) {
    if (!time) {
        time = 2;
    }
    tupletspan = span;
    if (!tupletspan) {
        tupletspan = notes;
    }
    tupletratio = static_cast<float>(time) / notes;
}

Sequence *AbcVoice::sequence() {
    if (activeseq != mainseq) {
        throw std::logic_error("unterminated repeat section");
    }
    return mainseq;
}

uint8_t AbcVoice::note2pitch(unsigned char note, char octave) {
    int pitch = AbcKey::pitches[note];
    return 12 + (octave * 12) + pitch + mkey->modify(note);
}

} // namespace midi
} // namespace bipscript
