/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef ABC_H
#define ABC_H

#include "midifile.h"
#include "midisequence.h"
#include "stresspattern.h"
#include "time/clock.h"

#include <map>

extern "C" {
#include "abcp.h"
}

namespace bipscript {
namespace midi {

class AbcKey {
    std::map<char, int> modmap;

  public:
    static std::map<char, int> pitches;
    AbcKey(char tonic, int accidental, std::string mode, std::vector<std::string> *extraaccs);
    AbcKey() : AbcKey('C', 0, "maj", nullptr) {}
    int modify(char note) { return modmap[note]; }

  private:
    int pitchdiff(int pitch, int natural);
};

class AbcVoice {
    Sequence *mainseq = new Sequence();   // main sequence
    Sequence *repeatseq = new Sequence(); // for repeated bars
    Sequence *activeseq = mainseq;        // active sequence, main by default
    AbcKey *mkey;
    time::Signature msig;
    StressPattern &stresspat;
    float mmeasure = 1.0;
    float lastbar = 1.0;
    std::vector<MessageEvent *> lastoffs; // last NOTEOFF messages
    float duration = 1.0;                 // last note duration
    float repstart = 0;                   // start of repeated section
    int mnotelen = 8;                     // default note is an 8th note
    bool ingroup = false;                 // notes are inside a grouping
    float grouplen = 0.0;                 // group length
    int endingnum = 0;                    // repeat ending number
    int broken = 0;                       // dotted notes
    float tupletratio = 1.0;              // for tuplets: triplets etc.
    int tupletspan = 0;                   // tuplet note remaining
    bool tiednote = false;                // current note tied to next

  public:
    AbcVoice(int notelen, AbcKey *key, time::Signature sig, StressPattern &sp)
        : mkey(key), msig(sig), stresspat(sp), mnotelen(notelen) {}
    ~AbcVoice() { delete repeatseq; }
    // setters
    void noteLength(int note) { mnotelen = note; }
    void key(AbcKey *key) { mkey = key; }
    void timeSignature(int num, int den) { msig = time::Signature(num, den); }
    // parser methods
    MessageEvent *add(Message &mesg) { return activeseq->add(mesg, mmeasure); }
    void bar(int starts, int ends, int size);
    void brokenRhythm(char type, int dots);
    void chordStart();
    void chordEnd();
    void ending(int num);
    void note(unsigned char pitch, char octave, unsigned char bending, short bend, double ticks);
    void rest(int type, double duration);
    void tieNote();
    void tuplet(int notes, int time, int span);
    // getters
    Sequence *sequence();
    float measure() { return mmeasure; }

  private:
    uint8_t note2pitch(unsigned char note, char octive);
};

typedef std::vector<midi::Sequence *> SequenceList;

class Abc {

  public:
    static std::map<int, std::string> notename;
    Sequence *sequence(const char *abc);
    File *file(const char *abc);
    std::string abc(midi::Note &note);

  private:
    void abcerror(abcScanner *s, const char *mesg);
    SequenceList *parse(abcScanner *s, Sequence *meta);
    abcScanner *scanner(const char *abc);
    AbcKey *parseKey(std::string &str);
    uint gcd(uint x, uint y);
};

} // namespace midi
} // namespace bipscript

#endif // ABC_H
