﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDIFILE_H
#define MIDIFILE_H

#include <cmath>
#include <stdint.h>
#include <string>

#include "io.h"
#include "midisequence.h"
#include "scriptarray.h"

namespace bipscript {
namespace midi {

class MetaEvent : public Message {
  public:
    uint8_t command() override { return 0xFF; }
    bool meta() override { return true; }
    virtual uint8_t metatype() const = 0;
};

class KeySignature : public MetaEvent {
    const uint8_t mcount;
    const uint8_t mminor;

  public:
    static const uint8_t METATYPE = 0x59;
    KeySignature(uint8_t count, uint8_t minor) : mcount(count), mminor(minor) {}
    uint8_t metatype() const override { return METATYPE; }
    int priority() override { return Priority::KEYSIG; }
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = command();
        buffer[1] = METATYPE;
        buffer[2] = 2; // 2 bytes data length
        buffer[3] = mcount;
        buffer[4] = mminor;
    }
    uint8_t count() { return mcount; }
    bool minor() { return mminor > 0; }
    uint16_t size() const override { return 5; }
};

class TempoChange : public MetaEvent {
    const float mtempo;

  public:
    static const uint8_t METATYPE = 0x51;
    TempoChange(uint32_t tempo) : mtempo(tempo) {}
    float tempo() { return mtempo; }
    uint8_t metatype() const override { return METATYPE; }
    int priority() override { return Priority::TEMPO; }
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = command();
        buffer[1] = METATYPE;
        buffer[2] = 3; // 3 bytes data length
        uint32_t value = 60000000L / mtempo;
        buffer[3] = (value >> 16) & 0xff;
        buffer[4] = (value >> 8) & 0xff;
        buffer[5] = (value >> 0) & 0xff;
    }
    uint16_t size() const override { return 6; }
};

class TimeSignature : public MetaEvent {
    uint8_t mnumerator;
    uint8_t mdenompower; // denominator as power of 2
    uint8_t mclockspq;   // MIDI clocks per quarter note
    uint8_t m32pq;       // 32nd notes per MIDI quarter
  public:
    static const uint8_t METATYPE = 0x58;
    TimeSignature(uint8_t top, uint8_t bottom, uint8_t clockspq, uint8_t i32pq)
        : mnumerator(top), mdenompower(bottom), mclockspq(clockspq), m32pq(i32pq) {}
    uint8_t metatype() const override { return METATYPE; }
    int priority() override { return Priority::TIMESIG; }
    uint8_t numerator() { return mnumerator; }
    uint8_t denompower() { return mdenompower; }
    uint8_t denominator() { return pow(2, mdenompower); }
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = command();
        buffer[1] = METATYPE;
        buffer[2] = 4; // 4 bytes data length
        buffer[3] = mnumerator;
        buffer[4] = mdenompower;
        buffer[5] = mclockspq;
        buffer[6] = m32pq;
    }
    uint16_t size() const override { return 7; }
};

class TextMetaEvent : public MetaEvent {
    char *mtext;
    uint32_t msize;

  public:
    static const uint32_t HSIZE = 3;
    TextMetaEvent(const char *text) : TextMetaEvent(text, strlen(text)) {}
    TextMetaEvent(const char *text, uint32_t size) : msize(size) {
        mtext = new char[size + 1];
        strncpy(mtext, text, size);
        mtext[size] = '\0';
    }
    ~TextMetaEvent() { delete[] mtext; }
    int priority() override { return Priority::METATEXT; }
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = command();
        buffer[1] = metatype();
        buffer[2] = msize;
        memcpy(&buffer[3], mtext, msize);
    }
    char *text() { return mtext; }
    uint16_t size() const override { return 3 + msize; }
};

class Text : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x01;
    uint8_t metatype() const override { return METATYPE; }
    Text(const char *text) : TextMetaEvent(text) {}
    Text(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class Copyright : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x02;
    uint8_t metatype() const override { return METATYPE; }
    Copyright(const char *text) : TextMetaEvent(text) {}
    Copyright(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class TrackName : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x03;
    uint8_t metatype() const override { return METATYPE; }
    TrackName(const char *text) : TextMetaEvent(text) {}
    TrackName(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class Instrument : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x04;
    uint8_t metatype() const override { return METATYPE; }
    Instrument(const char *text) : TextMetaEvent(text) {}
    Instrument(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class Lyric : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x05;
    uint8_t metatype() const override { return METATYPE; }
    Lyric(const char *text) : TextMetaEvent(text) {}
    Lyric(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class Marker : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x06;
    uint8_t metatype() const override { return METATYPE; }
    Marker(const char *text) : TextMetaEvent(text) {}
    Marker(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class CuePoint : public TextMetaEvent {
  public:
    static const uint8_t METATYPE = 0x07;
    uint8_t metatype() const override { return METATYPE; }
    CuePoint(const char *text) : TextMetaEvent(text) {}
    CuePoint(const char *text, uint32_t size) : TextMetaEvent(text, size) {}
};

class File {
    uint32_t mppq;
    uint8_t mtype;
    ScriptArray<Sequence> *mtrack;

  public:
    File();
    File(std::vector<Sequence *> seqs); // for cpp
    File(std::string path);
    File(io::File &file) : File(file.path()) {}
    ~File();
    uint32_t ppq() { return mppq; }
    void ppq(uint32_t ppq) { mppq = ppq; }
    uint8_t type() { return mtype; }
    void type(uint8_t type) { mtype = type; }
    uint32_t tracks() { return mtrack->len(); }
    HSQOBJECT track() { return mtrack->hsqobject(); }
    void track(HSQOBJECT track);
    void save(std::string path);
    void save(io::File &file) { save(file.path()); }
};

struct TpmMapEntry {
    float tpm;
    union {
        uint32_t tick;
        float measure;
    };
};

} // namespace midi
} // namespace bipscript

#endif // MIDIFILE_H
