/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "midisequence.h"
#include "bindinterface.h"
#include "bindmidi.h"
#include "bindutil.h"
#include "midipattern.h"
#include "module_midi.h"

namespace bipscript {
namespace midi {

MessageEvent::~MessageEvent() {
    if (mmessage->delref() == 0) {
        delete mmessage;
    }
}

bool MessageEvent::operator<(MessageEvent &other) {
    if (mmeasure == other.mmeasure) {
        return mmessage->priority() < other.mmessage->priority();
    }
    return mmeasure < other.mmeasure;
}

SequenceFilter::SequenceFilter(ScriptFunction &function) : ScriptFunction(function) {
    if (function.numargs() < 2 || function.numargs() > 3) {
        throw std::logic_error("Sequence filter function should take one or two arguments");
    }
}

bool SequenceFilter::execute(size_t index, MessageEvent *event) {
    // push function pointer as closure
    sq_pushobject(vm, function());
    // push associated closure root as "this" pointer
    sq_getclosureroot(vm, -1);
    // add other parameters
    if (numargs() == 2) { // single parameter: message event
        BindUtil::PushRef(vm, event, binding::MessageEventObject);
    } else { // two parameters: index + message event
        sq_pushinteger(vm, index);
        BindUtil::PushRef(vm, event, binding::MessageEventObject);
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, numargs(), SQTrue, SQTrue))) {
        throw std::logic_error("Sequence filter method failed\n");
    }
    SQBool retval;
    if (SQ_FAILED(sq_getbool(vm, -1, &retval))) {
        throw std::logic_error("Sequence filter method did not evaluate to boolean");
    }
    // pop retval + closure
    sq_pop(vm, 2);
    return retval;
}

Sequence::Sequence(Sequence &copy) {
    for (uint32_t i = 0; i < copy.eventlist.size(); i++) {
        eventlist.push_back(copy.eventlist[i]->addref());
    }
}

Sequence::~Sequence() {
    for (uint32_t i = 0; i < eventlist.size(); i++) {
        if (eventlist[i]->delref() == 0) {
            delete eventlist[i];
        }
    }
}

MessageEvent *Sequence::add(Message &mesg, float measure) {
    if (measure < 1.0) {
        throw std::logic_error("Cannot add a sequence event before measure 1.0");
    }
    MessageEvent *evt = (new MessageEvent(mesg, measure))->addref();
    eventlist.insert(std::upper_bound(eventlist.begin(), eventlist.end(), evt, MessageEvent::cmp),
                     evt);
    return evt;
}

void Sequence::add(Note &note, float measure) {
    add(*new NoteOn(note.pitch(), note.velocity(), note.channel()), measure);
    add(*new NoteOff(note.pitch(), 64, note.channel()), measure + note.duration());
}

void Sequence::add(Pattern &pattern, float measure) {
    if (measure < 1.0) {
        throw std::logic_error("Cannot add a sequence event before measure 1.0");
    }
    for (unsigned int i = 0; i < pattern.len(); i++) {
        const NoteEvent *pn = pattern[i];
        const Note &note = *pn->note();
        const float newmeasure = measure + pn->measure() - 1;

        NoteOn *noteOn = new NoteOn(note.pitch(), note.velocity(), note.channel());
        MessageEvent *evt = (new MessageEvent(*noteOn, newmeasure))->addref();
        eventlist.push_back(evt);

        NoteOff *noteOff = new NoteOff(note.pitch(), 64, note.channel());
        MessageEvent *offevt = (new MessageEvent(*noteOff, newmeasure + note.duration()))->addref();
        eventlist.push_back(offevt);
    }
    std::sort(eventlist.begin(), eventlist.end(), MessageEvent::cmp);
}

void Sequence::add(Sequence &sequence, float measure) {
    if (measure < 1.0) {
        throw std::logic_error("Cannot add a sequence event before measure 1.0");
    }
    for (unsigned int i = 0; i < sequence.len(); i++) {
        MessageEvent *evt = sequence[i];
        Message *mesg = evt->message();
        if (measure != 1.0) {
            const float newmeasure = measure + evt->measure() - 1;
            MessageEvent *newevt = (new MessageEvent(*mesg, newmeasure))->addref();
            eventlist.push_back(newevt);
        } else {
            // time is the same: no need for a new event here
            eventlist.push_back(evt->addref());
        }
    }
    std::sort(eventlist.begin(), eventlist.end(), MessageEvent::cmp);
}

void Sequence::add(MessageEvent *event) {
    eventlist.push_back(event->addref());
    std::sort(eventlist.begin(), eventlist.end(), MessageEvent::cmp);
}

void Sequence::clear() {
    for (uint32_t i = 0; i < eventlist.size(); i++) {
        if (eventlist[i]->delref() == 0) {
            delete eventlist[i];
        }
    }
    eventlist.clear();
}

Sequence *Sequence::filter(ScriptFunction func) {
    Sequence *ret = new Sequence();
    SequenceFilter filter(func);
    for (uint32_t i = 0; i < eventlist.size(); i++) {
        MessageEvent *evt = eventlist[i];
        if (filter.execute(i, evt)) {
            ret->add(*evt->message(), evt->measure());
        }
    }
    return ret;
}

Pattern *Sequence::pattern() {
    Pattern *ret = new Pattern();
    MessageEvent *start[128] = {nullptr};
    for (uint32_t i = 0; i < eventlist.size(); i++) {
        Message *mesg = eventlist[i]->message();
        if (mesg->command() == NoteOn::COMMAND) {
            NoteOn *note = static_cast<NoteOn *>(mesg);
            if (note->velocity() > 0) {
                start[note->pitch()] = eventlist[i];
            } else { // treat velocity zero as note off
                MessageEvent *evt = start[note->pitch()];
                if (evt) {
                    NoteOn *on = static_cast<NoteOn *>(evt->message());
                    float duration = eventlist[i]->measure() - evt->measure();
                    Note *note = new Note(on->pitch(), on->velocity(), duration);
                    ret->add(note, evt->measure());
                    start[note->pitch()] = nullptr;
                }
            }
        } else if (mesg->command() == NoteOff::COMMAND) {
            NoteOff *off = static_cast<NoteOff *>(mesg);
            MessageEvent *evt = start[off->pitch()];
            if (evt) {
                NoteOn *on = static_cast<NoteOn *>(evt->message());
                float duration = eventlist[i]->measure() - evt->measure();
                Note *note = new Note(on->pitch(), on->velocity(), duration);
                ret->add(note, evt->measure());
                start[off->pitch()] = nullptr;
            }
        }
    }
    return ret;
}

} // namespace midi
} // namespace bipscript
