/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef OSCINPUT_H
#define OSCINPUT_H

#include "bindosc.h"
#include "bindutil.h"
#include "closure.h"
#include "collector.h"
#include "lo/lo.h"
#include "net.h"
#include "objectcache.h"
#include "oscmessage.h"

#include <atomic>
#include <stdexcept>
#include <string>

namespace bipscript {
namespace osc {

class Input : public Listable<Input> {
    lo_server_thread st;
    std::atomic<ScriptFunction *> onReceiveHandler;

  public:
    Input();                            // UDP
    Input(int port, int protocol);      // UDP/TCP
    Input(int port, const char *group); // UDP multicast
    Input(const char *path);            // UNIX domain
    ~Input() { delete onReceiveHandler.load(); }
    int handle(const char *path, const char *types, lo_arg **argv, int argc, void *data);
    int port() { return lo_server_thread_get_port(st); }
    void onMessage(ScriptFunction &function);
    void onReceive(ScriptFunction &function) {
        fprintf(stderr, "onReceive method is deprecated, please use onMessage\n");
        onMessage(function);
    }
    void join();
    void restartScript();
};

class OnReceiveClosure;

class ClosureCollector {
    Collector<OnReceiveClosure> eventCollector;

  public:
    static ClosureCollector &instance() {
        static ClosureCollector instance;
        return instance;
    }
    void init() { CollectorThread::instance().addCollector(&eventCollector); }
    void recycle(OnReceiveClosure *evt) { eventCollector.recycle(evt); }
};

class OnReceiveClosure : public Closure {
    Message *message;

  protected:
    void addParameters() { BindUtil::PushCopy(vm, *message, binding::MessageObject); }

  public:
    OnReceiveClosure(ScriptFunction function, Message *message)
        : Closure(function, nullptr), message(message) {}
    ~OnReceiveClosure() { delete message; }
    void finished() { ClosureCollector::instance().recycle(this); }
};

class InputFactory : public ActiveCache<Input> {
    uint32_t minstcount; // instance count
    InputFactory() : minstcount(0) { ClosureCollector::instance().init(); }

  public:
    static InputFactory &instance() {
        static InputFactory instance;
        return instance;
    }
    ~InputFactory() { removeAll(); }
    Input *newOscInput();                                                // UDP
    Input *newOscInput(int port, int protocol);                          // UDP/TCP
    Input *newOscInput(int port) { return newOscInput(port, net::UDP); } // UDP
    Input *newOscInput(int port, const char *group);                     // UDP multicast
    Input *newOscInput(const char *path);                                // UNIX domain
    void newObject(Input *) override {}
    void removeObject(Input *input) override {
        input->join();
        delete input;
    }
    void restartCache() override { minstcount = 0; }
};

} // namespace osc
} // namespace bipscript

#endif // OSCINPUT_H
