/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef REGEXTYPE_H
#define REGEXTYPE_H

#include <regex>

#include "bindutil.h"

using namespace std;

namespace bipscript {

namespace binding {
extern HSQOBJECT regexObject;
}

class Regex {
    std::regex exp;
    string mpattern;

  public:
    static void bindType(HSQUIRRELVM vm) {
        BindUtil binding(vm);
        binding.startClass("regex", 5, binding::regexObject);
        binding.addMethod("constructor", 11, Ctor);
        binding.addMethod("_cloned", 7, BindUtil::Clone<Regex>);
        binding.addMethod("_tostring", 9, ToString);
        binding.addMethod("match", 5, Match);
        binding.addMethod("find", 4, Find);
        binding.endClass();
    }
    static Regex *getRegex(HSQUIRRELVM vm, SQInteger index) {
        return BindUtil::getObject<Regex>(vm, index, binding::regexObject);
    }
    std::regex &to_regex() { return exp; }
    Regex(const char *pattern) : exp(pattern), mpattern("/") {
        mpattern.append(pattern);
        mpattern.append("/");
    }
    void match(HSQUIRRELVM vm, const char *string) { // for squirrel
        cmatch cm;
        regex_match(string, cm, exp);
        size_t results = cm.size();
        if (results) {
            sq_newarray(vm, results);
            for (uint32_t i = 0; i < results; i++) {
                sq_pushinteger(vm, i);
                sq_pushstring(vm, _SC(cm[i].str().c_str()), -1);
                sq_set(vm, -3);
            }
        } else {
            sq_pushnull(vm);
        }
    }
    bool matches(const char *string) { // for c++
        cmatch cm;
        return regex_match(string, cm, exp);
    }
    void find(HSQUIRRELVM vm, const char *target) {
        std::string str(target);
        smatch sm;
        cmatch cm;
        sq_newarray(vm, 0);
        while (regex_search(str, sm, exp)) {
            sq_newtable(vm);
            sq_pushstring(vm, "index", 5);
            sq_pushinteger(vm, sm.position(0));
            sq_newslot(vm, -3, false);
            sq_pushstring(vm, "match", 5);
            sq_pushstring(vm, _SC(sm[0].str().c_str()), -1);
            sq_newslot(vm, -3, false);
            sq_arrayappend(vm, -2); // append array value
            str = sm.suffix().str();
        }
    }

  private:
    // binding methods
#define BIND_CLASS_NAME "regex"
#define BIND_METHOD_NAME "constructor"
    static SQInteger Ctor(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        BIND_GET_CHECK_STRING(pattern, 1)
        try {
            Regex *obj = new Regex(pattern);
            // return pointer to new object
            sq_setinstanceup(vm, 1, (SQUserPointer *)obj);
            sq_setreleasehook(vm, 1, BindUtil::Dtor<Regex>);
            return 1;
        } catch (std::exception const &e) {
            return sq_throwerror(vm, e.what());
        }
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "_tostring"
    static SQInteger ToString(HSQUIRRELVM vm) {
        BIND_GET_THIS(Regex, binding::regexObject)
        sq_pushstring(vm, obj->mpattern.c_str(), -1);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "match"
    static SQInteger Match(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        BIND_GET_THIS(Regex, binding::regexObject)
        BIND_GET_CHECK_STRING(string, 1)
        BIND_WRAP_TRY(obj->match(vm, string))
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "find"
    static SQInteger Find(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        BIND_GET_THIS(Regex, binding::regexObject)
        BIND_GET_CHECK_STRING(string, 1)
        BIND_WRAP_TRY(obj->find(vm, string))
        return 1;
    }
#undef BIND_METHOD_NAME
#undef BIND_CLASS_NAME
};

} // namespace bipscript

#endif // REGEXTYPE_H
