/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "scripthost.h"

#include "sqstdaux.h"
#include "sqstdio.h"

#include "bindings.h"
#include "buffertype.h"
#include "clock.h"
#include "closure.h"
#include "connectionop.h"
#include "extension.h"
#include "iterator.h"
#include "methodqueue.h"
#include "printfunc.h"
#include "regextype.h"
#include "scriptcollector.h"
#include "statusserver.h"
#include "stdpipes.h"
#include "stringtype.h"
#include "tabletype.h"
#include "virtualmachine.h"

#include <boost/filesystem.hpp>
#include <iostream>
#include <pwd.h>

namespace bipscript {

HSQOBJECT binding::bufferObject;
HSQOBJECT binding::iteratorObject;
HSQOBJECT binding::readerObject;
HSQOBJECT binding::regexObject;
HSQOBJECT binding::writerObject;

SQInteger bip_import_function(HSQUIRRELVM vm) { return ScriptHost::instance().imports(vm); }

void ScriptHost::resetProcessorThreads(bool restart) {
    if (StatusServer::instance().active()) {
        StatusServer::instance().scriptReset(restart);
    }
    ProcessorEngine *thread = engines.first();
    // process thread reset
    if (audioengine) {
        audioengine->restart();
    }
    while (thread) {
        thread->restart();
        thread = engines.next(thread);
    }
    // script thread reset
    for (uint16_t i = 0; i < objectCacheCount; i++) {
        objectCacheList[i]->reset(restart);
    }
    // reset default clock
    time::Clock::defaultClock(nullptr);
    // wait for process threads
    while (true) {
        bool restarting = false;
        ProcessorEngine *thread = engines.first();
        while (thread) {
            if (thread->restarting()) {
                restarting = true;
            }
            thread = engines.next(thread);
        }
        MethodQueue &methodQueue = MethodQueue::instance();
        methodQueue.scriptReset();
        if (!restarting) {
            break;
        }
        // sleep
        struct timespec req = {0, 10000};
        while (nanosleep(&req, &req) == -1) {
            continue;
        }
    }
    // wait for audio engine last as it holds the clocks
    if (audioengine) {
        while (audioengine->restarting()) {
            struct timespec req = {0, 10000};
            while (nanosleep(&req, &req) == -1) {
                continue;
            }
        }
    }
}

void ScriptHost::executeClosures() {
    MethodQueue &methodQueue = MethodQueue::instance();
    Closure *closure = methodQueue.next();
    while (closure && !(reloadRequested || exitRequested)) {
        closure->execute();
        methodQueue.finished(closure);
        closure = methodQueue.next();
    }
}

bool ScriptHost::eventLoop() {
    while (!exitRequested) {
        // run any dispatched methods
        executeClosures();

        // run any collectors
        ScriptCollector::instance().collect();

        // restart main script if requested by Script.reload
        if (reloadRequested) {
            reloadRequested = false;
            runningFlag.store(true);
            imported.clear();
            resetProcessorThreads(true);
            return true;
        }

        // check for active objects, exit if none
        bool activeObjects = audioengine ? audioengine->active() : false;
        ProcessorEngine *thread = engines.first();
        while (thread && !activeObjects) {
            if (thread->active()) {
                activeObjects = true;
            }
            thread = engines.next(thread);
        }
        // TODO: extensions request processor threads instead
        activeObjects = ExtensionManager::instance().eventLoop() || activeObjects;
        if (!activeObjects) {
            // run any remaining closures
            executeClosures();
            // reset all processor threads
            resetProcessorThreads(false);
            // return false, script is complete
            return false;
        }

        // sleep
        struct timespec req = {0, 25000};
        while (nanosleep(&req, &req) == -1) {
            continue;
        }
    }
    resetProcessorThreads(false);
    return false;
}

ScriptHost::ScriptHost() : audioengine(0), reloadRequested(false), exitCode(0), runningFlag(true) {
    struct passwd *userinfo = getpwuid(getuid());
    const char *home;
    if (userinfo) {
        home = userinfo->pw_dir;
    } else {
        home = getenv("HOME");
    }
    mhomedir = home;
}

SQInteger ScriptHost::imports(HSQUIRRELVM vm) {
    SQInteger numargs = sq_gettop(vm);
    sq_push(vm, 1); // repush the this
    for (int i = 1; i < numargs; i += 2) {
        const SQChar *token;
        SQInteger local;
        sq_getstring(vm, i + 1, &token);
        sq_getinteger(vm, i + 2, &local);
        // printf("got import: %s local=%lld\n", token, local);
        try {
            import(vm, token, local);
        } catch (std::exception const &e) {
            return sq_throwerror(vm, e.what());
        }
    }
    return 1;
}

void ScriptHost::import(HSQUIRRELVM vm, const char *token, bool local) {
    if (local) {
        // strip dots
        regex re("\\.\\.");
        std::string impath = regex_replace(token, re, "");
        // add extension if none
        string ext = bfs::path(impath).extension().string();
        if (ext.length() == 0) {
            impath = impath + ".bip";
        }
        // resolve import
        bfs::path resolved = resolveImport(impath);
        // file could not be found
        if (!bfs::exists(resolved)) {
            std::string mesg("Import file could not be found: ");
            mesg.append(impath);
            throw std::logic_error(mesg);
        }
        if (bfs::is_directory(resolved)) {
            std::string mesg("Import file could not be read: ");
            mesg.append(impath);
            throw std::logic_error(mesg);
        }
        // convert to canonical form
        bfs::path cpath = bfs::canonical(resolved);
        const char *canonical = cpath.c_str();
        // import only if not already imported
        std::set<std::string>::iterator it = imported.find(canonical);
        if (it == imported.end()) {
            string parent = currentfile;
            currentfile = canonical;
            SQBool printerror = SQFalse;
            if (SQ_FAILED(sqstd_dofile(vm, canonical, SQTrue, printerror))) {
                currentfile = parent;
                std::string mesg = "reading imported file failed";
                throw std::logic_error(mesg);
            }
            imported.insert(canonical);
            currentfile = parent;
        }
    } else {
        ExtensionManager::instance().import(vm, token);
    }
}

boost::filesystem::path ScriptHost::resolveImport(std::string &filepath) {
    if (filepath[0] != '/') {
        // relative import
        bfs::path currentdir = bfs::path(currentfile).parent_path();
        return bfs::absolute(filepath, currentdir);
    } else {
        // absolute "root" import
        return resolveRootImport(filepath);
    }
}

boost::filesystem::path ScriptHost::resolveRootImport(std::string &filepath) {
    string relative = filepath.substr(1);
    // try with root as main script location
    bfs::path scriptloc = bfs::absolute(relative, mscriptroot);
    if (!bfs::exists(scriptloc)) {
        // look for .biproot file
        string biproot(".biproot");
        bfs::path dir = mfilepath.parent_path();
        while (bfs::exists(dir)) {
            if (bfs::exists(bfs::absolute(biproot, dir))) {
                scriptloc = bfs::absolute(relative, dir);
                if (bfs::exists(scriptloc)) {
                    return scriptloc;
                }
            }
            dir = dir.parent_path();
        }
        // local has failed, try with user home library
        bfs::path homelib = bfs::path(mhomedir + "/.bip/lib/");
        scriptloc = bfs::absolute(relative, homelib);
    }
    return scriptloc;
}

io::File *ScriptHost::resolve(string filepath) {
    SQStackInfos stackinfo;
    if (SQ_SUCCEEDED(sq_stackinfos(vm, 1, &stackinfo))) {
        const SQChar *srcfile = stackinfo.source;
        if (srcfile) {
            bfs::path currentdir = bfs::path(srcfile).parent_path();
            bfs::path absolute = bfs::absolute(filepath, currentdir);
            return new io::File(absolute.c_str());
        }
    }
    throw std::logic_error("Failed to resolve path: " + filepath);
}

void ScriptHost::shutdown() {
    resetProcessorThreads(false);
    ScriptCollector::instance().collect();
    sq_close(vm);
}

void ScriptHost::init(bfs::path &filepath) {
    // init squirrel
    vm = sq_open(1024);
    VirtualMachine::instance().vm(vm);
    // set print function
    sqstd_seterrorhandlers(vm);
    sq_setprintfunc(vm, PrintFunc::Print, PrintFunc::Print);
    bip_setimportfunc(vm, bip_import_function);
    bip_setdconnectfunc(vm, bip_dconnect_operator);
    bip_setmconnectfunc(vm, bip_mconnect_operator);
    // standard library modules
    sq_pushroottable(vm);
    sq_getstackobj(vm, -1, &apiTable);
    sq_addref(vm, &apiTable);
    // standard library modules
    sqstd_register_iolib(vm);
    // local types
    Buffer::bindType(vm);
    Iterator::bindType(vm);
    Reader::bindType(vm);
    Regex::bindType(vm);
    String::bindType(vm);
    Table::bindType(vm);
    PrintFunc::bindType(vm);
    Writer::bindType(vm);
    // local objects
    StdIn::bindObject(vm);
    StdOut::bindObject(vm);
    StdErr::bindObject(vm);
    // API objects
    binding::bindAll(vm);
    // pop root table
    sq_pop(vm, 1);
    // store filepath
    mfilepath = filepath;
    mscriptroot = mfilepath;
    mscriptroot.remove_filename();
    mfilename = filepath.filename().c_str();
}

uint8_t ScriptHost::run() {
    bool rerun = true;
    while (rerun) {
        // push original api table and clone
        sq_pushobject(vm, apiTable);
        sq_clone(vm, 1);
        // set clone as new root table (and pop)
        sq_setroottable(vm);
        // pop original api table and push clone
        sq_pop(vm, 1);
        sq_pushroottable(vm);
        // do file
        currentfile = mfilepath.lexically_normal().string();
        if (!SQ_SUCCEEDED(sqstd_dofile(vm, currentfile.c_str(), 0, SQTrue))) {
            const SQChar *error;
            sq_getlasterror(vm);
            if (SQ_SUCCEEDED(sq_getstring(vm, -1, &error))) {
                std::cerr << error << std::endl;
            }
            sq_close(vm);
            return 1;
        }

        // pop clone root table
        sq_pop(vm, 1);

        // notify object caches script complete
        runningFlag.store(false);
        for (uint16_t i = 0; i < objectCacheCount; i++) {
            objectCacheList[i]->scriptComplete();
        }

        // wait for restart
        rerun = eventLoop();
    }

    // collectors
    ScriptCollector::instance().collect();

    // shut down squirrel
    sq_close(vm);
    return exitCode;
}

} // namespace bipscript
