/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRINGTYPE_H
#define STRINGTYPE_H

#include "regextype.h"

#include "squirrel.h"
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/regex.hpp>
#include <boost/foreach.hpp>
#include <boost/tokenizer.hpp>

namespace bipscript {

// adds methods to existing string class
class String {
    std::regex exp;

  public:
    static void bindType(HSQUIRRELVM vm) {
        BindUtil binding(vm);
        // pull up existing delegate table
        sq_getdefaultdelegate(vm, OT_STRING);
        // add methods
        binding.addMethod("endsWith", 8, EndsWith);
        binding.addMethod("find", 4, Find);
        binding.addMethod("ltrim", 5, Ltrim);
        binding.addMethod("matches", 7, Matches);
        binding.addMethod("replace", 7, Replace);
        binding.addMethod("rtrim", 5, Rtrim);
        binding.addMethod("split", 5, Split);
        binding.addMethod("startsWith", 10, StartsWith);
        binding.addMethod("trim", 4, Trim);
        // pop delegate table
        sq_poptop(vm);
    }

  private:
#define BIND_CLASS_NAME "string"
#define BIND_METHOD_NAME "endsWith"
    static SQInteger EndsWith(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        SQInteger thissize = sq_getsize(vm, 1);
        BIND_GET_CHECK_STRING(str, 1)
        SQInteger strlen = sq_getsize(vm, 2);
        bool ret = (thissize >= strlen) && !strncmp(&thisstr[thissize - strlen], str, strlen);
        sq_pushbool(vm, ret);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "matches"
    static SQInteger Matches(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        SQBool ret; // return value
        SQObjectType overrideType = sq_gettype(vm, 2);
        if (overrideType == OT_STRING) {
            const SQChar *pattern;
            sq_getstring(vm, 2, &pattern);
            BIND_WRAP_TRY(ret = regex_match(thisstr, std::regex(pattern)))
        } else if (Regex *regex = Regex::getRegex(vm, 2)) {
            BIND_WRAP_TRY(ret = regex_match(thisstr, regex->to_regex()))
        } else {
            return sq_throwerror(vm, "argument 1 is not of type {string, regex}");
        }

        // push return value
        sq_pushbool(vm, ret);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "find"
    static SQInteger Find(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_MIN(1)
        BIND_NARGS_CHECK_MAX(2)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        size_t pos = 0;
        if (numargs == 3) {
            BIND_GET_CHECK_INTEGER(start, 2)
            size_t thislen = strlen(thisstr);
            if (start < 0) {
                if (static_cast<long>(thislen) + start < 0) {
                    return sq_throwerror(vm, "string find argument 2 start is too small");
                }
                pos = thislen + start;
            } else {
                if (static_cast<size_t>(start) >= thislen) {
                    return sq_throwerror(vm, "string find argument 2 start is too large");
                }
                pos = start;
            }
        }
        // return value
        SQInteger ret = -1;
        SQObjectType overrideType = sq_gettype(vm, 2);
        if (overrideType == OT_STRING) {
            const SQChar *pattern;
            sq_getstring(vm, 2, &pattern);
            const char *found = scstrstr(&thisstr[pos], pattern);
            if (found) {
                ret = found - thisstr;
            }
        } else if (Regex *regex = Regex::getRegex(vm, 2)) {
            std::cmatch cm;
            bool found;
            BIND_WRAP_TRY(found = regex_search(&thisstr[pos], cm, regex->to_regex()))
            if (found) {
                ret = cm.position(0) + pos;
            }
        } else {
            return sq_throwerror(vm, "string find argument 1 is not of type {string, regex}");
        }
        // push return value
        sq_pushinteger(vm, ret);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "replace"
    static SQInteger Replace(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_MIN(2)
        BIND_NARGS_CHECK_MAX(3)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        BIND_GET_CHECK_STRING(replace, 2) // replacement always a string
        bool all = true;
        if (BIND_NARGS == 3) { // optional arg global
            BIND_GET_CHECK_BOOL(global, 3)
            all = global;
        }
        // return value
        string ret(thisstr);
        SQObjectType overrideType = sq_gettype(vm, 2);
        if (overrideType == OT_STRING) {
            const SQChar *pattern;
            sq_getstring(vm, 2, &pattern);
            if (all) {
                boost::replace_all(ret, pattern, replace);
            } else {
                boost::replace_first(ret, pattern, replace);
            }
        } else if (Regex *regex = Regex::getRegex(vm, 2)) {
            if (all) {
                ret = regex_replace(ret, regex->to_regex(), replace);
            } else {
                ret = regex_replace(ret, regex->to_regex(), replace,
                                    regex_constants::format_first_only);
            }
        } else {
            return sq_throwerror(vm, "string replace argument 1 is not of type {string, regex}");
        }
        // push return value
        sq_pushstring(vm, ret.c_str(), -1);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "split"
    static SQInteger Split(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        sq_newarray(vm, 0); // return value
        SQObjectType overrideType = sq_gettype(vm, 2);
        if (overrideType == OT_INTEGER) {
            SQInteger idelimiter;
            sq_getinteger(vm, 2, &idelimiter);
            SQChar delimiter[2];
            delimiter[0] = static_cast<SQChar>(idelimiter);
            delimiter[1] = '\0';
            std::string str(thisstr);
            boost::char_separator<char> sep(delimiter);
            boost::tokenizer<boost::char_separator<char>> tokens(str, sep);
            boost::tokenizer<boost::char_separator<char>>::iterator it;
            for (it = tokens.begin(); it != tokens.end(); it++) {
                sq_pushstring(vm, (*it).c_str(), -1);
                sq_arrayappend(vm, -2);
            }
        } else if (overrideType == OT_STRING) {
            const SQChar *pattern;
            sq_getstring(vm, 2, &pattern);
            std::string str(thisstr);
            std::list<std::string> results;
            boost::iter_split(results, str, boost::first_finder(pattern));
            BOOST_FOREACH (std::string token, results) {
                sq_pushstring(vm, token.c_str(), token.length());
                sq_arrayappend(vm, -2);
            }
        } else if (Regex *regex = Regex::getRegex(vm, 2)) {
            std::string s(thisstr);
            std::regex_token_iterator<std::string::iterator> it(s.begin(), s.end(),
                                                                regex->to_regex(), -1);
            std::regex_token_iterator<std::string::iterator> rend;
            while (it != rend) {
                sq_pushstring(vm, (*it).str().c_str(), -1);
                sq_arrayappend(vm, -2);
                ++it;
            }
        } else {
            return sq_throwerror(vm, "argument 1 is not of type {string, regex}");
        }
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "startsWith"
    static SQInteger StartsWith(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(1)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        SQInteger thissize = sq_getsize(vm, 1);
        BIND_GET_CHECK_STRING(str, 1)
        SQInteger strlen = sq_getsize(vm, 2);
        bool ret = (thissize >= strlen) && !strncmp(thisstr, str, strlen);
        sq_pushbool(vm, ret);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "ltrim"
    static SQInteger Ltrim(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        const SQChar *ret = thisstr;
        while ((*ret) != '\0' && isspace(*ret)) {
            ret++;
        }
        sq_pushstring(vm, ret, -1);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "rtrim"
    static SQInteger Rtrim(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        SQInteger newlen = sq_getsize(vm, 1);
        while (newlen > 0 && isspace(thisstr[newlen - 1])) {
            newlen--;
        }
        sq_pushstring(vm, thisstr, newlen);
        return 1;
    }
#undef BIND_METHOD_NAME
#define BIND_METHOD_NAME "trim"
    static SQInteger Trim(HSQUIRRELVM vm) {
        BIND_GET_NARGS()
        BIND_NARGS_CHECK_EXACT(0)
        const SQChar *thisstr; // get "this" pointer as string
        sq_getstring(vm, 1, &thisstr);
        const SQChar *ret = thisstr;
        SQInteger newlen = sq_getsize(vm, 1);
        while (*ret != '\0' && isspace(*ret)) {
            ret++;
            newlen--;
        }
        while (newlen > 0 && isspace(ret[newlen - 1])) {
            newlen--;
        }
        sq_pushstring(vm, ret, newlen);
        return 1;
    }

#undef BIND_METHOD_NAME
#undef BIND_CLASS_NAME
};

} // namespace bipscript

#endif // STRINGTYPE_H
