﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CLOCK_H
#define CLOCK_H

#include "collector.h"
#include "memorypool.h"
#include "queuelist.h"
#include "refcountable.h"
#include "sortedqueuelist.h"
#include "types.h"

namespace bipscript {
namespace time {

/**
 * Represents a musical time signature.
 */
class Signature : public RefCountable<Signature> {
    uint16_t mnumerator;
    uint16_t mdenominator;

  public:
    Signature(uint16_t numerator, uint16_t denominator)
        : mnumerator(numerator), mdenominator(denominator) {}
    Signature() : mnumerator(4), mdenominator(4) {}
    uint16_t numerator() { return mnumerator; }
    void numerator(uint16_t numerator) { mnumerator = numerator; }
    uint16_t denominator() { return mdenominator; }
    void denominator(uint16_t denominator) { mdenominator = denominator; }
    float beatsPerMeasure() { // TODO: account for compound signatures eg 6/8 ?
        return 4.0 * mnumerator / mdenominator;
    }
};

/**
 * Represents a period of clock time with constant time signature and tempo.
 * An audio period will have more than one window if e.g. a tempo change is
 * made on a particular frame within that period.
 */
class Window {
    static samples_t msampleRate; // samples per second
    double mmeasure;
    samples_t mframe;
    samples_t mstart;
    samples_t msize;
    Signature msignature;
    float mbpm;
    float framesPerMeasure;
    bool mreposition;

  public:
    Window *next;
    // static methods
    static void sampleRate(samples_t value) { msampleRate = value; }
    static samples_t sampleRate() { return msampleRate; }
    // ctor
    Window()
        : mmeasure(1.0), mframe(0), mstart(0), msize(0), mbpm(120.0f), mreposition(true),
          next(nullptr) {
        calculateFpm();
    }
    // getters and setters
    double measure() const { return mmeasure; } // static_cast<float>(mmeasure); }
    void measure(double measure) { mmeasure = measure; }
    samples_t frame() const { return mframe; }
    void frame(samples_t frame) { mframe = frame; }
    samples_t start() { return mstart; }
    void start(samples_t start) { mstart = start; }
    samples_t size() const { return msize; }
    void size(samples_t size) { msize = size; }
    Signature &signature() { return msignature; }
    void signature(Signature &signature) {
        msignature = signature;
        calculateFpm();
    }
    float bpm() { return mbpm; }
    void bpm(float bpm) {
        mbpm = bpm;
        calculateFpm();
    }
    bool reposition() { return mreposition; }
    void reposition(bool reposition) { mreposition = reposition; }
    /**
     * Recalculated frames per measure.
     */
    void calculateFpm() {
        framesPerMeasure = msampleRate * 60 * msignature.beatsPerMeasure() / mbpm;
    }
    /**
     * returns the frame offset from the start of this window
     */
    samples_t offset(float measure) {
        float delta = measure - this->mmeasure;
        return delta < 0 ? 0 : static_cast<samples_t>(delta * framesPerMeasure + 0.5f);
    }
    /**
     * returns the offset in measures from the start of this window
     */
    double moffset(samples_t frames) { return static_cast<double>(frames / framesPerMeasure); }
    // debug
    void print() {
        printf("%u @%f %.1fbpm %d/%d %d-%d\n", mframe, mmeasure, mbpm, msignature.numerator(),
               msignature.denominator(), mstart, mstart + msize);
    }
};

/**
 * Superclass for all instant clock change requests
 */
class ClockChange {
  public:
    enum Type { RELOCATION, RESTART, SIGNATURE, STATE, TEMPO };

  private:
    Type mtype;

  public:
    ClockChange(enum Type type) : mtype(type) {}
    Type type() { return mtype; }
};

class Relocation : public ClockChange {
    float mlocation;

  public:
    Relocation(float location) : ClockChange(Type::RELOCATION), mlocation(location) {
        if (location < 1.0) {
            throw std::logic_error("cannot reposition before 1.0");
        }
    }
    float location() { return mlocation; }
};

class Restart : public ClockChange {
  public:
    Restart() : ClockChange(Type::RESTART) {}
};

class SignatureChange : public ClockChange {
    Signature msignature;

  public:
    SignatureChange(Signature *signature) : ClockChange(Type::SIGNATURE), msignature(*signature) {}
    Signature &signature() { return msignature; }
};

class StateChange : public ClockChange {
    bool mstart;

  public:
    StateChange(bool start) : ClockChange(Type::STATE), mstart(start) {}
    bool start() { return mstart; }
};

class TempoChange : public ClockChange {
    float mtempo;

  public:
    TempoChange(float tempo) : ClockChange(Type::TEMPO), mtempo(tempo) {}
    float tempo() { return mtempo; }
};

/**
 * Iterator for time windows in the current period. This local version is used
 * in the process thread.
 */
class Timer {
    Window *top;
    // running flag
    bool mrunning;

  public:
    Timer(Window *top) : top(top), mrunning(false) {}
    /**
     * Returns the first time window for the current clock period.
     */
    virtual Window *firstpos() {
        //        printf("%c P %c %d [%d] @%f %.2fbpm %d/%d %p\n", mrunning ? '+' : '-',
        //               top->reposition() ? '!' : '>', top->frame(), top->size(), top->measure(),
        //               top->bpm(), top->signature().numerator(), top->signature().denominator(),
        //               top);
        return top;
    }
    /**
     * Returns the following time window or null if there are no more time windows in this clock
     * period.
     */
    virtual Window *nextpos(Window *window) {
        //        Window *ret = window->next;
        //        if (ret) {
        //            printf("%c - %c %d [%d] @%f %.2fbpm %d/%d %p\n", mrunning ? '+' : '-',
        //                   ret->reposition() ? '!' : '>', ret->frame(), ret->size(),
        //                   ret->measure(), ret->bpm(), ret->signature().numerator(),
        //                   ret->signature().denominator(), ret);
        //        }
        return window->next;
    }
    /**
     * Returns true when the timer has completed its lifecycle, useful for remote timers only
     */
    virtual bool finished() { return true; }
    void running(bool value) { mrunning = value; }
    bool running() { return mrunning; }
    float offsetToMeasure(samples_t offset);
};

/**
 * Window sent to remote timers via thread queue. Created in the process
 * thread, uses memory pool for allocation.
 */
class RemoteWindow : public Window {
    static Collector<RemoteWindow> collector;

  public:
    static void recycle(RemoteWindow *window) { collector.recycle(window); }
    static void free() { collector.free(); }
    RemoteWindow(Window &window) : Window(window) {}
    void *operator new(size_t size) { return MemoryPool::processPool().malloc(size); }
    void operator delete(void *p) { MemoryPool::processPool().free(p); }
};

/**
 *  Remote iterator for time windows in the current clock period.
 *  Used by remotetimeset to give timing to another thread. Note instances
 *  are created by the associated clock but recycled by the client timeline.
 */
class RemoteTimer : public Timer, public Listable<RemoteTimer> {
    Timer &original;
    ThreadQueue<RemoteWindow> queue;
    std::atomic<bool> mfinished;
    RemoteWindow *last;

  public:
    RemoteTimer(Timer &orig)
        : Timer(nullptr), original(orig), queue(256), mfinished(false), last(nullptr) {}
    void finish() { mfinished.store(true); }
    bool finished() override;
    void update();
    // clock internal timeline API
    Window *firstpos() override;
    Window *nextpos(Window *) override;
};

/**
 * Superclass for all clocks, provides timing infomation to timelines.
 * Current clients include Timeline, BufferEventTimeline.
 */
class Clock : public Listable<Clock> {
    static Clock *defclock;   // default clock
    static samples_t mperiod; // audio engine period size
    QueueList<RemoteTimer> remoteTimers;

  protected:
    Window current;
    std::atomic<float> mtempo; // updated at the start of each period
    ThreadQueue<ClockChange> commands;
    Timer mtimer;
    uint16_t ticksPerBeat = 1920;
    /**
     * Update all remote timers for the current period.
     */
    void updateRemotes();
    void clearRemotes() { remoteTimers.clear(); }
    void validateTempo(float bpm) {
        if (bpm <= 0.0) {
            throw std::logic_error("Clock tempo must be positive and non-zero");
        }
    }

  public:
    static Clock &defaultClock();
    static void defaultClock(Clock *clock);
    static samples_t period() { return mperiod; }
    static void period(samples_t nframes) { mperiod = nframes; }
    Clock(float bpm) : remoteTimers(32), mtempo(bpm), commands(8), mtimer(&current) {
        validateTempo(bpm);
        current.bpm(bpm);
    }
    virtual ~Clock() {}
    float tempo() { return mtempo.load(); }
    void tempo(float bpm) {
        validateTempo(bpm);
        commands.add(new TempoChange(bpm));
    }
    /**
     * Script API method: start the clock running
     */
    virtual void start() = 0;
    /**
     * Script API method: stop the clock
     */
    virtual void stop() = 0;
    // internal API (process thread)
    Timer &timer() { return mtimer; }
    void running(bool value);
    virtual void update(samples_t time) = 0;
    /**
     * Returns true if this clock can map previous measures back to frames.
     */
    virtual bool hasHistory() = 0;
    /**
     * Map a given measure back to a frame location.
     */
    virtual samples_t toFrame(float measure) { return 0; }
    /**
     * Create and return a new remote timer for this clock.
     */
    RemoteTimer &remoteTimer() {
        RemoteTimer *timer = new RemoteTimer(mtimer);
        remoteTimers.add(timer);
        return *timer;
    }
    // end of lifecycle
    void finished();
    virtual void recycle() = 0;
    Clock *addref() { return this; } // needed for bindings
};

/**
 * Time API package class
 */
class Time {
  public:
    /**
     * The default clock.
     */
    static Clock *def() { return &Clock::defaultClock(); }
    static void def(Clock *clock) { Clock::defaultClock(clock); }
};

} // namespace time
} // namespace bipscript

#endif // CLOCK_H
