/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "link.h"

#include "clockcollector.h"
#include <iomanip>

using namespace std::chrono;

namespace bipscript {
namespace time {

Link::Link() : Clock(120), mlink(120), nextEvent(nullptr), initbpm(0), mframe(0), mquantum(4) {
    mlink.enable(true);
    mlink.enableStartStopSync(true);
}

Link::Link(float bpm, int quantum)
    : Clock(bpm), mlink(bpm), nextEvent(nullptr), initbpm(bpm), mframe(0), mquantum(quantum) {
    mlink.enable(true);
    mlink.enableStartStopSync(true);
    commands.add(new TempoChange(bpm));
    startWindow.bpm(bpm);
}

Link::Link(float bpm, Signature &signature) : Link(bpm, signature.beatsPerMeasure()) {
    float beats = signature.beatsPerMeasure();
    if (ceilf(beats) != beats) {
        throw std::logic_error("Time signature must have integer beats per measure");
    }
    current.signature(signature);
    startWindow.signature(signature);
}

Link::~Link() {
    mlink.enableStartStopSync(false);
    mlink.enable(false);
}

void Link::start() { commands.add(new StateChange(true)); }

void Link::stop() { commands.add(new StateChange(false)); }

void Link::update(samples_t time) {

    // grab link state
    ableton::Link::SessionState sessionState = mlink.captureAudioSessionState();
    // host time ms
    microseconds hostTime = mHostTimeFilter.sampleTimeToHostTime(time);

    // apply immediate request to change tempo, running state
    ClockChange *command;
    bool statusupdate = false;
    while (commands.pop(command)) {
        if (command->type() == ClockChange::Type::STATE) {
            sessionState.setIsPlaying(static_cast<StateChange *>(command)->start(), hostTime);
        } else if (command->type() == ClockChange::Type::TEMPO) {
            sessionState.setTempo(static_cast<TempoChange *>(command)->tempo(), hostTime);
        }
        ClockCollector::instance().recycle(command);
        statusupdate = true;
    }

    // update timer running state if changed
    current.reposition(false);
    bool playing = mtimer.running();
    if (playing != sessionState.isPlaying()) {
        if (!playing) {
            sessionState.requestBeatAtStartPlayingTime(0, mquantum);
            current.reposition(true);
            mframe = 0;
        }
        mtimer.running(sessionState.isPlaying());
        statusupdate = true;
    }

    // latency from audio engine
    samples_t latency = AudioEngine::instance().latency();
    microseconds mslatency =
        std::chrono::microseconds(llround(1.0e6 * latency / current.sampleRate()));

    // get beat from link, convert to measures
    const double linkBeat = sessionState.beatAtTime(hostTime + mslatency, mquantum);
    const double measure = 1.0 + linkBeat / mquantum;

    // set period initial bpm directly from link
    float tempo = sessionState.tempo();
    if (tempo != mtempo.load()) {
        mtempo.store(tempo);
        statusupdate = true;
    }
    current.bpm(tempo);

    // note bar change for status reporting
    uint32_t bar = static_cast<uint32_t>(measure);
    bool barflip = bar != static_cast<uint32_t>(current.measure());

    // update and reset event list
    if (eventBuffer.update()) {
        nextEvent = eventBuffer.firstEvent();
    }

    if (mtimer.running()) {
        samples_t range = period();

        // update measure and frame
        current.measure(measure);
        current.frame(mframe);

        // skip over past events
        while (nextEvent && nextEvent->measure() < measure) {
            nextEvent = eventBuffer.nextEvent(nextEvent);
        }

        // deal with scheduled events for this period
        Window *last = &current;
        while (nextEvent && nextEvent->offset(last) < range) {
            // update window attached to this event
            Window *window = nextEvent->window();
            samples_t subrange = nextEvent->start();
            window->start(period() - (range - subrange));
            // event-specific update
            if (nextEvent->type() == ClockChangeEvent::TEMPO) {
                // set frame via offset from last
                window->frame(last->frame() + subrange);
                window->signature(last->signature());
                // update link
                microseconds changeTime =
                    mHostTimeFilter.sampleTimeToHostTime(time + nextEvent->start());
                sessionState.setTempo(window->bpm(), changeTime);
            } else if (nextEvent->type() == ClockChangeEvent::SIGNATURE) {
                //                window->frame(last->frame() + subrange);
                //                window->bpm(last->bpm());
            }
            // set size and next on last pos
            last->size(subrange);
            last->next = window;
            // shrink range
            range -= subrange;
            mframe += subrange;
            // iterate
            nextEvent = eventBuffer.nextEvent(nextEvent);
            last = window;
            statusupdate = true;
        }
        last->size(range);
        last->next = nullptr;

        // deal with count-in
        if (measure < 1.0) {
            current.size(0);
            samples_t zeroframe = current.offset(1.0); // frames till 1.0
            if (zeroframe < range) {                   // 1.0 happens in this period
                // set initial bpm if any
                if (initbpm) {
                    sessionState.setTempo(initbpm, hostTime);
                }
                // reset next event
                nextEvent = eventBuffer.firstEvent();
                // chop up and add extra window
                current.next = &startWindow;
                startWindow.start(zeroframe);
                uint32_t size = range - zeroframe;
                startWindow.size(size);
                statusupdate = true;
                mframe += size;
            }
        } else {
            mframe += range;
        }
    }

    // commit session
    mlink.commitAudioSessionState(sessionState);

    // update status server
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        if (statusupdate) {
            time::Signature &sig = current.signature();
            statusserver.clockUpdate(this, mtimer.running(), current.measure(), current.bpm(),
                                     sig.numerator(), sig.denominator());
        }
        if (barflip) {
            statusserver.clockTime(this, current.measure());
        }
    }
    updateRemotes();
}

// process thread
void Link::recycle() {
    // recycle events
    nextEvent = eventBuffer.firstEvent();
    while (nextEvent) {
        ClockChangeEvent *done = nextEvent;
        nextEvent = eventBuffer.pop();
        ClockCollector::instance().recycle(done);
    }
    // clear remote timers
    clearRemotes();
}

Link *LinkFactory::newLink() {
    AudioEngine &engine = AudioEngine::instance();
    if (!mlink) {
        mlink = new Link();
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(mlink, "Link", mlink->tempo());
        }
    }
    engine.addClock(mlink);
    return mlink;
}

Link *LinkFactory::newLink(float bpm) {
    AudioEngine &engine = AudioEngine::instance();
    if (!mlink) {
        mlink = new Link(bpm);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(mlink, "Link", bpm);
        }
    } else {
        mlink->tempo(bpm);
    }
    engine.addClock(mlink);
    return mlink;
}

Link *LinkFactory::newLink(float bpm, Signature &sig) {
    AudioEngine &engine = AudioEngine::instance();
    if (!mlink) {
        mlink = new Link(bpm, sig);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(mlink, "Link", bpm, sig.numerator(), sig.denominator());
        }
    } else {
        // TODO: how do we handle this
    }
    engine.addClock(mlink);
    return mlink;
}

} // namespace time
} // namespace bipscript
