/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef VARIABLECLOCK_H
#define VARIABLECLOCK_H

#include "audioengine.h"
#include "clock.h"
#include "clockchangeevent.h"
#include "clockcollector.h"
#include "statusserver.h"

namespace bipscript {
namespace time {

class AdvancedClock : public Clock {
    SortedQueueList<ClockChangeEvent> eventBuffer;
    ClockChangeEvent *nextEvent;
    Window initial;
    Window following;

  protected:
    float tempoChange; // TODO: can we remove this
    AdvancedClock(float bpm) : Clock(bpm), nextEvent(nullptr), tempoChange(-1.0) {
        initial.bpm(bpm);
        following.bpm(bpm);
    }
    AdvancedClock(float bpm, Signature signature)
        : Clock(bpm), nextEvent(nullptr), tempoChange(-1.0) {
        initial.bpm(bpm);
        following.bpm(bpm);
        initial.signature(signature);
        following.signature(signature);
    }
    ~AdvancedClock();
    void relocate(samples_t location);
    void relocate(float measure);
    void tick();
    void reposition(float location, float measure) { // scheduled reposition
        eventBuffer.addEvent(ClockChangeEvent::relocationEvent(location, measure));
    }
    void statusUpdate(float bpm);

  public:
    void schedule(float tempo, float measure) {
        validateTempo(tempo);
        eventBuffer.addEvent(ClockChangeEvent::tempoChangeEvent(tempo, measure));
    }
    void schedule(Signature &signature, unsigned int bar) {
        eventBuffer.addEvent(ClockChangeEvent::timeChangeEvent(signature, bar));
    }
    // internal API
    bool hasHistory() override { return true; }
    samples_t toFrame(float measure) override;

  private:
    void calculatePeriod();
    ClockChangeEvent *calculateWindow(Window *);
};

class VariableClock : public AdvancedClock {
  public:
    VariableClock(float bpm) : AdvancedClock(bpm) {}
    VariableClock(float bpm, Signature &signature) : AdvancedClock(bpm, signature) {}
    void update(samples_t) override;
    void start() override { commands.add(new StateChange(true)); }
    void stop() override { commands.add(new StateChange(false)); }
    // immediate reposition request
    void reposition(float location) { commands.add(new Relocation(location)); }
    // scheduled reposition
    void reposition(float location, float measure) { AdvancedClock::reposition(location, measure); }
    void recycle() override {
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.finishClock(this);
        }
        ClockCollector::instance().recycle(this);
    }
};

class VariableClockFactory {
  public:
    static VariableClockFactory &instance() {
        static VariableClockFactory instance;
        return instance;
    }
    VariableClock *variableClock(float bpm) {
        AudioEngine &engine = AudioEngine::instance();
        VariableClock *clock = new VariableClock(bpm);
        engine.addClock(clock);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(clock, "var", bpm);
        }
        return clock;
    }
    VariableClock *variableClock(float bpm, Signature signature) {
        AudioEngine &engine = AudioEngine::instance();
        VariableClock *clock = new VariableClock(bpm, signature);
        engine.addClock(clock);
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.newClock(clock, "var", bpm, signature.numerator(),
                                  signature.denominator());
        }
        return clock;
    }
};

} // namespace time
} // namespace bipscript

#endif // VARIABLECLOCK_H
