﻿/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef TIMELINE_H
#define TIMELINE_H

#include "clock.h"
#include "clockcollector.h"
#include "event.h"
#include "listable.h"
#include "scriptcollector.h"

#include <map>

#define UPDATE_MAX_EVENTS 256 // TODO: function of period size

namespace bipscript {

using time::Clock;
using time::Timer;

template <class T> class RemoveCommand : public Command {
    Event<T> *mevent;

  public:
    RemoveCommand(Event<T> *event) : Command(REMOVE), mevent(event) {}
    Event<T> *event() { return mevent; }
};

/**
 * Template class that holds events in a sorted list with associated clock.
 * Contains methods to allow accessing events at the scheduled clock time.
 */
template <class T> class Timeline : public Listable<Timeline<T>> {
    ThreadQueue<Command> eventQueue; // script thread -> process thread
    List<Event<T>> sortedEvents;     // local to process thread
    List<Event<T>> finishedEvents;   // local to process thread
    bool eventsPending;
    Timer &mtimer;
    time::Window *window;
    bool mfinished;
    Event<T> *last;
    Event<T> *mnextEvent; // next event in this period or null if none
    Collector<Command> cmdcollector;
    RefCollector<Event<T>> evtcollector;

  public:
    Timeline(Timer &clock)
        : mtimer(clock), window(nullptr), eventQueue(2048), mfinished(false), last(nullptr),
          mnextEvent(nullptr) {
        CollectorThread::instance().addCollector(&cmdcollector);
        // collect events in script thread
        ScriptCollector::instance().addCollector(&evtcollector);
    }
    ~Timeline();
    Timer &timer() { return mtimer; }
    // script thread
    void add(Command *cmd) { eventQueue.add(cmd); }
    // process thread
    void startPeriod() { mnextEvent = periodFirst(); }
    void popEvent() { mnextEvent = periodNext(); }
    Event<T> *nextEvent() { return mnextEvent; }
    bool finished() { return mfinished; }
    bool operator>(Timeline<T> &other);

  private:
    Event<T> *periodFirst();
    Event<T> *periodNext();
    int update();
};

template <class T> Event<T> *Timeline<T>::periodFirst() {
    bool newEvents = update();
    window = mtimer.firstpos();
    while (window) {
        if (window->reposition()) {
            last = sortedEvents.first();
            while (last && last->measure() < window->measure()) {
                last = sortedEvents.next(last);
            }
        } else if (newEvents) {
            last = mnextEvent ? mnextEvent : sortedEvents.first();
            while (last && last->measure() < window->measure()) {
                last = sortedEvents.next(last);
            }
        }
        if (!mtimer.running()) {
            return nullptr;
        }
        if (last && last->within(window)) {
            if (last->etype() == Command::TRANSIENT) {
                sortedEvents.remove(last);
            }
            return last;
        }
        mfinished = !last && !eventsPending;
        window = mtimer.nextpos(window);
    }
    return nullptr;
}

template <class T> Event<T> *Timeline<T>::periodNext() {
    last = sortedEvents.next(last);
    if (last && last->within(window)) {
        if (last->etype() == Command::TRANSIENT) {
            sortedEvents.remove(last);
        }
        return last;
    }
    // no events left in this position, try next
    window = mtimer.nextpos(window);
    while (window) {
        if (window->reposition()) {
            last = sortedEvents.first();
            while (last && last->measure() < window->measure()) {
                last = sortedEvents.next(last);
            }
        }
        if (last && last->within(window)) {
            return last;
        }
        window = mtimer.nextpos(window);
    }
    mfinished = !last && !eventsPending;
    return nullptr;
}

/**
 * Pull new events from the script thread and add them to the sorted event
 * list. Runs in the process thread.
 */
template <class T> int Timeline<T>::update() {
    // remove
    Event<T> *evt = finishedEvents.first();
    while (evt) {
        Event<T> *done = evt;
        evt = finishedEvents.pop();
        evtcollector.recycle(done); // TODO: send list instead of one at a time
    }
    // run commands
    Command *cmd;
    int counter = 0;
    while (eventQueue.pop(cmd)) {
        if (cmd->etype() == Command::CLEAR) {
            Event<T> *top = sortedEvents.clear();
            while (top) {
                Event<T> *done = top;
                top = top->mnext;
                evtcollector.recycle(done); // TODO: send list instead of one at a time
            }
            last = nullptr;
            cmdcollector.recycle(cmd);
        } else if (cmd->etype() == Command::REMOVE) {
            RemoveCommand<T> *rcmd = static_cast<RemoveCommand<T> *>(cmd);
            Event<T> *evt = rcmd->event();
            if (evt == last) {
                last = sortedEvents.next(last);
            }
            if (sortedEvents.remove(evt)) {
                evtcollector.recycle(evt);
            }
            cmdcollector.recycle(cmd);
        } else {
            sortedEvents.insert(static_cast<Event<T> *>(cmd));
            // TODO: should be factor of existing list size
            if (++counter == UPDATE_MAX_EVENTS) {
                eventsPending = true;
                return counter;
            }
        }
    }
    eventsPending = false;
    return counter;
}

template <class T> Timeline<T>::~Timeline() {
    CollectorThread::instance().removeCollector(&cmdcollector);
    ScriptCollector::instance().removeCollector(&evtcollector);
    Event<T> *evt = sortedEvents.first();
    while (evt) {
        Event<T> *done = evt;
        evt = sortedEvents.pop();
        if (done->delref() == 0) {
            delete done;
        }
    }
}

template <class T> bool Timeline<T>::operator>(Timeline<T> &other) {
    if (!mnextEvent) {
        return true;
    }
    if (!other.mnextEvent) {
        return false;
    }
    return mnextEvent->start() > other.mnextEvent->start();
}

/**
 * An iterable set of timelines (of class T) that dispatches events to each timeline by clock
 */
template <class T> class TimelineSet {
    Timeline<T> *topline;
    Collector<Timeline<T>> collector;

  protected:
    std::map<Clock *, Timeline<T> *> timelines;
    QueueList<Timeline<T>> activeList;

  public:
    TimelineSet() : activeList(16) { ScriptCollector::instance().addCollector(&collector); }
    void addEvent(Event<T> *evt, Clock &clock) { timeline(clock)->add(evt->addref()); }
    Event<T> *createEvent(T *mesg, float measure, Clock &clock);
    void removeEvent(Event<T> *event) {
        timeline(event->clock())->add(new RemoveCommand<T>(event));
    }
    Event<T> *sendEvent(T *mesg, float measure, Clock &clock);
    void clear(Clock &clock) { timeline(clock)->add(new Command(Command::CLEAR)); }
    Event<T> *firstEvent();
    Event<T> *nextEvent();
    bool active();
    bool restart();
    void restartScript() { timelines.clear(); }

  private:
    Timeline<T> *timeline(Clock &clock);
};

/**
 * Add a new event to the timeline, timed by the given clock. Runs in the
 * script thread.
 */
template <class T> Event<T> *TimelineSet<T>::createEvent(T *mesg, float measure, Clock &clock) {
    Event<T> *evt = new Event<T>(mesg, Command::Type::SCHEDULE, measure, clock);
    timeline(clock)->add(evt->addref());
    return evt;
}

/**
 * Add a new transient event to the timeline, timed by the given clock. Runs in the
 * script thread.
 */
template <class T> Event<T> *TimelineSet<T>::sendEvent(T *mesg, float measure, Clock &clock) {
    Event<T> *evt = new Event<T>(mesg, Command::Type::TRANSIENT, measure, clock);
    timeline(clock)->add(evt->addref());
    return evt;
}

/**
 * Return the first event for this period for this set of timelines.
 * Collates timelines with multiple clocks. Runs in the process thread.
 */
template <class T> Event<T> *TimelineSet<T>::firstEvent() {
    Timeline<T> *timeline = activeList.first();
    while (timeline) {
        timeline->startPeriod();
        timeline = activeList.next(timeline);
    }
    topline = activeList.sort();
    return topline ? topline->nextEvent() : nullptr;
}

/**
 * Return the next event for this period for this set of timelines.
 * Collates timelines with multiple clocks. Runs in the process thread.
 */
template <class T> Event<T> *TimelineSet<T>::nextEvent() {
    topline->popEvent();
    topline = activeList.sort();
    return topline ? topline->nextEvent() : nullptr;
}

/**
 * Returns false if all events in all timelines have passed.
 */
template <class T> bool TimelineSet<T>::active() {
    Timeline<T> *timeline = activeList.first();
    while (timeline) {
        if (!timeline->finished()) {
            return true;
        }
        timeline = activeList.next(timeline);
    }
    return false;
}

/**
 * Runs in the process thread.
 */
template <class T> bool TimelineSet<T>::restart() {
    Timeline<T> *timeline = activeList.first();
    while (timeline) {
        Timeline<T> *done = timeline;
        if (!done->timer().finished()) {
            return false;
        }
        timeline = activeList.next(timeline);
        activeList.remove(done);
        collector.recycle(done);
    }
    return true;
}

template <class T> Timeline<T> *TimelineSet<T>::timeline(time::Clock &clock) {
    Timeline<T> *timeline = timelines[&clock];
    if (timeline == nullptr) {
        timeline = new Timeline<T>(clock.timer());
        activeList.add(timeline);
        timelines[&clock] = timeline;
    }
    return timeline;
}

} // namespace bipscript

#endif // TIMELINE_H
