/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "xtprocessor.h"

#include "audioengine.h"
#include "midi/midievent.h"

namespace bipscript {

namespace audio {
namespace binding {
audio::Sink *resolve_sink(SQUserPointer obj) { return XtAudioProcessorMap::instance().find(obj); }
audio::Source *resolve_source(SQUserPointer obj) {
    return XtAudioProcessorMap::instance().find(obj);
}
} // namespace binding
} // namespace audio

namespace midi {
namespace binding {
midi::Sink *resolve_sink(SQUserPointer obj) { return XtMidiProcessorMap::instance().find(obj); }
midi::Source *resolve_source(SQUserPointer obj) { return XtMidiProcessorMap::instance().find(obj); }
} // namespace binding
} // namespace midi

XtAudioProcessor::XtAudioProcessor(BXTInstance instance, const char *name, uint32_t inputs,
                                   uint32_t outputs, BXTAudioProcess func)
    : audio::Sink(inputs), audio::Source(outputs), mprocessfunc(func), minstance(instance) {
    inputaudio = new float *[inputs];
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.newProcessor(this, name, inputs, outputs);
        reportConnections(this);
        reportConnectors(this);
    }
}

bool XtAudioProcessor::doProcess(samples_t time) {
    // process audio inputs
    samples_t nframes = Clock::period();
    for (uint32_t i = 0; i < audio::Sink::channels(); i++) {
        inputaudio[i] = Sink::inputaudio(i, time, nframes);
    }
    bool active = mprocessfunc(nframes, minstance);
    reportLevels(this, time);
    return audio::Sink::activeEvents() || active;
}

XtAudioProcessor *XtAudioProcessorMap::add(BXTInstance instance, const char *name, uint32_t inputs,
                                           uint32_t outputs, BXTAudioProcess func) {
    AudioEngine &engine = AudioEngine::instance();
    XtAudioProcessor *processor = new XtAudioProcessor(instance, name, inputs, outputs, func);
    engine.addProcessor(processor);
    mmap[instance] = processor;
    return processor;
}

void XtAudioProcessorMap::reset(bool restart) {
    if (mmap.size()) {
        AudioEngine &engine = AudioEngine::instance();
        StatusServer &statusserver = StatusServer::instance();
        for (auto const &entry : mmap) {
            XtAudioProcessor *processor = entry.second;
            engine.removeProcessor(processor);
            if (restart && statusserver.active()) {
                statusserver.removeProcessor(processor);
            }
        }
        mmap.clear();
    }
}

XtMidiProcessor::XtMidiProcessor(BXTInstance instance, const char *name, uint32_t inputs,
                                 uint32_t outputs)
    : midi::Sink(inputs), midi::Source(outputs), minstance(instance), mevent(nullptr) {
    maudioprocessor = XtAudioProcessorMap::instance().find(instance);
    StatusServer &statusserver = StatusServer::instance();
    if (statusserver.active()) {
        statusserver.addProcessorMidi(maudioprocessor, inputs, outputs);
        reportConnections(maudioprocessor);
        reportConnectors(maudioprocessor);
    }
}

uint32_t XtMidiProcessor::nextEvent(BXTIndex index, uint32_t *offset, unsigned char *buffer,
                                    uint32_t bufsize) {
    // TODO: index?
    if (mevent) {
        midi::RawMessage &mesg = mevent->message();
        *offset = mevent->start();
        uint16_t size = mesg.size();
        if (size <= bufsize) {
            std::memcpy(buffer, (unsigned char *)mesg.bytes(), size);
        }
        mevent = mconnector->nextEvent();
        return size;
    }
    return 0;
}

bool XtMidiProcessor::doProcess(samples_t time) {
    midi::Sink::process(time);
    mconnector = &Sink::connector(0);
    mevent = mconnector->firstEvent();
    //    if (StatusServer::instance().active()) {
    //        mmeter.update(this, 0, event != nullptr, time);
    //    }
    return mconnector->active();
}

XtMidiProcessor *XtMidiProcessorMap::add(XtAudioProcessor *audioproc, uint32_t inputs,
                                         uint32_t outputs) {
    // create processor and add to audio engine and instance map
    AudioEngine &engine = AudioEngine::instance();
    BXTInstance instance = XtAudioProcessorMap::instance().findByProcessor(audioproc);
    XtMidiProcessor *processor = new XtMidiProcessor(instance, nullptr, inputs, outputs);
    engine.addProcessor(processor);
    mmap[instance] = processor;
    // is there an audio processor for this instance?
    return processor;
}

void XtMidiProcessorMap::reset(bool restart) {
    if (mmap.size()) {
        AudioEngine &engine = AudioEngine::instance();
        StatusServer &statusserver = StatusServer::instance();
        for (auto const &entry : mmap) {
            XtMidiProcessor *processor = entry.second;
            engine.removeProcessor(processor);
            if (restart && statusserver.active()) {
                statusserver.removeProcessor(processor);
            }
        }
        mmap.clear();
    }
}

} // namespace bipscript
