/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef XTPROCESSOR_H
#define XTPROCESSOR_H

#include "audiosink.h"
#include "bipxtapi.h"
#include "midisink.h"
#include "objectcache.h"
#include "processorthread.h"

namespace bipscript {

class XtProcessor : public ProcessorThread {
    BXTProcess mprocess;
    BXTProcess mrestart;
    void *marg;

    const static int DEF_PERIOD = 5;

  public:
    XtProcessor(BXTBoolean resetproof, BXTProcess process, BXTProcess restart, void *arg)
        : ProcessorThread(resetproof, DEF_PERIOD), mprocess(process), mrestart(restart), marg(arg) {
        start();
    }
    bool process() override { return mprocess(marg); }
    bool restarted() override { return mrestart(marg); }
};

struct XtMessage {
    void *mdata;
    XtMessage(void *data) : mdata(data) {}
};

typedef TimelineSet<XtMessage> XtTimelineSet;

class XtAudioProcessor : public audio::Sink, public audio::Source {
    BXTAudioProcess mprocessfunc;
    BXTInstance minstance;
    float **inputaudio; // pointers to input buffers

  public:
    XtAudioProcessor(BXTInstance instance, const char *name, uint32_t inputs, uint32_t outputs,
                     BXTAudioProcess func);
    float *audioinputbuffer(int i) { return inputaudio[i]; }
    // processor api
    bool doProcess(samples_t time) override;
    void restart() override {}
    bool connectsTo(AbstractSink *sink) override {
        return sink == this || audio::Sink::connectsTo(sink);
    }
};

class XtAudioProcessorMap : public ObjectCache {
    std::map<BXTInstance, XtAudioProcessor *> mmap;
    XtAudioProcessorMap() {}
    virtual ~XtAudioProcessorMap() {}

  public:
    static XtAudioProcessorMap &instance() {
        static XtAudioProcessorMap instance;
        return instance;
    }
    XtAudioProcessor *find(BXTInstance instance) {
        return mmap.find(instance) != mmap.end() ? mmap[instance] : nullptr;
    }
    BXTInstance findByProcessor(XtAudioProcessor *proc) {
        for (auto it = mmap.begin(); it != mmap.end(); ++it) {
            if (it->second == proc) {
                return it->first;
            }
        }
        return nullptr;
    }
    XtAudioProcessor *add(BXTInstance instance, const char *name, uint32_t inputs, uint32_t outputs,
                          BXTAudioProcess func);
    void scriptComplete() override {}
    void reset(bool restart) override;
};

class XtMidiProcessor : public midi::Sink, public midi::Source {
    BXTInstance minstance;
    midi::Connector *mconnector;
    midi::RawEvent *mevent; // next event
    XtAudioProcessor *maudioprocessor;

  public:
    XtMidiProcessor(BXTInstance instance, const char *name, uint32_t inputs, uint32_t outputs);
    uint32_t nextEvent(BXTIndex index, uint32_t *offset, unsigned char *buffer, uint32_t insize);
    // processor api
    bool doProcess(samples_t) override;
    void restart() override {}
    bool connectsTo(AbstractSink *sink) override {
        return sink == this || midi::Sink::connectsTo(sink);
    }
};

class XtMidiProcessorMap : public ObjectCache {
    std::map<BXTInstance, XtMidiProcessor *> mmap;
    XtMidiProcessorMap() {}
    virtual ~XtMidiProcessorMap() {}

  public:
    static XtMidiProcessorMap &instance() {
        static XtMidiProcessorMap instance;
        return instance;
    }
    XtMidiProcessor *find(BXTInstance instance) {
        return mmap.find(instance) != mmap.end() ? mmap[instance] : nullptr;
    }
    XtMidiProcessor *add(XtAudioProcessor *audioproc, uint32_t inputs, uint32_t outputs);
    void scriptComplete() override {}
    void reset(bool restart) override;
};

} // namespace bipscript

#endif // XTPROCESSOR_H
